           (*-------------------------------------------*
            |                CSP-Prover                 |
            |                April 2005                 |
            |               June   2005  (modified)     |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Alpha_parallel = CSP_proc:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*********************************************************
       Preparation (traces operated by par and hide)
 *********************************************************)

(*** par hide ***)

(*** if ***)

lemma par_tr_hide_tr_if_all:
  "u ~= None & sett(u) <= insert Tick (Ev ` (X Un Y))
   --> u : u --tr (- X) |[X Int Y]|tr u --tr (- Y)"
apply (induct_tac u rule: induct_trace)
apply (simp_all)
apply (simp add: par_tr_head)

 apply (case_tac "a : X")
 apply (case_tac "a : Y")
 apply (auto)
done

lemma par_tr_hide_tr_if:
  "[| u ~= None ; sett(u) <= insert Tick (Ev ` (X Un Y)) |]
   ==> u : u --tr (- X) |[X Int Y]|tr u --tr (- Y)"
by (simp add: par_tr_hide_tr_if_all)

(*** only if ***)

lemma par_tr_hide_tr_only_if_all:
  "ALL s t. (sett s <= insert Tick (Ev ` X) & sett t <= insert Tick (Ev ` Y) & 
             u : s |[X Int Y]|tr t)
            --> (s = u --tr (- X) & t = u --tr (- Y) &
                 sett(u) <= insert Tick (Ev ` (X Un Y)) &
                 u ~= None)"
apply (induct_tac u rule: induct_trace)
apply (simp_all)

apply (intro allI impI)
apply (simp add: par_tr_head)
apply (elim conjE exE)

 (* sync *)
 apply (erule disjE)
 apply (elim conjE exE)
 apply (drule_tac x="s'" in spec)
 apply (drule_tac x="t'" in spec)
 apply (simp)
 apply (case_tac "s' = None", simp)
 apply (case_tac "t' = None", simp)
 apply (simp add: not_None)

 (* left *)
 apply (erule disjE)
 apply (elim conjE exE)
 apply (drule_tac x="s'" in spec)
 apply (drule_tac x="t" in spec)
 apply (simp)
 apply (case_tac "s' = None", simp)
 apply (simp add: not_None)

 apply (case_tac "a : X")
 apply (simp)
 apply (fast)

 (* right *)
 apply (elim conjE exE)
 apply (drule_tac x="sa" in spec)
 apply (drule_tac x="t'" in spec)
 apply (simp)
 apply (case_tac "t' = None", simp)
 apply (simp add: not_None)

 apply (case_tac "a : Y")
 apply (simp)
 apply (fast)
done

(*** par hide ***)

lemma par_tr_hide_tr:
  "[| sett s <= insert Tick (Ev ` X) ; sett t <= insert Tick (Ev ` Y) |]
   ==> u : s |[X Int Y]|tr t 
    = (s = u --tr (- X) & t = u --tr (- Y) &
       sett(u) <= insert Tick (Ev ` (X Un Y)) &
       u ~= None)"
apply (rule iffI)
apply (insert par_tr_hide_tr_only_if_all[of X Y u])
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (simp add: par_tr_hide_tr_if)
done

(*********************************************************
             Alphabetized Parallel eval
 *********************************************************)

(*---*
 | T |
 *---*)

(*** Par and SKIP ***)

lemma Parallel_SKIP_memT:
  "(u :t [[P |[X]| SKIP]]T) = 
   (u :t [[P]]T & sett(u) Int (Ev ` X) = {})"
apply (simp add: Parallel_memT)
apply (simp add: SKIP_memT)
apply (rule iffI)

(* => *)
 apply (elim conjE exE)
 apply (erule disjE)

 (* t = []t *)
  apply (simp add: par_tr_nil)
  apply (elim conjE, simp)

 (* t = [Tick]t *)
  apply (simp add: par_tr_Tick)
  apply (elim conjE, simp)

(* <= *)
 apply (case_tac "Tick ~: sett u")

 (* t = []t *)
  apply (rule_tac x="u" in exI)
  apply (rule_tac x="[]t" in exI)
  apply (simp add: par_tr_nil)
  apply (erule conjE)
  apply (simp add: not_None_T)

 (* t = [Tick]t *)
  apply (rule_tac x="u" in exI)
  apply (rule_tac x="[Tick]t" in exI)
  apply (simp add: par_tr_Tick)
  apply (erule conjE)
  apply (simp add: not_None_T)
done

(*** complement ***)

lemma Parallel_SKIP_comp_memT:
  "(u :t [[P |[- X]| SKIP]]T) = 
   (u :t [[P]]T &  sett u <= insert Tick (Ev ` X))"
apply (simp add: Parallel_SKIP_memT)
apply (auto)
apply (case_tac "x = Tick", simp)
apply (simp add: not_tick_to_Ev)
apply (auto)
done

(*** Alpha_parallel_evalT ***)

lemma Alpha_parallel_memT:
  "(u :t [[P |[X1,X2]| Q]]T) = 
   ((u --tr (- X1)) :t [[P]]T & (u --tr (- X2)) :t [[Q]]T &
    sett(u) <= insert Tick (Ev ` (X1 Un X2)))"

apply (simp add: Alpha_parallel_def)
apply (simp add: Parallel_memT[of _ _ "X1 Int X2"])
apply (simp add: Parallel_SKIP_comp_memT)
apply (rule iffI)

(* => *)
 apply (elim conjE exE)
 apply (simp add: par_tr_hide_tr)

(* <= *)
 apply (rule_tac x="u --tr (- X1)" in exI)
 apply (rule_tac x="u --tr (- X2)" in exI)
 apply (case_tac "u = None", simp)
 apply (simp add: par_tr_hide_tr_if)
 apply (simp add: hide_tr_sett_comp_subseteq)
done

(*** Semantics for alphabetized parallel on T ***)

lemma Alpha_parallel_evalT:
  "[[P |[X1,X2]| Q]]T = 
   {u. ((u --tr (- X1)) :t [[P]]T & (u --tr (- X2)) :t [[Q]]T &
        sett(u) <= insert Tick (Ev ` (X1 Un X2)))}t"
apply (simp add: Alpha_parallel_memT[THEN sym])
apply (simp add: CollectT_open)
done

(*---*
 | F |
 *---*)

(*** Par and SKIP ***)

lemma Parallel_SKIP_memF_lm1: 
  "[| Y - insert Tick (Ev ` X) = Z - insert Tick (Ev ` X) ; 
     (Tick ~: Z | Z <= Evset) |]
   ==> Z - Y <= Ev ` X"
by (auto simp add: Evset_def)

lemma Parallel_SKIP_memF_lm2: 
  "Y - insert Tick (Ev ` X) = Z - insert Tick (Ev ` X)  
   ==> Z - insert Tick Y <= Ev ` X"
by (auto)

(* Para SKIP *)

lemma Parallel_SKIP_memF:
  "(f :f [[P |[X]| SKIP]]F) = 
   (EX u Y Z. f = (u, Y Un Z) & (u, Y) :f [[P]]F & 
              sett(u) Int (Ev ` X) = {} & Z <= Ev ` X)"
apply (simp add: Parallel_memF)
apply (simp add: SKIP_memF)
apply (rule iffI)

(* => *)
 apply (elim conjE exE)
 apply (erule disjE)

 (* t = []t *)
apply (simp)

  apply (simp add: par_tr_nil)
  apply (elim conjE, simp)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z - Y" in exI)
  apply (simp add: Parallel_SKIP_memF_lm1)

 (* t = [Tick]t *)
  apply (simp add: par_tr_Tick)
  apply (elim conjE, simp)
  apply (case_tac "Tick ~: Z")

   apply (rule_tac x="Y" in exI)
   apply (rule_tac x="Z - Y" in exI)
   apply (simp add: Parallel_SKIP_memF_lm1)

   (* Tick : Z *)
   apply (rule_tac x="insert Tick Y" in exI)
   apply (rule_tac x="(Z - insert Tick Y)" in exI)
   apply (rule conjI, force)
   apply (simp add: Tick_in_sett)
   apply (erule exE)
   apply (rule conjI)
   apply (simp)
   apply (rule proc_T2_T3)
   apply (simp)
   apply (simp add: Parallel_SKIP_memF_lm2)

(* <= *)
 apply (elim conjE exE)
 apply (case_tac "Tick ~: sett u")

 (* t = []t *)
  apply (rule_tac x="u" in exI)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z Un (Y - {Tick})" in exI)
  apply (simp)
  apply (rule conjI)
  apply (fast)
  apply (rule conjI)
  apply (fast)
  apply (rule_tac x="u" in exI)
  apply (rule_tac x="[]t" in exI)
  apply (simp add: par_tr_nil)
  apply (case_tac "u = None", simp, simp)
  apply (simp add: Evset_def)
  apply (force)

 (* t = [Tick]t *)
  apply (rule_tac x="u" in exI)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z Un Y" in exI)
  apply (simp)
  apply (rule conjI)
  apply (fast)
  apply (rule conjI)
  apply (fast)
  apply (rule_tac x="u" in exI)
  apply (rule_tac x="[Tick]t" in exI)
  apply (simp add: par_tr_Tick)
  apply (case_tac "u = None", simp, simp)
done

(*** complement ***)

lemma Parallel_SKIP_comp_memF:
  "(f :f [[P |[- X]| SKIP]]F) = 
   (EX u Y Z. f = (u, Y Un Z) & (u, Y) :f [[P]]F & 
              sett(u) <= insert Tick (Ev ` X) & Z Int insert Tick (Ev ` X) = {})"
apply (simp add: Parallel_SKIP_memF)
apply (auto)
apply (rule_tac x="Y" in exI)
apply (rule_tac x="Z" in exI)
apply (auto)
apply (case_tac "x = Tick", simp)
apply (simp add: not_tick_to_Ev)
apply (auto)
apply (rule_tac x="Y" in exI)
apply (rule_tac x="Z" in exI)
apply (auto)
apply (case_tac "x = Tick", simp)
apply (simp add: not_tick_to_Ev)
apply (auto)
done

(*** Alpha_parallel_evalF ***)

lemma Alpha_parallel_memF_lm1:
  "[| Tick ~: Za ; Za Int Ev ` X1 = {}; 
      Tick ~: Zb ; Zb Int Ev ` X2 = {};
      Ya Un Za - insert Tick (Ev ` (X1 Int X2)) =
      Yb Un Zb - insert Tick (Ev ` (X1 Int X2)) |]
       ==>
       (Ya Un Za Un (Yb Un Zb)) Int insert Tick (Ev ` (X1 Un X2)) =
         (Ya Int insert Tick (Ev ` X1)) Un
         (Yb Int insert Tick (Ev ` X2))"
by (auto)

lemma Alpha_parallel_memF_lm2:
  "X Int insert Tick (Ev ` (X1 Un X2)) <= Y Un Z
   ==> X = X Int Ev ` (X1 - X2) Un X Int Y Int insert Tick (Ev ` (X1 Int X2)) Un
               (X - insert Tick (Ev ` X1)) Un
               (X Int Ev ` (X2 - X1) Un X Int Z Int insert Tick (Ev ` (X1 Int X2)) Un
                (X - insert Tick (Ev ` X2)))"
by (auto)

lemma Alpha_parallel_memF_lm3:
  "X Int Ev ` (X1 - X2) Un X Int Y Int insert Tick (Ev ` (X1 Int X2)) Un
           (X - insert Tick (Ev ` X1)) -
           insert Tick (Ev ` (X1 Int X2)) =
           X Int Ev ` (X2 - X1) Un X Int Z Int insert Tick (Ev ` (X1 Int X2)) Un
           (X - insert Tick (Ev ` X2)) -
           insert Tick (Ev ` (X1 Int X2))"
by (auto)

lemma Alpha_parallel_memF_lm4:
 "[| X Int insert Tick (Ev ` (X1 Un X2)) =
     Y Int insert Tick (Ev ` X1) Un Z Int insert Tick (Ev ` X2) |]
  ==> X Int Ev ` (X1 - X2) Un X Int Y Int insert Tick (Ev ` (X1 Int X2)) <= Y"
by (auto)

lemma Alpha_parallel_memF_lm5:
 "[| X Int insert Tick (Ev ` (X1 Un X2)) =
     Y Int insert Tick (Ev ` X1) Un Z Int insert Tick (Ev ` X2) |]
  ==> X Int Ev ` (X2 - X1) Un X Int Z Int insert Tick (Ev ` (X1 Int X2)) <= Z"
by (auto)

(* F *)

lemma Alpha_parallel_memF:
  "(f :f [[P |[X1,X2]| Q]]F) = 
   (EX u X.
     f = (u, X) &
     (EX Y Z.
         X Int insert Tick (Ev ` (X1 Un X2)) =
         Y Int insert Tick (Ev ` X1) Un Z Int insert Tick (Ev ` X2) &
         (u --tr (- X1), Y) :f [[P]]F &
         (u --tr (- X2), Z) :f [[Q]]F & sett u <= insert Tick (Ev ` (X1 Un X2))))"
apply (simp add: Alpha_parallel_def)
apply (simp add: Parallel_memF[of _ _ "X1 Int X2"])
apply (simp add: Parallel_SKIP_comp_memF)
apply (rule iffI)

(* => *)
 apply (elim conjE exE)
 apply (simp add: par_tr_hide_tr)
 apply (rule_tac x="Ya" in exI, simp)
 apply (rule_tac x="Yb" in exI, simp)
 apply (simp add: Alpha_parallel_memF_lm1)

(* <= *)
 apply (elim conjE exE)
 apply (simp)
 apply (rule_tac x=
   "(X Int (Ev ` (X1 - X2))) Un 
    (X Int Y Int insert Tick (Ev ` (X1 Int X2))) Un
    (X - insert Tick (Ev ` X1))" in exI)
 apply (rule_tac x=
   "(X Int (Ev ` (X2 - X1))) Un 
    (X Int Z Int insert Tick (Ev ` (X1 Int X2))) Un
    (X - insert Tick (Ev ` X2))" in exI)

(* X = Y' Un Z' *)
  apply (rule conjI)
  apply (rule Alpha_parallel_memF_lm2)
  apply (fast)

(* Parallel condition *)
  apply (rule conjI)
  apply (simp add: Alpha_parallel_memF_lm3)

(* [[ ]]F *)
  apply (rule_tac x="u --tr (- X1)" in exI)
  apply (rule_tac x="u --tr (- X2)" in exI)
  apply (case_tac "u = None", simp)
  apply (simp add: par_tr_hide_tr_if)
  apply (simp add: hide_tr_sett_comp_subseteq)

  apply (rule conjI)
   apply (rule_tac x=
     "(X Int (Ev ` (X1 - X2))) Un 
      (X Int Y Int insert Tick (Ev ` (X1 Int X2)))" in exI)
   apply (rule_tac x= "(X - insert Tick (Ev ` X1))" in exI)
   apply (simp)
   apply (rule conjI)
   apply (rule proc_F2)
   apply (simp)
   apply (simp add: Alpha_parallel_memF_lm4)
   apply (force)

   apply (rule_tac x=
     "(X Int (Ev ` (X2 - X1))) Un 
      (X Int Z Int insert Tick (Ev ` (X1 Int X2)))" in exI)
   apply (rule_tac x= "(X - insert Tick (Ev ` X2))" in exI)
   apply (simp)
   apply (rule conjI)
   apply (rule proc_F2)
   apply (simp)
   apply (simp add: Alpha_parallel_memF_lm5)
   apply (force)
done

(*** Semantics for alphabetized parallel on F ***)

lemma Alpha_parallel_evalF:
  "[[P |[X1,X2]| Q]]F = 
   {f. (EX u X. f = (u, X) &
     (EX Y Z.
         X Int insert Tick (Ev ` (X1 Un X2)) =
         Y Int insert Tick (Ev ` X1) Un Z Int insert Tick (Ev ` X2) &
         (u --tr (- X1), Y) :f [[P]]F &
         (u --tr (- X2), Z) :f [[Q]]F & sett u <= insert Tick (Ev ` (X1 Un X2))))}f"
apply (simp only: Alpha_parallel_memF[THEN sym])
apply (simp add: CollectF_open)
done

(****************** to add it again ******************)

declare disj_not1   [simp] 
declare not_None_eq [simp]

end




