           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June     2005  (modified)   |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_law_decompo = CSP_proc:

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*------------------------------------------------*
 |                                                |
 |      laws for monotonicity and congruence      |
 |                                                |
 *------------------------------------------------*)

(*********************************************************
                        Act_prefix mono
 *********************************************************)

lemma domSF_Act_prefix_mono:
  "[| a = b ; [[P]]SF <= [[Q]]SF |]
     ==> [[a -> P]]SF <= [[b -> Q]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Act_prefix_evalT_mono)
apply (simp add: evalF_def Act_prefix_evalF_mono)
done

lemma domSF_Act_prefix_cong:
  "[| a = b ; [[P]]SF = [[Q]]SF |]
     ==> [[a -> P]]SF = [[b -> Q]]SF"
apply (rule order_antisym)
apply (simp_all add: domSF_Act_prefix_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Act_prefix_mono:
  "[| a = b ; P <=SF Q |] ==> a -> P <=SF b -> Q"
by (simp add: refSF_def domSF_Act_prefix_mono)

lemma csp_Act_prefix_cong:
  "[| a = b ; P =SF Q |] ==> a -> P =SF b -> Q"
by (simp add: csp_eq_ref_iff csp_Act_prefix_mono)

(*********************************************************
                   Ext_pre_choice mono 
 *********************************************************)

lemma domSF_Ext_pre_choice_mono:
  "[|  X = Y ; !! a. a:Y ==> [[Pf a]]SF <= [[Qf a]]SF |]
     ==> [[? :X -> Pf]]SF <= [[? :Y -> Qf]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Ext_pre_choice_evalT_mono)
apply (simp add: evalF_def Ext_pre_choice_evalF_mono)
done

lemma domSF_Ext_pre_choice_cong:
  "[|  X = Y ; !! a. a:Y ==> [[Pf a]]SF = [[Qf a]]SF |]
     ==> [[? :X -> Pf]]SF = [[? :Y -> Qf]]SF"
apply (rule order_antisym)
apply (simp_all add: domSF_Ext_pre_choice_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_pre_choice_mono:
  "[| X = Y ; !! a. a:Y ==> Pf a <=SF Qf a |] ==> ? :X -> Pf <=SF ? :Y -> Qf"
by (simp add: refSF_def domSF_Ext_pre_choice_mono)

lemma csp_Ext_pre_choice_cong:
  "[| X = Y ; !! a. a:Y ==> Pf a =SF Qf a |] ==> ? :X -> Pf =SF ? :Y -> Qf"
by (simp add: csp_eq_ref_iff csp_Ext_pre_choice_mono)

(*********************************************************
                      Ext choice mono
 *********************************************************)

lemma domSF_Ext_choice_mono:
  "[| [[P1]]SF <= [[Q1]]SF ; [[P2]]SF <= [[Q2]]SF |]
     ==> [[P1 [+] P2]]SF <= [[Q1 [+] Q2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Ext_choice_evalT_mono)
apply (simp add: evalF_def Ext_choice_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_mono:
  "[| P1 <=SF Q1 ; P2 <=SF Q2 |] ==> P1 [+] P2 <=SF Q1 [+] Q2"
by (simp add: refSF_def domSF_Ext_choice_mono)

lemma csp_Ext_choice_cong:
  "[| P1 =SF Q1 ; P2 =SF Q2 |] ==> P1 [+] P2 =SF Q1 [+] Q2"
by (simp add: csp_eq_ref_iff csp_Ext_choice_mono)

(*********************************************************
                      Int choice mono
 *********************************************************)

lemma domSF_Int_choice_mono:
  "[| [[P1]]SF <= [[Q1]]SF ; [[P2]]SF <= [[Q2]]SF |]
     ==> [[P1 |~| P2]]SF <= [[Q1 |~| Q2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Int_choice_evalT_mono)
apply (simp add: evalF_def Int_choice_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Int_choice_mono:
  "[| P1 <=SF Q1 ; P2 <=SF Q2 |] ==> P1 |~| P2 <=SF Q1 |~| Q2"
by (simp add: refSF_def domSF_Int_choice_mono)

lemma csp_Int_choice_cong:
  "[| P1 =SF Q1 ; P2 =SF Q2 |] ==> P1 |~| P2 =SF Q1 |~| Q2"
by (simp add: csp_eq_ref_iff csp_Int_choice_mono)

(*********************************************************
               replicated internal choice
 *********************************************************)

lemma domSF_Rep_int_choice_mono:
  "[|  X = Y ; !! a. a:Y ==> [[Pf a]]SF <= [[Qf a]]SF |]
     ==> [[!! :X .. Pf]]SF <= [[!! :Y .. Qf]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Rep_int_choice_evalT_mono)
apply (simp add: evalF_def Rep_int_choice_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Rep_int_choice_mono0:
  "[| X = Y ; !! a. a:Y ==> Pf a <=SF Qf a |]
      ==> !! :X .. Pf <=SF !! :Y .. Qf"
by (simp add: refSF_def domSF_Rep_int_choice_mono)

lemma csp_Rep_int_choice_cong0:
  "[| X = Y ; !! a. a:Y ==> Pf a =SF Qf a |]
      ==> !! :X .. Pf =SF !! :Y .. Qf"
by (simp add: csp_eq_ref_iff csp_Rep_int_choice_mono0)

(*** OR1 ***)

lemma csp_Rep_int_choice_mono1:
  "[| X = Y ; !! a. a:Y ==> Pf a <=SF Qf a |]
      ==> ! :X .. Pf <=SF ! :Y .. Qf"
apply (simp add: Rep_int_choice_OR1_def)
apply (rule csp_Rep_int_choice_mono0)
by (auto)

lemma csp_Rep_int_choice_cong1:
  "[| X = Y ; !! a. a:Y ==> Pf a =SF Qf a |]
      ==> ! :X .. Pf =SF ! :Y .. Qf"
by (simp add: csp_eq_ref_iff csp_Rep_int_choice_mono1)

(*** OR2 ***)

lemma csp_Rep_int_choice_mono2:
  "[| X = Y ; !! a. a:Y ==> Pf a <=SF Qf a |]
      ==> !nat :X .. Pf <=SF !nat :Y .. Qf"
apply (simp add: Rep_int_choice_OR2_def)
apply (rule csp_Rep_int_choice_mono0)
by (auto)

lemma csp_Rep_int_choice_cong2:
  "[| X = Y ; !! a. a:Y ==> Pf a =SF Qf a |]
      ==> !nat :X .. Pf =SF !nat :Y .. Qf"
by (simp add: csp_eq_ref_iff csp_Rep_int_choice_mono2)

lemmas csp_Rep_int_choice_mono = csp_Rep_int_choice_mono0
                                 csp_Rep_int_choice_mono1
                                 csp_Rep_int_choice_mono2

lemmas csp_Rep_int_choice_cong = csp_Rep_int_choice_cong0
                                 csp_Rep_int_choice_cong1
                                 csp_Rep_int_choice_cong2

(*********************************************************
                   Conditional mono
 *********************************************************)

lemma domSF_Conditional_mono:
  "[| b1 = b2 ; [[P1]]SF <= [[Q1]]SF ; [[P2]]SF <= [[Q2]]SF |]
     ==> [[IF b1 THEN P1 ELSE P2]]SF <= 
         [[IF b2 THEN Q1 ELSE Q2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Conditional_evalT_mono)
apply (simp add: evalF_def Conditional_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Conditional_mono:
  "[| b1 = b2 ; P1 <=SF Q1 ; P2 <=SF Q2 |]
           ==> IF b1 THEN P1 ELSE P2 <=SF
               IF b2 THEN Q1 ELSE Q2"
by (simp add: refSF_def domSF_Conditional_mono)

lemma csp_Conditional_cong:
  "[| b1 = b2 ; P1 =SF Q1 ;  P2 =SF Q2 |]
           ==> IF b1 THEN P1 ELSE P2 =SF
               IF b2 THEN Q1 ELSE Q2"
by (simp add: csp_eq_ref_iff csp_Conditional_mono)

(*********************************************************
                     Parallel mono 
 *********************************************************)

lemma domSF_Parallel_mono:
  "[| X = Y ; [[P1]]SF <= [[Q1]]SF ; [[P2]]SF <= [[Q2]]SF |]
     ==> [[P1 |[X]| P2]]SF <= [[Q1 |[X]| Q2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Parallel_evalT_mono)
apply (simp add: evalF_def Parallel_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Parallel_mono:
  "[| X = Y ; P1 <=SF Q1 ; P2 <=SF Q2 |]
           ==> P1 |[X]| P2 <=SF Q1 |[Y]| Q2"
by (simp add: refSF_def domSF_Parallel_mono)

lemma csp_Parallel_cong:
  "[| X = Y ; P1 =SF Q1 ; P2 =SF Q2 |]
           ==> P1 |[X]| P2 =SF Q1 |[Y]| Q2"
by (simp add: csp_eq_ref_iff csp_Parallel_mono)

(*********************************************************
                        Hiding mono
 *********************************************************)

lemma domSF_Hiding_mono:
  "[| X = Y ; [[P]]SF <= [[Q]]SF |]
     ==> [[P -- X]]SF <= [[Q -- X]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Hiding_evalT_mono)
apply (simp add: evalF_def Hiding_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Hiding_mono:
  "[| X = Y ; P <=SF Q |] ==> P -- X <=SF Q -- Y"
by (simp add: refSF_def domSF_Hiding_mono)

lemma csp_Hiding_cong:
  "[| X = Y ; P =SF Q |] ==> P -- X =SF Q -- Y"
by (simp add: csp_eq_ref_iff csp_Hiding_mono)

(*********************************************************
                        Renaming mono
 *********************************************************)

lemma domSF_Renaming_mono:
  "[| r1 = r2 ; [[P]]SF <= [[Q]]SF |]
     ==> [[P [[r1]]]]SF <= [[Q [[r2]]]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Renaming_evalT_mono)
apply (simp add: evalF_def Renaming_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Renaming_mono:
  "[| r1 = r2 ; P <=SF Q |] ==> P [[r1]] <=SF Q [[r2]]"
by (simp add: refSF_def domSF_Renaming_mono)

lemma csp_Renaming_cong:
  "[| r1 = r2 ; P =SF Q |] ==> P [[r1]] =SF Q [[r2]]"
by (simp add: csp_eq_ref_iff csp_Renaming_mono)

(*********************************************************
               Sequential composition mono 
 *********************************************************)

lemma domSF_Seq_compo_mono:
  "[| [[P1]]SF <= [[Q1]]SF ; [[P2]]SF <= [[Q2]]SF |]
     ==> [[P1 ;; P2]]SF <= [[Q1 ;; Q2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: evalT_def Seq_compo_evalT_mono)
apply (simp add: evalF_def Seq_compo_evalF_mono)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Seq_compo_mono:
  "[| P1 <=SF Q1 ; P2 <=SF Q2 |]
           ==> P1 ;; P2 <=SF Q1 ;; Q2"
by (simp add: refSF_def domSF_Seq_compo_mono)

lemma csp_Seq_compo_cong:
  "[| P1 =SF Q1 ; P2 =SF Q2 |] ==> P1 ;; P2 =SF Q1 ;; Q2"
by (simp add: csp_eq_ref_iff csp_Seq_compo_mono)

(*********************************************************
                        Timeout mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Timeout_mono:
  "[| P1 <=SF Q1 ; P2 <=SF Q2 |]
      ==> P1 [> P2 <=SF Q1 [> Q2"
apply (rule csp_Ext_choice_mono)
apply (rule csp_Int_choice_mono)
by (simp_all)

lemma csp_Timeout_cong:
  "[| P1 =SF Q1 ; P2 =SF Q2 |]
      ==> P1 [> P2 =SF Q1 [> Q2"
by (simp add: csp_eq_ref_iff csp_Timeout_mono)

(*------------------------------------------------------*
 |                       alias                          |
 *------------------------------------------------------*)

lemmas csp_free_mono =
   csp_Ext_choice_mono csp_Int_choice_mono csp_Parallel_mono
   csp_Hiding_mono csp_Renaming_mono csp_Seq_compo_mono

lemmas csp_mono = csp_free_mono
   csp_Act_prefix_mono csp_Ext_pre_choice_mono 
   csp_Rep_int_choice_mono csp_Conditional_mono 

lemmas csp_free_cong =
   csp_Ext_choice_cong csp_Int_choice_cong csp_Parallel_cong
   csp_Hiding_cong csp_Renaming_cong csp_Seq_compo_cong

lemmas csp_cong = csp_free_cong
   csp_Act_prefix_cong csp_Ext_pre_choice_cong 
   csp_Rep_int_choice_cong csp_Conditional_cong 

lemmas csp_free_decompo = csp_free_mono csp_free_cong
lemmas csp_decompo = csp_mono csp_cong

lemmas csp_rm_head_mono = csp_Act_prefix_mono csp_Ext_pre_choice_mono 
lemmas csp_rm_head_cong = csp_Act_prefix_cong csp_Ext_pre_choice_cong
lemmas csp_rm_head      = csp_rm_head_mono csp_rm_head_cong

(****************** to add them again ******************)

declare not_None_eq    [simp]

end
