           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June     2005  (modified)   |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_law_dist = CSP_law_commut + CSP_law_decompo:

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

      distribution of internal choice over ...

         1. (P1 |~| P2) [+] Q
         2. Q [+] (P1 |~| P2)
         3. (P1 |~| P2) |[X]| Q
         4. Q |[X]| (P1 |~| P2)
         5. (P1 |~| P2) -- X
         6. (P1 |~| P2) [[r]]

 *****************************************************************)

(*********************************************************
                dist law for Ext_choice (l)
 *********************************************************)

lemma domSF_Ext_choice_dist_l: 
  "[[(P1 |~| P2) [+] Q]]SF = [[(P1 [+] Q) |~| (P2 [+] Q)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
 apply (rule, simp add: Ext_choice_mem Int_choice_mem, fast)+

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Ext_choice_mem Int_choice_mem
                        memF_IntF memF_UnF memT_UnT, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_dist_l: 
  "(P1 |~| P2) [+] Q =SF
   (P1 [+] Q) |~| (P2 [+] Q)"
by (simp add: eqSF_def domSF_Ext_choice_dist_l)

(*********************************************************
                dist law for Ext_choice (r)
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_dist_r: 
  "P [+] (Q1 |~| Q2) =SF
   (P [+] Q1) |~| (P [+] Q2)"
apply (rule csp_rw_left)
apply (rule csp_commut)
apply (rule csp_rw_left)
apply (rule csp_Ext_choice_dist_l)
apply (rule csp_decompo)
apply (rule csp_commut)+
done

(*********************************************************
                dist law for Parallel (l)
 *********************************************************)

lemma domSF_Parallel_dist_l: 
  "[[(P1 |~| P2) |[X]| Q]]SF = [[(P1 |[X]| Q) |~| (P2 |[X]| Q)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
 apply (rule, simp add: Parallel_mem Int_choice_mem, fast)+

(* domF *)
 apply (rule eq_iffI)
  apply (rule, simp add: Parallel_mem Int_choice_mem)
  apply (elim disjE conjE exE)
  apply (rule disjI1)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)
  apply (rule disjI2)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)

  apply (rule, simp add: Parallel_mem Int_choice_mem)
  apply (elim disjE conjE exE)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Parallel_dist_l: 
  "(P1 |~| P2) |[X]| Q =SF
   (P1 |[X]| Q) |~| (P2 |[X]| Q)"
by (simp add: eqSF_def domSF_Parallel_dist_l)

(*********************************************************
                dist law for Parallel (r)
 *********************************************************)

lemma csp_Parallel_dist_r: 
  "P |[X]| (Q1 |~| Q2) =SF
   (P |[X]| Q1) |~| (P |[X]| Q2)"
apply (rule csp_rw_left)
apply (rule csp_commut)
apply (rule csp_rw_left)
apply (rule csp_Parallel_dist_l)
apply (rule csp_decompo)
apply (rule csp_commut)+
done

(*********************************************************
                dist law for Hiding
 *********************************************************)

lemma domSF_Hiding_dist: 
  "[[(P1 |~| P2) -- X]]SF =
   [[(P1 -- X) |~| (P2 -- X)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
 apply (rule, simp add: Hiding_mem Int_choice_mem, fast)+

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Hiding_mem Int_choice_mem, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Hiding_dist: 
  "(P1 |~| P2) -- X =SF
   (P1 -- X) |~| (P2 -- X)"
by (simp add: eqSF_def domSF_Hiding_dist)

(*********************************************************
               dist law for Renaming
 *********************************************************)

lemma domSF_Renaming_dist: 
  "[[(P1 |~| P2) [[r]]]]SF =
   [[(P1 [[r]]) |~| (P2 [[r]])]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
 apply (rule, simp add: Renaming_mem Int_choice_mem, fast)+

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Renaming_mem Int_choice_mem, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Renaming_dist: 
  "(P1 |~| P2) [[r]] =SF
   (P1 [[r]]) |~| (P2 [[r]])"
by (simp add: eqSF_def domSF_Renaming_dist)

(*********************************************************
         dist law for Sequential composition
 *********************************************************)

lemma domSF_Seq_compo_dist: 
  "[[(P1 |~| P2) ;; Q]]SF =
   [[(P1 ;; Q) |~| (P2 ;; Q)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
 apply (rule, simp add: Seq_compo_mem Int_choice_mem, fast)+

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Seq_compo_mem Int_choice_mem, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Seq_compo_dist: 
  "(P1 |~| P2) ;; Q =SF
   (P1 ;; Q) |~| (P2 ;; Q)"
by (simp add: eqSF_def domSF_Seq_compo_dist)

(*********************************************************
                     dist laws
 *********************************************************)

lemmas csp_dist = csp_Ext_choice_dist_l csp_Ext_choice_dist_r
                  csp_Parallel_dist_l   csp_Parallel_dist_r
                  csp_Hiding_dist       csp_Renaming_dist
                  csp_Seq_compo_dist

(*****************************************************************

      distribution of internal bind over ...

         1. (!! :X .. Pf) [+] Q
         2. Q [+] (!! :X .. Pf)
         3. (!! :X .. Pf) |[X]| Q
         4. Q |[X]| (!! :X .. Pf)
         5. (!! :X .. Pf) -- X
         6. (!! :X .. Pf) [[r]]

 *****************************************************************)

(*********************************************************
                Rep_dist law for Ext_choice (l)
 *********************************************************)

lemma domSF_Ext_choice_Rep_dist0_l: 
  "X ~= {} ==> [[(!! :X .. Pf) [+] Q]]SF = [[!! x:X .. (Pf x [+] Q)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
 apply (rule, simp add: Ext_choice_mem Rep_int_choice_mem
                        memF_IntF memF_UnF memT_UnT, fast)+

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Ext_choice_mem Rep_int_choice_mem
                        memF_IntF memF_UnF memT_UnT, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_Rep_dist0_l: 
  "X ~= {} ==> (!! :X .. Pf) [+] Q =SF
               !! x:X .. (Pf x [+] Q)"
by (simp add: eqSF_def domSF_Ext_choice_Rep_dist0_l)

(*********************************************************
                Rep_dist0 law for Ext_choice (r)
 *********************************************************)

lemma csp_Ext_choice_Rep_dist0_r: 
  "X ~= {} ==> P [+] (!! :X .. Qf) =SF
               !! x:X .. (P [+] Qf x)"
apply (rule csp_rw_left)
apply (rule csp_commut)
apply (rule csp_rw_left)
apply (rule csp_Ext_choice_Rep_dist0_l, simp)
apply (rule csp_decompo, simp)
apply (rule csp_commut)
done

(*********************************************************
                Rep_dist0 law for Parallel (l)
 *********************************************************)

lemma domSF_Parallel_Rep_dist0_l: 
  "Y ~= {} ==> [[(!! :Y .. Pf) |[X]| Q]]SF = [[!! x:Y .. (Pf x |[X]| Q)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
  apply (rule, simp add: Parallel_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE bexE)
   apply (subgoal_tac "EX a. a:Y")
   apply (elim exE)
   apply (rule disjI2)
   apply (rule_tac x="a" in bexI)
   apply (simp)
   apply (rule_tac x="[]t" in exI)
   apply (rule_tac x="ta" in exI)
   apply (simp)
   apply (simp)
   apply (fast)
  (* *)
   apply (rule disjI2)
   apply (rule_tac x="a" in bexI)
   apply (fast)
   apply (simp)

  apply (rule, simp add: Parallel_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE bexE)
   apply (rule_tac x="[]t" in exI)
   apply (rule_tac x="[]t" in exI)
   apply (simp)
   apply (fast)

(* domF *)
 apply (rule eq_iffI)
  apply (rule, simp add: Parallel_mem Rep_int_choice_mem)
  apply (elim conjE exE bexE)
   apply (rule_tac x="a" in bexI)
   apply (rule_tac x="Ya" in exI)
   apply (rule_tac x="Z" in exI)
   apply (fast)
   apply (simp)
  (* *)
  apply (rule, simp add: Parallel_mem Rep_int_choice_mem)
  apply (elim conjE exE bexE)
   apply (rule_tac x="Ya" in exI)
   apply (rule_tac x="Z" in exI)
   apply (fast)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Parallel_Rep_dist0_l: 
  "Y ~= {} ==>
     (!! :Y .. Pf) |[X]| Q =SF
     !! x:Y .. (Pf x |[X]| Q)"
by (simp add: eqSF_def domSF_Parallel_Rep_dist0_l)

(*********************************************************
                Rep_dist0 law for Parallel (r)
 *********************************************************)

lemma csp_Parallel_Rep_dist0_r: 
  "Y ~= {} ==>
     P |[X]| (!! :Y .. Qf) =SF
     !! x:Y .. (P |[X]| Qf x)"
apply (rule csp_rw_left)
apply (rule csp_commut)
apply (rule csp_rw_left)
apply (rule csp_Parallel_Rep_dist0_l, simp)
apply (rule csp_decompo, simp)
apply (rule csp_commut)
done

(*********************************************************
                Rep_dist0 law for Hiding
 *********************************************************)

lemma domSF_Hiding_Rep_dist0: 
  "[[(!! :Y .. Pf) -- X]]SF = [[!! x:Y .. (Pf x -- X)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
  apply (rule, simp add: Hiding_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE, simp, fast)

  apply (rule, simp add: Hiding_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE bexE)
  apply (rule_tac x="[]t" in exI, simp)
  apply (rule_tac x="s" in exI, fast)

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Hiding_mem Rep_int_choice_mem, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Hiding_Rep_dist0: 
  "(!! :Y .. Pf) -- X =SF
   !! x:Y .. (Pf x -- X)"
by (simp add: eqSF_def domSF_Hiding_Rep_dist0)

(*********************************************************
                Rep_dist0 law for Renaming
 *********************************************************)

lemma domSF_Renaming_Rep_dist0: 
  "[[(!! :X .. Pf) [[r]]]]SF = [[!! x:X .. (Pf x [[r]])]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
  apply (rule, simp add: Renaming_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE, simp, fast)
  apply (rule, simp add: Renaming_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE, simp, fast)

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Renaming_mem Rep_int_choice_mem, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Renaming_Rep_dist0: 
  "(!! :X .. Pf) [[r]] =SF
   !! x:X .. (Pf x [[r]])"
by (simp add: eqSF_def domSF_Renaming_Rep_dist0)

(*********************************************************
          Rep_dist0 law for Sequential composition
 *********************************************************)

lemma domSF_Seq_compo_Rep_dist0: 
  "[[(!! :X .. Pf) ;; Q]]SF = [[!! x:X .. (Pf x ;; Q)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* domT *)
 apply (rule eq_iffI)
  apply (rule, simp add: Seq_compo_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE, simp, fast, fast)
  apply (rule, simp add: Seq_compo_mem Rep_int_choice_mem)
  apply (elim disjE conjE exE bexE)
  apply (rule disjI1)
  apply (rule_tac x="[]t" in exI, simp, fast, fast)

(* domF *)
 apply (rule eq_iffI)
 apply (rule, simp add: Seq_compo_mem Rep_int_choice_mem, fast)+
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Seq_compo_Rep_dist0: 
  "(!! :X .. Pf) ;; Q =SF
   !! x:X .. (Pf x ;; Q)"
by (simp add: eqSF_def domSF_Seq_compo_Rep_dist0)

(*********************************************************
                     Rep_dist0 laws
 *********************************************************)

lemmas csp_Rep_dist0 = csp_Ext_choice_Rep_dist0_l csp_Ext_choice_Rep_dist0_r
                       csp_Parallel_Rep_dist0_l   csp_Parallel_Rep_dist0_r
                       csp_Hiding_Rep_dist0       csp_Renaming_Rep_dist0
                       csp_Seq_compo_Rep_dist0

(*****************************************************************

      for convenience

         1. (! :X .. Pf) [+] Q
         2. Q [+] (! :X .. Pf)
         3. (! :X .. Pf) |[X]| Q
         4. Q |[X]| (! :X .. Pf)
         5. (! :X .. Pf) -- X
         6. (! :X .. Pf) [[r]]

 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_Rep_dist1_l: 
  "X ~= {} ==> (! :X .. Pf) [+] Q =SF
               ! x:X .. (Pf x [+] Q)"
by (simp add: Rep_int_choice_OR1_def csp_Rep_dist0)

lemma csp_Ext_choice_Rep_dist1_r: 
  "X ~= {} ==> P [+] (! :X .. Qf) =SF
               ! x:X .. (P [+] Qf x)"
by (simp add: Rep_int_choice_OR1_def csp_Rep_dist0)

lemma csp_Parallel_Rep_dist1_l: 
  "Y ~= {} ==>
     (! :Y .. Pf) |[X]| Q =SF
     ! x:Y .. (Pf x |[X]| Q)"
by (simp add: Rep_int_choice_OR1_def csp_Rep_dist0)

lemma csp_Parallel_Rep_dist1_r: 
  "Y ~= {} ==>
     P |[X]| (! :Y .. Qf) =SF
     ! x:Y .. (P |[X]| Qf x)"
by (simp add: Rep_int_choice_OR1_def csp_Rep_dist0)

lemma csp_Hiding_Rep_dist1: 
  "(! :Y .. Pf) -- X =SF
   ! x:Y .. (Pf x -- X)"
by (simp add: Rep_int_choice_OR1_def csp_Rep_dist0)

lemma csp_Renaming_Rep_dist1: 
  "(! :X .. Pf) [[r]] =SF
   ! x:X .. (Pf x [[r]])"
by (simp add: Rep_int_choice_OR1_def csp_Rep_dist0)

lemma csp_Seq_compo_Rep_dist1: 
  "(! :X .. Pf) ;; Q =SF
   ! x:X .. (Pf x ;; Q)"
by (simp add: Rep_int_choice_OR1_def csp_Rep_dist0)

(*********************************************************
                     Rep_dist laws
 *********************************************************)

lemmas csp_Rep_dist1 = csp_Ext_choice_Rep_dist1_l csp_Ext_choice_Rep_dist1_r
                       csp_Parallel_Rep_dist1_l   csp_Parallel_Rep_dist1_r
                       csp_Hiding_Rep_dist1       csp_Renaming_Rep_dist1
                       csp_Seq_compo_Rep_dist1

(*****************************************************************

      for convenience

         1. (!nat :X .. Pf) [+] Q
         2. Q [+] (!nat :X .. Pf)
         3. (!nat :X .. Pf) |[X]| Q
         4. Q |[X]| (!nat :X .. Pf)
         5. (!nat :X .. Pf) -- X
         6. (!nat :X .. Pf) [[r]]

 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_Rep_dist2_l: 
  "X ~= {} ==> (!nat :X .. Pf) [+] Q =SF
               !nat x:X .. (Pf x [+] Q)"
by (simp add: Rep_int_choice_OR2_def csp_Rep_dist0)

lemma csp_Ext_choice_Rep_dist2_r: 
  "X ~= {} ==> P [+] (!nat :X .. Qf) =SF
               !nat x:X .. (P [+] Qf x)"
by (simp add: Rep_int_choice_OR2_def csp_Rep_dist0)

lemma csp_Parallel_Rep_dist2_l: 
  "Y ~= {} ==>
     (!nat :Y .. Pf) |[X]| Q =SF
     !nat x:Y .. (Pf x |[X]| Q)"
by (simp add: Rep_int_choice_OR2_def csp_Rep_dist0)

lemma csp_Parallel_Rep_dist2_r: 
  "Y ~= {} ==>
     P |[X]| (!nat :Y .. Qf) =SF
     !nat x:Y .. (P |[X]| Qf x)"
by (simp add: Rep_int_choice_OR2_def csp_Rep_dist0)

lemma csp_Hiding_Rep_dist2: 
  "(!nat :Y .. Pf) -- X =SF
   !nat x:Y .. (Pf x -- X)"
by (simp add: Rep_int_choice_OR2_def csp_Rep_dist0)

lemma csp_Renaming_Rep_dist2: 
  "(!nat :X .. Pf) [[r]] =SF
   !nat x:X .. (Pf x [[r]])"
by (simp add: Rep_int_choice_OR2_def csp_Rep_dist0)

lemma csp_Seq_compo_Rep_dist2: 
  "(!nat :X .. Pf) ;; Q =SF
   !nat x:X .. (Pf x ;; Q)"
by (simp add: Rep_int_choice_OR2_def csp_Rep_dist0)

(*********************************************************
                     Rep_dist laws
 *********************************************************)

lemmas csp_Rep_dist2 = csp_Ext_choice_Rep_dist2_l csp_Ext_choice_Rep_dist2_r
                       csp_Parallel_Rep_dist2_l   csp_Parallel_Rep_dist2_r
                       csp_Hiding_Rep_dist2       csp_Renaming_Rep_dist2
                       csp_Seq_compo_Rep_dist2

(*** all rules ***)

lemmas csp_Rep_dist = csp_Rep_dist0 csp_Rep_dist1 csp_Rep_dist2

(*****************************************************************

      additional distribution of internal bind over ...

         1. (! :X .. (a -> P))
         2. (! :Y .. (? :X -> P))

 *****************************************************************)

(*********************************************************
              Rep_dist law for Act_prefix
 *********************************************************)

lemma domT_Act_prefix_Rep_dist0:
  "X ~= {} ==> 
   [[a -> (!! :X .. Pf)]]T = [[!! x:X .. (a -> Pf x)]]T"
apply (rule eq_iffI)
(* <= *)
 apply (rule)
 apply (simp add: Act_prefix_mem)
 apply (erule disjE)
  apply (simp)
  apply (elim conjE exE)
  apply (simp add: Rep_int_choice_mem)
  apply (erule disjE)
   apply (simp add: Act_prefix_mem)
   apply (force)
   apply (elim conjE bexE)
   apply (rule_tac x="aa" in bexI)
   apply (simp add: Act_prefix_mem)
   apply (force)
   apply (simp)

(* => *)
 apply (rule)
 apply (simp add: Act_prefix_mem)
 apply (simp add: Rep_int_choice_mem)
 apply (erule disjE)
  apply (simp)
  apply (elim conjE bexE)
  apply (simp add: Act_prefix_mem)
  apply (erule disjE)
   apply (simp)
   apply (elim conjE exE)
   apply (rule disjI2)
   apply (rule_tac x="s" in exI)
   apply (simp)
   apply (rule disjI2)
   apply (rule_tac x="aa" in bexI)
   apply (simp)
   apply (simp)
done

lemma domF_Act_prefix_Rep_dist0:
  "X ~= {} ==> 
   [[a -> (!! :X .. Pf)]]F = [[!! x:X .. (a -> Pf x)]]F"
apply (rule eq_iffI)

(* <= *)
 apply (rule)
 apply (simp add: Act_prefix_mem)
 apply (erule disjE)
  apply (elim conjE)
  apply (simp add: Rep_int_choice_mem)
  apply (simp add: Act_prefix_mem)
  apply (force)

  apply (simp add: Rep_int_choice_mem)
  apply (elim conjE exE bexE)
  apply (rule_tac x="aa" in bexI)
  apply (simp add: Act_prefix_mem)
  apply (force)
  apply (simp)

(* => *)
 apply (rule)
 apply (simp add: Act_prefix_mem)
 apply (simp add: Rep_int_choice_mem)
 apply (simp add: Act_prefix_mem)
 apply (elim disjE conjE bexE exE)
  apply (simp)
  apply (simp)
  apply (rule_tac x="sa" in exI)
  apply (simp)
  apply (rule_tac x="aa" in bexI)
  apply (simp)
  apply (simp)
done

lemma domSF_Act_prefix_Rep_dist0:
  "X ~= {} ==> 
   [[a -> (!! :X .. Pf)]]SF = [[!! x:X .. (a -> Pf x)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Act_prefix_Rep_dist0)
apply (simp add: domF_Act_prefix_Rep_dist0)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Act_prefix_Rep_dist0:
  "X ~= {} ==> 
   a -> (!! :X .. Pf) =SF
   !! x:X .. (a -> Pf x)"
by (simp add: eqSF_def domSF_Act_prefix_Rep_dist0)

(*********************************************************
              Rep_dist0 law for Ext_pre_choice
 *********************************************************)

lemma domT_Ext_pre_choice_Rep_dist0:
  "Y ~= {} ==> 
   [[? x:X -> (!! y:Y .. (Pf y) x)]]T = [[!! y:Y .. (? :X -> (Pf y))]]T"

apply (rule eq_iffI)
(* <= *)
 apply (rule)
 apply (simp add: Ext_pre_choice_mem)
 apply (erule disjE)
  apply (simp)
  apply (elim conjE exE)
  apply (simp add: Rep_int_choice_mem)
  apply (erule disjE)
   apply (simp add: Ext_pre_choice_mem)
   apply (force)
   apply (elim conjE bexE)
   apply (rule_tac x="aa" in bexI)
   apply (simp add: Ext_pre_choice_mem)
   apply (force)
   apply (simp)

(* => *)
 apply (rule)
 apply (simp add: Ext_pre_choice_mem)
 apply (simp add: Rep_int_choice_mem)
 apply (erule disjE)
  apply (simp)
  apply (elim conjE bexE)
  apply (simp add: Ext_pre_choice_mem)
  apply (erule disjE)
   apply (simp)
   apply (elim conjE exE)
   apply (rule disjI2)
   apply (rule_tac x="aa" in exI)
   apply (rule_tac x="s" in exI)
   apply (simp)
   apply (rule disjI2)
   apply (rule_tac x="a" in bexI)
   apply (simp)
   apply (simp)
done

lemma domF_Ext_pre_choice_Rep_dist0:
  "Y ~= {} ==> 
   [[? x:X -> (!! y:Y .. (Pf y) x)]]F = [[!! y:Y .. (? :X -> (Pf y))]]F"
apply (rule eq_iffI)

(* <= *)
 apply (rule)
 apply (simp add: Ext_pre_choice_mem)
 apply (erule disjE)
  apply (elim conjE)
  apply (simp add: Rep_int_choice_mem)
  apply (simp add: Ext_pre_choice_mem)
  apply (force)

  apply (simp add: Rep_int_choice_mem)
  apply (elim conjE exE bexE)
  apply (rule_tac x="aa" in bexI)
  apply (simp add: Ext_pre_choice_mem)
  apply (force)
  apply (simp)

(* => *)
 apply (rule)
 apply (simp add: Ext_pre_choice_mem)
 apply (simp add: Rep_int_choice_mem)
 apply (simp add: Ext_pre_choice_mem)
 apply (elim disjE conjE exE bexE)
  apply (simp)
  apply (simp)
  apply (rule_tac x="aa" in exI)
  apply (rule_tac x="sa" in exI)
  apply (simp)
  apply (rule_tac x="a" in bexI)
  apply (simp)
  apply (simp)
done

lemma domSF_Ext_pre_choice_Rep_dist0:
  "Y ~= {} ==> 
   [[? x:X -> (!! y:Y .. (Pf y) x)]]SF = [[!! y:Y .. (? :X -> (Pf y))]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Ext_pre_choice_Rep_dist0)
apply (simp add: domF_Ext_pre_choice_Rep_dist0)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_pre_choice_Rep_dist0:
  "Y ~= {} ==> 
   ? x:X -> (!! y:Y .. (Pf y) x) =SF
   !! y:Y .. (? :X -> (Pf y))"
by (simp add: eqSF_def domSF_Ext_pre_choice_Rep_dist0)

(*****************************************************************

      for convenience

         1. (! :X .. (a -> P))
         2. (! :Y .. (? :X -> P))

 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Act_prefix_Rep_dist1:
  "X ~= {} ==> 
   a -> (! :X .. Pf) =SF
   ! x:X .. (a -> Pf x)"
by (simp add: Rep_int_choice_OR1_def csp_Act_prefix_Rep_dist0)

lemma csp_Ext_pre_choice_Rep_dist1:
  "Y ~= {} ==> 
   ? x:X -> (! y:Y .. (Pf y) x) =SF
   ! y:Y .. (? :X -> (Pf y))"
by (simp add: Rep_int_choice_OR1_def csp_Ext_pre_choice_Rep_dist0)

lemma csp_Act_prefix_Rep_dist2:
  "X ~= {} ==> 
   a -> (!nat :X .. Pf) =SF
   !nat x:X .. (a -> Pf x)"
by (simp add: Rep_int_choice_OR2_def csp_Act_prefix_Rep_dist0)

lemma csp_Ext_pre_choice_Rep_dist2:
  "Y ~= {} ==> 
   ? x:X -> (!nat y:Y .. (Pf y) x) =SF
   !nat y:Y .. (? :X -> (Pf y))"
by (simp add: Rep_int_choice_OR2_def csp_Ext_pre_choice_Rep_dist0)

(*** arias ***)

lemmas csp_Act_prefix_Rep_dist 
     = csp_Act_prefix_Rep_dist0
       csp_Act_prefix_Rep_dist1
       csp_Act_prefix_Rep_dist2

lemmas csp_Ext_pre_choice_Rep_dist
     = csp_Ext_pre_choice_Rep_dist0
       csp_Ext_pre_choice_Rep_dist1
       csp_Ext_pre_choice_Rep_dist2

(****************** to add them again ******************)

declare not_None_eq    [simp]

end
