           (*-------------------------------------------*
            |                CSP-Prover                 |
            |              December 2004                |
            |              June     2005 (modified)     |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_law_etc = CSP_law_SKIP   + CSP_law_ref    +
                     CSP_law_dist   + CSP_law_par2   +
                     CSP_law_step1  + CSP_law_step2  +
                     CSP_law_fp:

(*****************************************************************

         1. Conditional
         2. Internal ... simp
         3. External ... simp
         4. Mix      ... simp
         5. Extended reflex

 *****************************************************************)

(*********************************************************
                       IF bool
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Conditional_split: 
  "IF b THEN P ELSE Q =SF (if b then P else Q)"
apply (simp add: eqSF_def)
apply (simp add: proc_eqSF_decompo)
apply (simp add: evalT_def evalt_def)
apply (simp add: evalF_def evalf_def)
done

lemma csp_Conditional_True:
  "IF True THEN P ELSE Q =SF P"
apply (rule csp_rw_left)
apply (rule csp_Conditional_split)
by (simp)

lemma csp_Conditional_False:
  "IF False THEN P ELSE Q =SF Q"
apply (rule csp_rw_left)
apply (rule csp_Conditional_split)
by (simp)

(*****************************************************************
                       Internal ... simp                         
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_DIV_unwind:
   "DIV =SF ! x:{} .. DIV"
apply (simp add: eqSF_def)
apply (simp add: proc_eqSF_decompo)
apply (simp add: Rep_int_choice_OR1_def)
apply (simp add: evalT_def evalF_def evalt_def evalf_def)
apply (simp add: empF_def)
done

(*** ! :{a} ***)

lemma domSF_Rep_int_choice_unique:
  "[[!! :{a} .. Pf]]SF = [[Pf a]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule)
 apply (rule eq_iffI)
  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (force)
  apply (rule)
  apply (simp add: Rep_int_choice_mem)

 apply (rule eq_iffI)
  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (rule)
  apply (simp add: Rep_int_choice_mem)
done

lemma csp_Rep_int_choice0_unique:
  "!! :{a} .. Pf =SF Pf a"
by (simp_all add: eqSF_def domSF_Rep_int_choice_unique)

lemma csp_Rep_int_choice1_unique:
  "! :{a} .. Pf =SF Pf a"
apply (simp add: Rep_int_choice_OR1_def)
by (simp_all add: eqSF_def domSF_Rep_int_choice_unique)

lemma csp_Rep_int_choice2_unique:
  "!nat :{a} .. Pf =SF Pf a"
apply (simp add: Rep_int_choice_OR2_def)
by (simp_all add: eqSF_def domSF_Rep_int_choice_unique)

lemmas csp_Rep_int_choice_unique = csp_Rep_int_choice0_unique
                                   csp_Rep_int_choice1_unique
                                   csp_Rep_int_choice2_unique

(* Int choice idempotence *)

lemma csp_Int_choice_idem: 
  "P |~| P =SF P"
apply (simp add: eqSF_def)
apply (simp add: proc_eqSF_decompo)
apply (rule)
apply (rule eq_iffI,
      (rule, simp add: Int_choice_mem)+)+
done

lemma csp_Int_choice_idem_weak: 
  "P =SF Q ==> P |~| Q =SF P"
apply (rule csp_rw_left)
apply (rule csp_decompo)
apply (rule csp_reflex)
apply (erule csp_symE)
apply (simp)
apply (simp add: csp_Int_choice_idem)
done

(*****************************************************************
                      External ... simp
 *****************************************************************)

(*** STOP [+] P ***)

lemma domSF_Ext_choice_unit_l: 
  "[[STOP [+] P]]SF = [[P]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)

(* T *)
 apply (rule eq_iffI)

 (* <= *)
  apply (rule)
  apply (simp add: Ext_choice_mem)
  apply (erule disjE)
  apply (simp add: STOP_mem)
  apply (simp)
 (* => *)
  apply (rule)
  apply (simp add: Ext_choice_mem)

(* F *)
 apply (rule eq_iffI)

 (* <= *)
  apply (rule)
  apply (simp add: Ext_choice_mem)
  apply (elim disjE)
   apply (simp add: memF_IntF)
   apply (simp add: STOP_mem, fast)
   apply (simp add: memF_UnF)
   apply (elim conjE disjE)
   apply (simp_all add: STOP_mem)
   apply (simp add: memT_UnT)
   apply (elim conjE disjE)
   apply (simp add: STOP_mem)
   apply (rule domSF_F2_F4)
   apply (simp_all)
 (* => *)
  apply (rule)
  apply (simp add: Ext_choice_mem)
  apply (case_tac "s = []t")
   apply (simp add: memF_IntF)
   apply (simp add: STOP_mem)
  (* s ~= []t *)
   apply (simp add: memF_UnF)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_unit_l: 
  "STOP [+] P =SF P"
by (simp add: eqSF_def domSF_Ext_choice_unit_l)

lemma csp_Ext_choice_STOP_r: 
  "P [+] STOP =SF P"
apply (rule csp_rw_left)
apply (rule csp_Ext_choice_commut)
apply (simp add: csp_Ext_choice_unit_l)
done

(*** weak ***)

lemma csp_Ext_choice_unit_l_weak: 
  "Q =SF STOP 
   ==> Q [+] P =SF P"
apply (rule csp_rw_left)
apply (rule csp_decompo)
apply (assumption)
apply (rule csp_reflex)
apply (simp add: csp_Ext_choice_unit_l)
done

lemma csp_Ext_choice_unit_r_weak: 
  "Q =SF STOP 
   ==> P [+] Q =SF P"
apply (rule csp_rw_left)
apply (rule csp_commut)
apply (simp add: csp_Ext_choice_unit_l_weak)
done

(*****************************************************************
                      Mix ... simp
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

(*** <= Timeout ***)

lemma csp_Timeout_right:
  "[| P <=SF Q1 ; P <=SF Q2 |] ==> P <=SF Q1 [> Q2"
apply (rule csp_rw_right)
apply (rule csp_dist)
apply (rule csp_Int_choice_right)
apply (rule csp_Ext_choice_right, simp_all)
apply (rule csp_rw_right)
apply (rule csp_Ext_choice_unit_l, simp_all)
done

(*** STOP [> P  =  P ***)

lemma csp_STOP_Timeout:
  "STOP [> P =SF P"
apply (rule csp_rw_left)
apply (rule csp_dist)
apply (rule csp_rw_left)
apply (rule csp_Int_choice_idem)
apply (rule csp_Ext_choice_unit_l_weak)
by (simp)

lemma csp_STOP_Timeout_weak:
  "[| P1 =SF STOP ;
      P2 =SF STOP |]
   ==> (P1 |~| P2) [+] Q =SF Q"
apply (rule csp_rw_left)
apply (rule csp_decompo)
apply (rule csp_decompo)
apply (simp_all)
apply (rule csp_reflex)
apply (simp add: csp_STOP_Timeout)
done

(*****************************************************************
                Extended reflex by empty sets
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_pre_choice_empty: 
  "? :{} -> Pf =SF ? x:{} -> DIV"
apply (simp add: eqSF_def proc_eqSF_decompo)
apply (simp add: evalt_def evalf_def evalT_def evalF_def)
done

lemma csp_Ext_pre_choice_STOP:
   "? :{} -> Pf =SF STOP"
apply (rule csp_rw_right)
apply (rule csp_STOP_step)
apply (rule csp_Ext_pre_choice_empty)
done

lemmas csp_Ext_pre_choice_STOP_sym = csp_Ext_pre_choice_STOP[THEN csp_sym]

lemma csp_Rep_int_choice0_DIV:
   "!! :{} .. Pf =SF DIV"
apply (simp add: eqSF_def proc_eqSF_decompo)
apply (simp add: evalt_def evalf_def evalT_def evalF_def)
apply (simp add: empF_def)
done

lemma csp_Rep_int_choice1_DIV:
   "! :{} .. Pf =SF DIV"
apply (simp add: Rep_int_choice_OR1_def)
apply (simp add: csp_Rep_int_choice0_DIV)
done

lemma csp_Rep_int_choice2_DIV:
   "!nat :{} .. Pf =SF DIV"
apply (simp add: Rep_int_choice_OR2_def)
apply (simp add: csp_Rep_int_choice0_DIV)
done

lemmas csp_Rep_int_choice_DIV = csp_Rep_int_choice0_DIV
                                csp_Rep_int_choice1_DIV
                                csp_Rep_int_choice2_DIV

lemmas csp_Rep_int_choice_DIV_sym = csp_Rep_int_choice0_DIV[THEN csp_sym]

lemma csp_Ext_pre_choice_empty_reflex:
   "? :{} -> Pf =SF ? :{} -> Qf"
apply (rule csp_trans[of _ "STOP"])
apply (simp add: csp_Ext_pre_choice_STOP)
apply (simp add: csp_Ext_pre_choice_STOP_sym)
done

lemma csp_Rep_int_choice0_empty_reflex:
   "!! :{} .. Pf =SF !! :{} .. Qf"
apply (rule csp_trans[of _ "DIV"])
apply (simp add: csp_Rep_int_choice_DIV)
apply (simp add: csp_Rep_int_choice_DIV_sym)
done

lemma csp_Rep_int_choice1_empty_reflex:
   "! :{} .. Pf =SF ! :{} .. Qf"
apply (simp add: Rep_int_choice_OR1_def)
apply (simp add: csp_Rep_int_choice0_empty_reflex)
done

lemma csp_Rep_int_choice2_empty_reflex:
   "!nat :{} .. Pf =SF !nat :{} .. Qf"
apply (simp add: Rep_int_choice_OR2_def)
apply (simp add: csp_Rep_int_choice0_empty_reflex)
done

lemmas csp_Rep_int_choice_empty_reflex 
     = csp_Rep_int_choice0_empty_reflex
       csp_Rep_int_choice1_empty_reflex
       csp_Rep_int_choice2_empty_reflex

lemmas csp_reflex_ex = csp_Ext_pre_choice_STOP
                       csp_Ext_pre_choice_STOP_sym
                       csp_Rep_int_choice_DIV
                       csp_Rep_int_choice_DIV_sym
                       csp_Ext_pre_choice_empty_reflex
                       csp_Rep_int_choice_empty_reflex

declare csp_reflex_ex [simp]

end
