           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               February 2005               |
            |               June     2005  (modified)   |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_law_fp = CSP_proc_cpo + CSP_proc_cms:

(*****************************************************************

         1. cpo fixed point theory in CSP-Prover
         2. cms fixed point theory in CSP-Prover
         3.
         4. 

 *****************************************************************)

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*=======================================================*
 |                                                       |
 |                        CPO                            |
 |                                                       |
 *=======================================================*)

(*-----------*
 |    Bot    |
 *-----------*)

lemma evalSF_prod_Bot: "(%pn. [[DIV]]SF) = Bot"
apply (simp add: prod_Bot_def)
apply (simp add: expand_fun_eq)
apply (simp add: evalSF_def)
apply (simp add: bottomSF_def)
apply (simp add: evalT_def evalF_def)
apply (simp add: evalt_def evalf_def)
done

(*-----------*
 |   FIX P   |
 *-----------*)

lemma evalSFX_FIX_LUB_SF_lm1:
   "Union (Rep_domT ` fst ` Rep_domSF ` 
           {y. EX x. (EX n. x = ([[P]]SFX ^ n) Bot) & y = x pn})
    = {t. EX n. t :t (fst (Rep_domSF ((([[P]]SFX ^ n) Bot) pn)))}"
by (auto simp add: Union_image_eq memT_def)

lemma evalSFX_FIX_LUB_SF_lm2:
   "Union (Rep_domF ` snd ` Rep_domSF ` 
           {y. EX x. (EX n. x = ([[P]]SFX ^ n) Bot) & y = x pn})
    = {f. EX n. f :f (snd (Rep_domSF ((([[P]]SFX ^ n) Bot) pn)))}"
by (auto simp add: Union_image_eq memF_def)

lemma evalSFX_FIX_LUB_SF: 
    "P : ProcX ==>
     [[(FIX P) pn]]SF
     = LUB_SF {y. EX x. (EX n. x = ([[P]]SFX ^ n) Bot) & y = x pn}"
apply (simp add: evalSF_def)
apply (simp add: pairSF_def)
apply (simp add: LUB_SF_def)
apply (subgoal_tac "EX x n. x = ([[P]]SFX ^ n) Bot")
apply (simp add: Abs_domSF_inject LUB_TF_in)
apply (simp add: LUB_TF_def)
apply (simp add: UnionT_def evalSFX_FIX_LUB_SF_lm1)
apply (simp add: UnionF_def evalSFX_FIX_LUB_SF_lm2)

apply (simp add: FIX_def FIXn_def)
apply (rule) 

 (* T *)
 apply (rule order_antisym)

  (* <= *)
  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (erule disjE, simp)
  apply (elim conjE exE)
  apply (simp add: UNIV_Union_proc_memT)
  apply (rule_tac x="a" in exI)
  apply (simp add: evalSFX_interation_ProcX[THEN sym])
  apply (simp add: fold_fstSF)
  apply (simp add: iteration_in_ProcX evalT_ProcX_compo)
  apply (simp add: evalSF_prod_Bot)

  (* => *)
  apply (rule)
  apply (simp add: evalSFX_interation_ProcX[THEN sym])
  apply (simp add: Rep_int_choice_mem)
  apply (simp add: UNIV_Union_proc_memT)
  apply (elim exE)
  apply (simp add: fold_fstSF)
  apply (simp add: iteration_in_ProcX evalT_ProcX_compo)
  apply (simp add: evalSF_prod_Bot)
  apply (rule disjI2)
  apply (rule_tac x="n" in exI)
  apply (simp)

 (* F *)
 apply (rule order_antisym)

  (* <= *)
  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (elim conjE exE)
  apply (simp add: UNIV_Union_proc_memF)
  apply (rule_tac x="a" in exI)
  apply (simp add: evalSFX_interation_ProcX[THEN sym])
  apply (simp add: fold_sndSF)
  apply (simp add: iteration_in_ProcX evalF_ProcX_compo)
  apply (simp add: evalSF_prod_Bot)

  (* => *)
  apply (rule)
  apply (simp add: evalSFX_interation_ProcX[THEN sym])
  apply (simp add: Rep_int_choice_mem)
  apply (simp add: UNIV_Union_proc_memF)
  apply (elim exE)
  apply (simp add: fold_sndSF)
  apply (simp add: iteration_in_ProcX evalF_ProcX_compo)
  apply (simp add: evalSF_prod_Bot)
  apply (rule_tac x="n" in exI)
  apply (simp)

by (auto)

(*** FIX is LUB ***)

lemma evalSFX_FIX_isLUB:
   "P : ProcX ==>
    (%pn. [[(FIX P) pn]]SF) isLUB {x. EX n. x = ([[P]]SFX^n) Bot}"
apply (simp add: evalSFX_FIX_LUB_SF)
apply (simp add: prod_LUB_decompo)
apply (intro allI)
apply (simp add: proj_fun_def)
apply (simp add: image_def)
apply (rule LUB_SF_isLUB)
apply (auto)
done

lemma evalSFX_FIX_LUB:
   "P : ProcX ==>
    (%pn. [[(FIX P) pn]]SF) = LUB {x. EX n. x = ([[P]]SFX^n) Bot}"
apply (rule sym)
apply (rule isLUB_LUB)
apply (simp add: evalSFX_FIX_isLUB)
done

lemma evalSFX_FIX_LFP:
   "P : ProcX ==> (%pn. [[(FIX P) pn]]SF) = LFP [[P]]SFX"
by (simp add: evalSFX_FIX_LUB evalSFX_Tarski_thm_LUB)

lemma evalSFX_FIX_LFP_pn:
   "P : ProcX ==> [[(FIX P) pn]]SF = LFP [[P]]SFX pn"
apply (insert evalSFX_FIX_LFP[of P])
apply (simp add: expand_fun_eq)
done

lemma evalSFX_FIX_isLFP:
   "P : ProcX ==> (%pn. [[(FIX P) pn]]SF) isLFP [[P]]SFX"
apply (simp add: evalSFX_FIX_LFP)
apply (simp add: evalSFX_Tarski_thm LFP_is)
done

lemma evalSFX_FIX_LFP_fixed_point:
   "P : ProcX ==>  [[P]]SFX (%pn. [[(FIX P) pn]]SF) = (%pn. [[(FIX P) pn]]SF)"
apply (insert evalSFX_FIX_isLFP[of P], simp)
apply (simp add: isLFP_def)
done

lemma evalSFX_FIX_LFP_least:
   "P : ProcX ==> 
    ALL SF. [[P]]SFX SFn = SFn --> (%pn. [[(FIX P) pn]]SF) <= SFn"
apply (insert evalSFX_FIX_isLFP[of P], simp)
apply (simp add: isLFP_def)
done

(*---------------------------------------------------------------------*
 |   FIX is the greatest fixed point of process-expression-functions   |
 |   (semantically least fixed point)                                  |
 *---------------------------------------------------------------------*)

lemma FIX_isGFP:
  "P : ProcX ==> 
   (FIX P) ==SF (P (FIX P)) & (ALL Q. Q ==SF P Q --> Q <==SF (FIX P))"
apply (simp add: refSF_prod_def eqSF_prod_def)
apply (simp add: refSF_def eqSF_def)
apply (insert evalSFX_FIX_isLFP[of P], simp)
apply (rule)

 (* FP *)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: evalSF_ProcX_compo_fun)
 apply (simp add: isLFP_def)

 (* LFP *)
 apply (intro allI impI)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (drule sym)
 apply (simp add: expand_fun_eq)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: evalSF_ProcX_compo_fun)
 apply (simp add: isLFP_def)
 apply (erule conjE)
 apply (drule_tac x="(%n. [[Q n]]SF)" in spec)
 apply (simp add: order_prod_def)
done

(*-------------------------------------------------------*
 |                                                       |
 |           Fixpoint unwind (CSP-Prover rule)           |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_unwind_lm:
  "PX : ProcX ==> ALL pn. (FIX PX) pn =SF PX (FIX PX) pn"
apply (simp add: eqSF_def)
apply (simp add: expand_fun_eq[THEN sym])
apply (simp add: evalSF_ProcX_compo_fun)
apply (simp add: evalSFX_FIX_LFP_fixed_point)
done

(*** csp rule ***)

lemma csp_unwind:
  "PX : ProcX ==> (FIX PX) pn =SF PX (FIX PX) pn"
by (simp add: csp_unwind_lm)

(*-------------------------------------------------------*
 |                                                       |
 |        Check fixpoint  (CSP-Prover intro rule)        |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_fp_induct_cpo_lm:
   "[| PX : ProcX ; Q <=SF Px pn ; ALL pn. Px pn <=SF (PX Px) pn |]
    ==> Q <=SF (FIX PX) pn"
apply (simp add: refSF_def)
apply (simp add: fold_order_prod_def)
apply (simp add: evalSF_ProcX_compo_fun)
apply (simp add: evalSFX_FIX_LFP_pn)
apply (insert cpo_fixpoint_induction_rev
       [of "[[PX]]SFX" "(%pn. [[Px pn]]SF)"])
apply (simp add: evalSFX_continuous)
apply (simp add: order_prod_def)
apply (rotate_tac -1)
apply (drule_tac x="pn" in spec)
apply (auto)
done

(*** csp rule ***)

lemma csp_fp_induct_cpo:
   "[| PX : ProcX ; Q <=SF Px pn ; !! pn. Px pn <=SF (PX Px) pn |]
    ==> Q <=SF (FIX PX) pn"
by (simp add: csp_fp_induct_cpo_lm)

(*=======================================================*
 |                                                       |
 |                        CMS                            |
 |                                                       |
 *=======================================================*)

lemma evalSFX_FIX_UFP:
   "[| P : ProcX ; nohideX P ; guardX P |] ==> 
    (%pn. [[(FIX P) pn]]SF) = UFP [[P]]SFX"
apply (subgoal_tac "[[P]]SFX hasUFP")
apply (insert evalSFX_FIX_isLFP[of P], simp)
apply (subgoal_tac "UFP [[P]]SFX isUFP [[P]]SFX")
apply (simp add: isUFP_def)
apply (simp add: isLFP_def)
apply (elim conjE exE)
apply (rotate_tac -1)
apply (drule_tac x="(%pn. [[(FIX P) pn]]SF)" in spec)
apply (simp)
apply (simp add: UFP_is)
apply (simp add: evalSFX_Banach_thm)
done

lemma evalSFX_FIX_UFP_pn:
   "[| P : ProcX ; nohideX P ; guardX P |] ==> 
    [[(FIX P) pn]]SF = UFP [[P]]SFX pn"
apply (insert evalSFX_FIX_UFP[of P])
apply (simp add: expand_fun_eq)
done

lemma evalSFX_FIX_isUFP:
   "[| P : ProcX ; nohideX P ; guardX P |] ==> 
    (%pn. [[(FIX P) pn]]SF) isUFP [[P]]SFX"
apply (simp add: evalSFX_FIX_UFP)
apply (simp add: evalSFX_Banach_thm UFP_is)
done

lemma evalSFX_FIX_UFP_unique:
   "[| P : ProcX ; nohideX P ; guardX P |] ==> 
    ALL SF. [[P]]SFX SFn = SFn --> (%pn. [[(FIX P) pn]]SF) = SFn"
apply (insert evalSFX_FIX_isUFP[of P], simp)
apply (simp add: isUFP_def)
done

(*------------------------------------------------------------------*
 |   FIX is the unique fixed point of process-expression-functions  |
 *------------------------------------------------------------------*)

lemma FIX_isUFP:
  "[| P : ProcX ; nohideX P ; guardX P |] ==> 
   (FIX P) ==SF (P (FIX P)) & (ALL Q. Q ==SF P Q --> Q ==SF (FIX P))"
apply (simp add: FIX_isGFP)

apply (simp add: refSF_prod_def eqSF_prod_def)
apply (simp add: refSF_def eqSF_def)
apply (insert evalSFX_FIX_isUFP[of P], simp)

 (* UFP *)
 apply (intro allI impI)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (drule sym)
 apply (simp add: expand_fun_eq)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: evalSF_ProcX_compo_fun)
 apply (simp add: isUFP_def)
 apply (erule conjE)
 apply (drule_tac x="(%n. [[Q n]]SF)" in spec)
 apply (simp add: expand_fun_eq)
done

(*-------------------------------------------------------*
 |                                                       |
 |        Check fixpoint  (CSP-Prover intro rule)        |
 |                                                       |
 *-------------------------------------------------------*)

(* refinement *)

lemma csp_fp_induct_cms_ref_lm:
   "[| PX : ProcX ; nohideX PX ; guardX PX ;
       Px pn <=SF Q ; ALL pn. (PX Px) pn <=SF Px pn |]
    ==> (FIX PX) pn <=SF Q"
apply (simp add: refSF_def)
apply (simp add: fold_order_prod_def)
apply (simp add: evalSF_ProcX_compo_fun)

apply (simp add: evalSFX_FIX_UFP_pn)
apply (insert cms_fixpoint_induction_ref
       [of "[[PX]]SFX" "(%pn. [[Px pn]]SF)" "UFP [[PX]]SFX"])
apply (simp add: evalSFX_constructive_rs)
apply (simp add: evalSFX_mono_fun)
apply (subgoal_tac "UFP [[PX]]SFX = [[PX]]SFX (UFP [[PX]]SFX)", simp)
apply (simp add: order_prod_def)
apply (rotate_tac -2)
apply (drule_tac x="pn" in spec)
apply (simp)

apply (insert UFP_is[of "[[PX]]SFX"])
apply (simp add: evalSFX_Banach_thm)
apply (simp add: isUFP_def)
done

(*** csp rule ***)

lemma csp_fp_induct_cms_ref:
   "[| PX : ProcX ; nohideX PX ; guardX PX ;
       Px pn <=SF Q ; !! pn. (PX Px) pn <=SF Px pn |]
    ==> (FIX PX) pn <=SF Q"
by (simp add: csp_fp_induct_cms_ref_lm)

(* equivalence *)

lemma csp_fp_induct_cms_eq_lm:
   "[| PX : ProcX ; nohideX PX ; guardX PX ;
       Px pn =SF Q ; ALL pn. (PX Px) pn =SF Px pn |]
    ==> (FIX PX) pn =SF Q"
apply (simp add: eqSF_def)
apply (simp add: expand_fun_eq[THEN sym])
apply (simp add: evalSF_ProcX_compo_fun)
apply (insert evalSFX_FIX_isUFP[of PX], simp)
apply (simp add: isUFP_def)
apply (elim conjE)
apply (drule_tac x="(%pn. [[Px pn]]SF)" in spec)
apply (simp add: expand_fun_eq)
done

(*** csp rule ***)

lemma csp_fp_induct_cms_eq:
   "[| PX : ProcX ; nohideX PX ; guardX PX ;
       Px pn =SF Q ; !! pn. (PX Px) pn =SF Px pn |]
    ==> (FIX PX) pn =SF Q"
by (simp add: csp_fp_induct_cms_eq_lm)

lemmas csp_fp_induct_cms = csp_fp_induct_cms_ref csp_fp_induct_cms_eq

(*-------------------------------------------------------*
 |                                                       |
 |         FIX expansion (CSP-Prover intro rule)         |
 |                                                       |
 *-------------------------------------------------------*)

lemma evalSF_FIXn_plus_sub: "P : ProcX ==>
  ALL n m pn. [[(FIX[n] P) pn]]SF <= [[(FIX[n+m] P) pn]]SF"
apply (rule)
apply (induct_tac n)
apply (simp add: FIXn_def)
apply (simp add: evalSF_def)
apply (simp add: evalT_def evalF_def)
apply (simp add: evalt_def evalf_def)
apply (fold bottomSF_def)
apply (simp add: bottomSF)

apply (rule allI)
apply (simp add: FIXn_def)
apply (drule_tac x="m" in spec)
apply (simp add: evalSF_ProcX_compo_pn)

apply (fold order_prod_def)
apply (simp add: evalSFX_mono)
done

lemma evalT_evalF_FIXn_plus_sub: "P : ProcX ==>
   [[(FIX[n] P) pn]]T <= [[(FIX[n+m] P) pn]]T &
   [[(FIX[n] P) pn]]F <= [[(FIX[n+m] P) pn]]F"
apply (simp add: proc_subsetSF_decompo[THEN sym])
apply (simp add: evalSF_FIXn_plus_sub)
done

lemma mem_evalT_FIXn_plus_sub: 
  "[| P : ProcX ; t :t [[(FIX[n] P) pn]]T |] ==> t :t [[(FIX[n+m] P) pn]]T"
apply (insert evalT_evalF_FIXn_plus_sub[of P n pn m])
by (auto)

lemma mem_evalF_FIXn_plus_sub: 
  "[| P : ProcX ; f :f [[(FIX[n] P) pn]]F |] ==> f :f [[(FIX[n+m] P) pn]]F"
apply (insert evalT_evalF_FIXn_plus_sub[of P n pn m])
apply (simp)
apply (insert surj_pair[of f])
apply (elim exE)
by (auto)

(*-----------------------------------------------------*
 |  sometimes FIX[n + f n] is useful more than FIX[n]  |
 *-----------------------------------------------------*)

lemma csp_FIX_plus_eq: "P : ProcX ==>
      ALL f pn. (FIX P) pn =SF (!nat n .. ((FIX[n + f n] P) pn))"
apply (simp add: FIX_def)
apply (simp add: eqSF_def)
apply (intro allI)
apply (simp add: proc_eqSF_decompo)
apply (rule)

 (* T *)
 apply (rule order_antisym)

  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (elim conjE exE disjE)
  apply (simp)
  apply (rule disjI2)
  apply (rule_tac x="a" in exI)
  apply (simp add: mem_evalT_FIXn_plus_sub)

  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (elim conjE exE disjE)
  apply (simp)
  apply (rule disjI2)
  apply (rule_tac x="a + f a" in exI)
  apply (simp)

 (* F *)
 apply (rule order_antisym)

  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (elim conjE exE)
  apply (rule_tac x="a" in exI)
  apply (simp add: mem_evalF_FIXn_plus_sub)

  apply (rule)
  apply (simp add: Rep_int_choice_mem)
  apply (elim conjE exE)
  apply (rule_tac x="a + f a" in exI)
  apply (simp)
done

lemma csp_FIX_plus_eq_fun: "P : ProcX ==>
      ALL f. (%pn. (FIX P) pn) ==SF (%pn. (!nat n .. ((FIX[n + f n] P) pn)))"
apply (simp add: eqSF_prod_def)
apply (simp add: csp_FIX_plus_eq)
done

end
