           (*-------------------------------------------*
            |                CSP-Prover                 |
            |                 May 2005                  |
            |                June 2005  (modified)      |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_law_par1 = Index_parallel:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. associativity of |[X,Y]|
         2. symmetricity of |[X,Y]|
         3. monotonicity of |[X,Y]|
         4. 

 *****************************************************************)

(*********************************************************
                        P |[X,Y]| Q
 *********************************************************)

(************************************
 |         SKIP and SKIP            |
 ************************************)

(*** domT ***)

lemma domT_Alpha_parallel_SKIP:
  "[[SKIP |[{}, {}]| SKIP]]T = [[SKIP]]T"
apply (rule order_antisym)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (case_tac "t = None", simp)
 apply (simp add: SKIP_memT)
 apply (simp add: sett_subset_Tick)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (case_tac "t = None", simp)
 apply (simp add: SKIP_memT)
 apply (simp add: sett_subset_Tick)
 apply (insert hide_tr_sett_comp_subseteq[of _ "{}"])
 apply (auto)
done

(*** domF ***)

lemma domF_Alpha_parallel_SKIP:
  "[[SKIP |[{}, {}]| SKIP]]F = [[SKIP]]F"
apply (rule order_antisym)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (elim conjE exE)
 apply (case_tac "s = None", simp)

 apply (simp add: SKIP_memF)
 apply (elim disjE)
 apply (simp_all)
 apply (simp add: Evset_def)
 apply (subgoal_tac "(s = []t | s = [Tick]t)")
 apply (erule disjE)
 apply (simp)
 apply (fast)
 apply (simp)
 apply (simp add: sett_subset_Tick)

 apply (subgoal_tac "(s = []t | s = [Tick]t)")
 apply (erule disjE)
 apply (simp)
 apply (simp)
 apply (simp add: sett_subset_Tick)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (rule_tac x="X" in exI)
 apply (rule_tac x="X" in exI)
 apply (simp)

 apply (simp add: SKIP_memF)
 apply (auto)
done

(*** domSF ***)

lemma domSF_Alpha_parallel_SKIP:
  "[[SKIP |[{}, {}]| SKIP]]SF = [[SKIP]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Alpha_parallel_SKIP)
apply (simp add: domF_Alpha_parallel_SKIP)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Alpha_parallel_SKIP:
  "(SKIP |[{}, {}]| SKIP) =SF SKIP"
by (simp add: eqSF_def domSF_Alpha_parallel_SKIP)

(************************************
 |          associativity           |
 ************************************)

(*** domT ***)

lemma domT_Alpha_parallel_ass:
  "[[(P1 |[X1, X2]| P2) |[X1 Un X2, X3]| P3]]T =
   [[P1 |[X1, X2 Un X3]| (P2 |[X2, X3]| P3)]]T"
apply (rule eq_iffI)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (elim conjE exE)
 apply (simp add: Int_lower1 Int_lower2 hide_tr_of_hide_tr_subset)
 apply (rule conjI)
 apply (simp only: Compl_Un[THEN sym])
 apply (simp only: hide_tr_sett_comp_subseteq)
 apply (fast)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (elim conjE exE)
 apply (simp add: Int_lower1 Int_lower2 hide_tr_of_hide_tr_subset)
 apply (rule conjI)
 apply (simp only: Compl_Un[THEN sym])
 apply (simp only: hide_tr_sett_comp_subseteq)
 apply (fast)
done

(*** domF ***)

lemma domF_Alpha_parallel_ass_lm1:
          "Ya Int insert Tick (Ev ` X1) Un
          (Za Int insert Tick (Ev ` X2) Un Z Int insert Tick (Ev ` X3)) =
           Ya Int insert Tick (Ev ` X1) Un
           (Za Int insert Tick (Ev ` X2) Un Z Int insert Tick (Ev ` X3)) Int
           insert Tick (Ev ` (X2 Un X3))"
by (auto)

lemma domF_Alpha_parallel_ass:
  "[[(P1 |[X1, X2]| P2) |[X1 Un X2, X3]| P3]]F =
   [[P1 |[X1, X2 Un X3]| (P2 |[X2, X3]| P3)]]F"
apply (rule eq_iffI)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (elim conjE exE)
 apply (simp add: Int_lower1 Int_lower2 hide_tr_of_hide_tr_subset)
 apply (rule_tac x="Ya" in exI, simp)
 apply (rule_tac x="(Za Int insert Tick (Ev ` X2)) Un 
                     (Z Int insert Tick (Ev ` X3))" in exI)
 apply (simp add: Un_assoc)
 apply (simp add: domF_Alpha_parallel_ass_lm1)

 apply (rule_tac x="Za" in exI, simp)
 apply (rule_tac x="Z" in exI, simp)
 apply (rule conjI)
 apply (force)
 apply (simp only: Compl_Un[THEN sym])
 apply (simp only: hide_tr_sett_comp_subseteq)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (elim conjE exE)
 apply (simp add: Int_lower1 Int_lower2 hide_tr_of_hide_tr_subset)
 apply (rule_tac x="Y Int insert Tick (Ev ` X1) Un
                  (Ya Int insert Tick (Ev ` X2))" in exI)
 apply (rule_tac x="Za" in exI)
 apply (simp add: Un_assoc)
 apply (rule conjI)
 apply (fast)

 apply (rule_tac x="Y" in exI, simp)
 apply (rule_tac x="Ya" in exI, simp)
 apply (rule conjI)
 apply (force)
 apply (simp only: Compl_Un[THEN sym])
 apply (simp only: hide_tr_sett_comp_subseteq)
done

(*** domSF ***)

lemma domSF_Alpha_parallel_ass:
  "[[(P1 |[X1, X2]| P2) |[X1 Un X2, X3]| P3]]SF =
   [[P1 |[X1, X2 Un X3]| (P2 |[X2, X3]| P3)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Alpha_parallel_ass)
apply (simp add: domF_Alpha_parallel_ass)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Alpha_parallel_ass:
  "(P1 |[X1, X2]| P2) |[X1 Un X2, X3]| P3 =SF
   P1 |[X1, X2 Un X3]| (P2 |[X2, X3]| P3)"
by (simp add: eqSF_def domSF_Alpha_parallel_ass)

lemma csp_Alpha_parallel_ass_sym:
  "P1 |[X1, X2 Un X3]| (P2 |[X2, X3]| P3) =SF
   (P1 |[X1, X2]| P2) |[X1 Un X2, X3]| P3"
apply (rule csp_sym)
by (simp add: csp_Alpha_parallel_ass)

(************************************
 |          symmetricity            |
 ************************************)

(*** domT ***)

lemma domT_Alpha_parallel_sym:
  "[[P1 |[X1, X2]| P2]]T = [[P2 |[X2, X1]| P1]]T"
apply (rule eq_iffI)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (fast)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (fast)
done

(*** domF ***)

lemma domF_Alpha_parallel_sym:
  "[[P1 |[X1, X2]| P2]]F = [[P2 |[X2, X1]| P1]]F"
apply (rule eq_iffI)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (elim conjE exE)
 apply (rule_tac x="Z" in exI)
 apply (rule_tac x="Y" in exI)
 apply (simp add: Un_sym)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (elim conjE exE)
 apply (rule_tac x="Z" in exI)
 apply (rule_tac x="Y" in exI)
 apply (simp add: Un_sym)
done

(*** domSF ***)

lemma domSF_Alpha_parallel_sym:
  "[[P1 |[X1, X2]| P2]]SF = [[P2 |[X2, X1]| P1]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Alpha_parallel_sym)
apply (simp add: domF_Alpha_parallel_sym)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Alpha_parallel_sym:
  "(P1 |[X1, X2]| P2) =SF (P2 |[X2, X1]| P1)"
by (simp add: eqSF_def domSF_Alpha_parallel_sym)

(************************************
 |          monotonicity            |
 ************************************)

(*** T ***)

lemma Alpha_parallel_evalT_mono:
  "[| X1 = X2 ; Y1 = Y2 ;
      [[P1]]T <= [[P2]]T ; [[Q1]]T <= [[Q2]]T |]
   ==> [[P1 |[X1,Y1]| Q1]]T <= [[P2 |[X2,Y2]| Q2]]T"
apply (simp add: Alpha_parallel_def)
apply (simp add: evalT_def)
apply (rule Parallel_evalT_mono)
apply (rule Parallel_evalT_mono)
apply (simp)
apply (simp)
apply (rule Parallel_evalT_mono)
apply (simp)
apply (simp)
done

lemma Alpha_parallel_evalT_cong:
  "[| X1 = X2 ; Y1 = Y2 ;
      [[P1]]T = [[P2]]T ; [[Q1]]T = [[Q2]]T |]
   ==> [[P1 |[X1,Y1]| Q1]]T = [[P2 |[X2,Y2]| Q2]]T"
apply (rule order_antisym)
by (simp_all add: Alpha_parallel_evalT_mono)

(*** F ***)

lemma Alpha_parallel_evalF_mono:
  "[| X1 = X2 ; Y1 = Y2 ;
      [[P1]]F <= [[P2]]F ; [[Q1]]F <= [[Q2]]F |]
   ==> [[P1 |[X1,Y1]| Q1]]F <= [[P2 |[X2,Y2]| Q2]]F"
apply (simp add: Alpha_parallel_def)
apply (simp add: evalF_def)
apply (rule Parallel_evalF_mono)
apply (rule Parallel_evalF_mono)
apply (simp)
apply (simp)
apply (rule Parallel_evalF_mono)
apply (simp)
apply (simp)
done

lemma Alpha_parallel_evalF_cong:
  "[| X1 = X2 ; Y1 = Y2 ;
      [[P1]]F = [[P2]]F ; [[Q1]]F = [[Q2]]F |]
   ==> [[P1 |[X1,Y1]| Q1]]F = [[P2 |[X2,Y2]| Q2]]F"
apply (rule order_antisym)
by (simp_all add: Alpha_parallel_evalF_mono)

(*** SF ***)

lemma Alpha_parallel_evalSF_mono:
  "[| X1 = X2 ; Y1 = Y2 ;
      [[P1]]SF <= [[P2]]SF ; [[Q1]]SF <= [[Q2]]SF |]
   ==> [[P1 |[X1,Y1]| Q1]]SF <= [[P2 |[X2,Y2]| Q2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (simp add: Alpha_parallel_evalT_mono)
apply (simp add: Alpha_parallel_evalF_mono)
done

lemma Alpha_parallel_evalSF_cong:
  "[| X1 = X2 ; Y1 = Y2 ;
      [[P1]]SF = [[P2]]SF ; [[Q1]]SF = [[Q2]]SF |]
   ==> [[P1 |[X1,Y1]| Q1]]SF = [[P2 |[X2,Y2]| Q2]]SF"
apply (rule order_antisym)
by (simp_all add: Alpha_parallel_evalSF_mono)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Alpha_parallel_mono:
  "[| X1 = X2 ; Y1 = Y2 ;
      P1 <=SF Q1 ; 
      P2 <=SF Q2 |]
           ==> P1 |[X1,Y1]| P2 <=SF Q1 |[X2,Y2]| Q2"
by (simp add: refSF_def Alpha_parallel_evalSF_mono)

lemma csp_Alpha_parallel_cong:
  "[| X1 = X2 ; Y1 = Y2 ;
      P1 =SF Q1 ; 
      P2 =SF Q2 |]
           ==> P1 |[X1,Y1]| P2 =SF Q1 |[X2,Y2]| Q2"
by (simp add: csp_eq_ref_iff csp_Alpha_parallel_mono)

(****************** to add them again ******************)

declare not_None_eq    [simp]

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end

