           (*-------------------------------------------*
            |                CSP-Prover                 |
            |                 May 2005                  |
            |                June 2005  (modified)      |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_law_par2 = CSP_law_par1:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. associativity of [||]:I
         2. symmetricity of [||]:I
         3. 
         4. 

 *****************************************************************)

(*****************************************************
   replace an index set with another equal index set
 *****************************************************)

(*** T ***)

lemma domT_Index_parallel_index_eq:
   "[| finite I1 ;
       EX f. I2 = f ` I1 &
             (ALL i:I1. PXf2 (f i) = PXf1 i) |]
     ==> [[[||]:I1 PXf1]]T = [[[||]:I2 PXf2]]T"
apply (case_tac "I1 = {}")
apply (simp)
apply (rule order_antisym)

 (* <= *)
 apply (rule)
 apply (elim conjE exE)
 apply (simp add: Index_parallel_memT)
 apply (subgoal_tac "Union (snd ` PXf2 ` f ` I1) = Union (snd ` PXf1 ` I1)")
 apply (simp)
 apply (simp add: Union_index_fun)

 (* => *)
 apply (rule)
 apply (elim conjE exE)
 apply (simp add: Index_parallel_memT)
 apply (subgoal_tac "Union (snd ` PXf2 ` f ` I1) = Union (snd ` PXf1 ` I1)")
 apply (simp)
 apply (simp add: Union_index_fun)
done

(*** F ***)

lemma domF_Index_parallel_index_eq_lm1:
   "[| inj_on f I1 ; ALL i:I1. PXf2 (f i) = PXf1 i |] ==>
    Union {Yf i Int insert Tick (Ev ` snd (PXf1 i)) |i. i : I1} =
    Union {Yf (inv_on I1 f i) Int insert Tick (Ev ` snd (PXf2 i)) |i. i : f ` I1}"
apply (rule) 

 (* => *)
 apply (rule, simp)
 apply (elim conjE exE)
 apply (rule_tac x="(Yf i) Int insert Tick (Ev ` snd (PXf2 (f i)))" in exI)
 apply (simp)
 apply (rule_tac x="f i" in exI)
 apply (simp add: inv_f_f_on)

 (* <= *)
 apply (rule, simp)
 apply (simp add: image_iff)
 apply (elim conjE exE bexE)
 apply (rule_tac x=
   "(Yf xb) Int insert Tick (Ev ` snd (PXf1 xb))" in exI)
 apply (simp)
 apply (simp add: inv_f_f_on)
 apply (rule_tac x="xb" in exI)
 apply (simp)
done

lemma domF_Index_parallel_index_eq_lm2:
   "ALL i:I1. PXf2 (f i) = PXf1 i ==>
    Union {Yf i Int insert Tick (Ev ` snd (PXf2 i)) |i. i : f ` I1} =
    Union {Yf (f i) Int insert Tick (Ev ` snd (PXf1 i)) |i. i : I1}"
apply (rule) 

 (* => *)
 apply (rule, simp)
 apply (simp add: image_iff)
 apply (elim conjE exE bexE)
 apply (simp)
 apply (rule_tac x="xa" in exI)
 apply (simp)
 apply (rule_tac x="xb" in exI)
 apply (simp)

 (* <= *)
 apply (rule, simp)
 apply (elim conjE exE)
 apply (rule_tac x="xa" in exI)
 apply (simp)
 apply (rule_tac x="f i" in exI)
 apply (simp)
done

(* main *)

lemma domF_Index_parallel_index_eq:
   "[| finite I1 ;
       EX f. I2 = f ` I1 & inj_on f I1 &
             (ALL i:I1. PXf2 (f i) = PXf1 i) |]
     ==> [[[||]:I1 PXf1]]F = [[[||]:I2 PXf2]]F"
apply (case_tac "I1 = {}")
apply (simp)
apply (rule order_antisym)

 (* <= *)
 apply (rule)
 apply (elim conjE exE)
 apply (simp add: Index_parallel_memF)
 apply (subgoal_tac "Union (snd ` PXf2 ` f ` I1) = Union (snd ` PXf1 ` I1)")
 apply (simp)
 apply (elim conjE exE)
 apply (rule_tac x="(%i. Yf (inv_on I1 f i))" in exI)
 apply (simp add: inv_f_f_on)
 apply (simp add: domF_Index_parallel_index_eq_lm1)
 apply (simp add: Union_index_fun)

 (* => *)
 apply (rule)
 apply (elim conjE exE)
 apply (simp add: Index_parallel_memF)
 apply (subgoal_tac "Union (snd ` PXf2 ` f ` I1) = Union (snd ` PXf1 ` I1)")
 apply (simp)
 apply (elim conjE exE)
 apply (rule_tac x="(%i. Yf (f i))" in exI)
 apply (simp)
 apply (simp add: domF_Index_parallel_index_eq_lm2)
 apply (simp add: Union_index_fun)
done

(*** domSF ***)

lemma domSF_Index_parallel_index_eq:
   "[| finite I1 ;
       EX f. I2 = f ` I1 & inj_on f I1 &
             (ALL i:I1. PXf2 (f i) = PXf1 i) |]
     ==> [[[||]:I1 PXf1]]SF = [[[||]:I2 PXf2]]SF"
apply (simp add: proc_eqSF_decompo)
apply (rule conjI)
apply (rule domT_Index_parallel_index_eq)
apply (simp)
apply (fast)
apply (simp add: domF_Index_parallel_index_eq)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Alpha_parallel_sym:
   "[| finite I1 ;
       EX f. I2 = f ` I1 & inj_on f I1 &
             (ALL i:I1. PXf2 (f i) = PXf1 i) |]
    ==> [||]:I1 PXf1 =SF [||]:I2 PXf2"
by (simp add: eqSF_def domSF_Index_parallel_index_eq)

(*********************************************************
                [||]:I PXf ==> [||] PXs
 *********************************************************)

(****** T ******)

lemma domT_Index_to_Rep_parallel:
  "[| finite I ; Is isListOf I |]
   ==> [[[||]:I PXf]]T = [[[||] (map PXf Is)]]T"
apply (case_tac "I = {}")
apply (simp)

apply (case_tac "map PXf Is = []")
apply (simp)
apply (rule order_antisym)

 (* <= *)
 apply (rule)
 apply (simp add: Index_parallel_memT)
 apply (simp add: Rep_parallel_memT_nth)
 apply (simp add: isListOf_set_eq)
 apply (intro allI impI)
 apply (elim conjE exE)
 apply (drule_tac x="Is!i" in bspec)
 apply (simp add: isListOf_nth_in_index)
 apply (simp)

 (* => *)
 apply (rule)
 apply (simp add: Index_parallel_memT)
 apply (simp add: Rep_parallel_memT_nth)
 apply (simp add: isListOf_set_eq)
 apply (intro ballI)
 apply (elim conjE)
 apply (erule isListOf_index_to_nthE)
 apply (drule_tac x="i" in bspec, simp)
 apply (elim exE conjE)
 apply (drule_tac x="n" in spec, simp)
done

(****** F ******)

lemma domF_Index_to_Rep_parallel:
  "[| finite I ; Is isListOf I |]
   ==> [[[||]:I PXf]]F = [[[||] (map PXf Is)]]F"
apply (case_tac "I = {}")
apply (simp)

apply (case_tac "map PXf Is = []")
apply (simp)
apply (rule order_antisym)

 (* <= *)
 apply (rule)
 apply (simp add: Index_parallel_memF)
 apply (simp add: Rep_parallel_memF_nth)
 apply (simp add: isListOf_set_eq)
 apply (elim conjE exE)
  apply (rule_tac x= "map Yf Is" in exI)
  apply (simp)
  apply (rule conjI)
  apply (rule Index_parallel_memF_lm2, simp)
  apply (intro allI impI)
  apply (drule_tac x="Is ! i" in bspec)
  apply (simp add: isListOf_nth_in_index)
  apply (simp)

 (* => *)
 apply (rule)
 apply (simp add: Index_parallel_memF)
 apply (simp add: Rep_parallel_memF_nth)
 apply (simp add: isListOf_set_eq)
 apply (elim conjE exE)
 apply (rule_tac x= "(%i. (Ys!(THE n. (Is!n) = i & n<length Is)))" in exI)
  apply (rule conjI)
  apply (simp add: Index_parallel_memF_lm1)
  apply (rule ballI)

  apply (erule isListOf_index_to_nthE)
  apply (drule_tac x="i" in bspec, simp)
  apply (elim conjE exE, simp)

  apply (drule_tac x="n" in spec, simp)
  apply (rotate_tac 4)
  apply (drule sym)
  apply (simp add: isListOf_THE_nth)
done

(*** domSF ***)

lemma domSF_Index_to_Rep_parallel:
  "[| finite I ; Is isListOf I |]
   ==> [[[||]:I PXf]]SF = [[[||] (map PXf Is)]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Index_to_Rep_parallel)
apply (simp add: domF_Index_to_Rep_parallel)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Index_to_Rep_parallel:
  "[| finite I ; Is isListOf I |] ==>
   [||]:I PXf =SF
   [||] (map PXf Is)"
by (simp add: eqSF_def domSF_Index_to_Rep_parallel)

(************************************
 |       [||]:I PXf and SKIP        |
 ************************************)

(*** domT ***)

lemma domT_Index_parallel_SKIP_right:
  " finite I ==>
     [[[||]:I PXf |[Union (snd ` PXf ` I), {}]| SKIP]]T
   = [[[||]:I PXf]]T"
apply (case_tac "I={}")
apply (simp add: domT_Alpha_parallel_SKIP)
apply (rule order_antisym)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (simp add: Index_parallel_memT)
 apply (simp add: hide_tr_sett_comp_subseteq)

 apply (intro ballI)
 apply (elim conjE)
 apply (drule_tac x="i" in bspec, simp)
 apply (subgoal_tac "(- Union (snd ` PXf ` I)) <= (- snd (PXf i))")
 apply (simp add: hide_tr_of_hide_tr_subset)
 apply (force)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (simp add: Index_parallel_memT)
 apply (simp add: hide_tr_sett_comp_subseteq)

 apply (case_tac "t = None", simp)
 apply (simp add: SKIP_memT)
 apply (simp add: hide_tr_UNIV)
 apply (intro ballI)
 apply (elim conjE)
 apply (drule_tac x="i" in bspec, simp)
 apply (subgoal_tac "(- Union (snd ` PXf ` I)) <= (- snd (PXf i))")
 apply (simp add: hide_tr_of_hide_tr_subset)
 apply (force)
done

(*** domF ***)

lemma domF_Index_parallel_SKIP_right_lm1: "B = {} ==> A Un B = A"
by (auto)

lemma domF_Index_parallel_SKIP_right_lm2:
  "I ~= {} ==>
   insert Tick (Union {Yf i Int insert Tick (Ev ` snd (PXf i)) |i. i : I})
   = Union {insert Tick (Yf i Int Ev ` snd (PXf i)) |i. i : I}"
by (auto)

(* main *)

lemma domF_Index_parallel_SKIP_right:
  "finite I ==>
   [[([||]:I PXf) |[Union (snd ` PXf ` I), {}]| SKIP]]F =
   [[[||]:I PXf ]]F"
apply (case_tac "I = {}")
apply (simp add: domF_Alpha_parallel_SKIP)
apply (rule order_antisym)

(* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (simp add: Index_parallel_memF)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (elim conjE exE)

  apply (case_tac "Tick ~: Z")
  apply (simp)
  apply (rule_tac x="Yf" in exI)
  apply (rule conjI)
  apply (rule domF_Index_parallel_SKIP_right_lm1)
  apply (force)
  apply (intro ballI)
  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)

  apply (simp add: SKIP_memF)
  apply (erule disjE)
  apply (simp add: Evset_def)
  apply (force)

  apply (rule_tac x="(%i. insert Tick (Yf i))" in exI)
  apply (simp)
  apply (subgoal_tac "Z Int {Tick} = {Tick}", simp)
  apply (simp add: domF_Index_parallel_SKIP_right_lm2)

  apply (intro ballI)
  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)

  apply (case_tac "s = None", simp)
  apply (simp add: hide_tr_Tick_sett)
  apply (elim conjE exE)
  apply (simp)
  apply (rule proc_T2_T3, simp)
  apply (force)
  apply (force)

(* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (rule_tac x="X" in exI)
 apply (rule_tac x="{}" in exI)
 apply (simp)
 apply (case_tac "s = None", simp)

 apply (simp add: SKIP_memF)
 apply (simp add: hide_tr_UNIV)
 apply (simp add: Index_parallel_memF)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (elim conjE exE)
 apply (rule_tac x="Yf" in exI)
 apply (simp)
 apply (intro ballI)
 apply (drule_tac x="i" in bspec, simp)

 apply (subgoal_tac "(- Union (snd ` PXf ` I)) <= (- snd (PXf i))")
 apply (simp add: hide_tr_of_hide_tr_subset)
 apply (force)
done

(*** domSF ***)

lemma domSF_Index_parallel_SKIP_right:
  "finite I ==>
   [[[||]:I PXf |[Union (snd ` PXf ` I), {}]| SKIP]]SF =
   [[[||]:I PXf ]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Index_parallel_SKIP_right)
apply (simp add: domF_Index_parallel_SKIP_right)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Index_parallel_SKIP_right:
  "finite I ==>
   (([||]:I PXf) |[Union (snd `  PXf ` I), {}]| SKIP) =SF
   ([||]:I PXf)"
by (simp add: eqSF_def domSF_Index_parallel_SKIP_right)

(************************************
 |        SKIP and [||]:I PXf       |
 ************************************)

(*** domT ***)

lemma domT_Index_parallel_SKIP_left:
  "finite I ==>
     [[SKIP |[{}, Union (snd ` PXf ` I)]| ([||]:I PXf)]]T
   = [[[||]:I PXf ]]T"
apply (subgoal_tac 
    "[[SKIP |[{}, Union (snd ` PXf ` I)]| [||]:I PXf]]T =
     [[[||]:I PXf |[Union (snd ` PXf ` I), {}]| SKIP]]T")
apply (simp add: domT_Index_parallel_SKIP_right)
apply (simp add: domT_Alpha_parallel_sym)
done

(*** domF ***)

lemma domF_Index_parallel_SKIP_left:
  "finite I ==>
     [[SKIP |[{}, Union (snd ` PXf ` I)]| ([||]:I PXf)]]F
   = [[[||]:I PXf ]]F"
apply (subgoal_tac 
    "[[SKIP |[{}, Union (snd ` PXf ` I)]| [||]:I PXf]]F =
     [[[||]:I PXf |[Union (snd ` PXf ` I), {}]| SKIP]]F")
apply (simp add: domF_Index_parallel_SKIP_right)
apply (simp add: domF_Alpha_parallel_sym)
done

(*** domSF ***)

lemma domSF_Index_parallel_SKIP_left:
  "finite I ==>
     [[SKIP |[{}, Union (snd ` PXf ` I)]| ([||]:I PXf)]]SF
   = [[[||]:I PXf ]]SF"
apply (subgoal_tac 
    "[[SKIP |[{}, Union (snd ` PXf ` I)]| [||]:I PXf]]SF =
     [[[||]:I PXf |[Union (snd ` PXf ` I), {}]| SKIP]]SF")
apply (simp add: domSF_Index_parallel_SKIP_right)
apply (simp add: domSF_Alpha_parallel_sym)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Index_parallel_SKIP_left:
  "finite I ==>
   (SKIP |[{}, Union (snd ` PXf ` I)]| ([||]:I PXf)) =SF
   ([||]:I PXf)"
by (simp add: eqSF_def domSF_Index_parallel_SKIP_left)

(*** left and right ***)

lemmas domT_Index_parallel_SKIP = domT_Index_parallel_SKIP_left
                                domT_Index_parallel_SKIP_right

lemmas domF_Index_parallel_SKIP = domF_Index_parallel_SKIP_left
                                domF_Index_parallel_SKIP_right

lemmas domSF_Index_parallel_SKIP = domSF_Index_parallel_SKIP_left
                                 domSF_Index_parallel_SKIP_right

lemmas csp_Index_parallel_SKIP = csp_Index_parallel_SKIP_left
                               csp_Index_parallel_SKIP_right

(************************************
 |          associativity           |
 ************************************)

(*** domT ***)

lemma domT_Index_parallel_ass:
  "[| finite I1 ; finite I2 |] ==>
   [[[||]:(I1 Un I2) PXf]]T =
   [[[||]:I1 PXf |[Union (snd ` PXf ` I1), Union (snd ` PXf ` I2)]| [||]:I2 PXf]]T"
apply (case_tac "I1 = {}")
apply (case_tac "I2 = {}")
apply (simp add: domT_Alpha_parallel_SKIP)
apply (simp add: domT_Index_parallel_SKIP)
apply (case_tac "I2 = {}")
apply (simp add: domT_Index_parallel_SKIP)
apply (rule eq_iffI)

 (* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (simp add: Index_parallel_memT)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (simp add: Union_snd_Un)
 apply (elim conjE)
 apply (rule conjI)

  apply (intro ballI)
  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I1)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)

  apply (intro ballI)
  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I2)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)

 (* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memT)
 apply (simp add: Index_parallel_memT)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (simp add: Union_snd_Un)
 apply (elim conjE)

 apply (intro ballI)
 apply (simp)
 apply (erule disjE)

  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I1)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)

  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I2)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)
done

(*** domF ***)

lemma domF_Index_parallel_ass_lm1:
   "Union {Yf i Int insert Tick (Ev ` snd (PXf i)) |i. i : I} Int
           insert Tick (Ev ` Union (snd ` PXf ` I))
    = Union {Yf i Int insert Tick (Ev ` snd (PXf i)) |i. i : I}"
by (auto)

lemma domF_Index_parallel_ass_lm2:
   "I1 Int I2 = {} ==>
    Union {Yf1 i Int insert Tick (Ev ` snd (PXf i)) |i. i : I1} Un
    Union {Yf2 i Int insert Tick (Ev ` snd (PXf i)) |i. i : I2} =
    Union {(if i : I1 then Yf1 i else Yf2 i) Int insert Tick (Ev ` snd (PXf i)) |i.
           i : I1 | i : I2}"
apply (rule)

 apply (rule)
 apply (simp)
 apply (elim disjE conjE exE)

  apply (rule_tac x="Yf1 i Int insert Tick (Ev ` snd (PXf i))" in exI, simp)
  apply (rule_tac x="i" in exI, simp)
  apply (rule_tac x="Yf2 i Int insert Tick (Ev ` snd (PXf i))" in exI, simp)
  apply (rule_tac x="i" in exI, simp)
  apply (force)

 apply (rule)
 apply (simp)
 apply (elim disjE conjE exE)

  apply (rule disjI1)
  apply (rule_tac x="Yf1 i Int insert Tick (Ev ` snd (PXf i))" in exI, simp)
  apply (rule_tac x="i" in exI, simp)
  apply (rule disjI2)
  apply (case_tac "i ~: I1")
  apply (rule_tac x="Yf2 i Int insert Tick (Ev ` snd (PXf i))" in exI, simp)
  apply (rule_tac x="i" in exI, simp)
  apply (force)
done

lemma domF_Index_parallel_ass:
  "[| I1 Int I2 = {} ; finite I1 ; finite I2 |] ==>
   [[[||]:(I1 Un I2) PXf]]F =
   [[[||]:I1 PXf |[Union (snd ` PXf ` I1), Union (snd ` PXf ` I2)]| [||]:I2 PXf]]F"
apply (case_tac "I1 = {}")
apply (case_tac "I2 = {}")
apply (simp add: domF_Alpha_parallel_SKIP)
apply (simp add: domF_Index_parallel_SKIP)
apply (case_tac "I2 = {}")
apply (simp add: domF_Index_parallel_SKIP)
apply (rule eq_iffI)

 (* => *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (simp add: Index_parallel_memF)
 apply (elim conjE exE)

 apply (rule_tac x="Union {Yf i Int insert Tick (Ev ` snd (PXf i)) |i. i : I1}" in exI)
 apply (rule_tac x="Union {Yf i Int insert Tick (Ev ` snd (PXf i)) |i. i : I2}" in exI)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (simp add: Union_snd_Un)
 apply (rule conjI)
  apply (simp add: domF_Index_parallel_ass_lm1)
  apply (blast)

 apply (rule conjI)

  (* I1 *)
  apply (rule_tac x="Yf" in exI)
  apply (simp add: domF_Index_parallel_ass_lm1)

  apply (intro ballI)
  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I1)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)

  (* I2 *)
  apply (rule_tac x="Yf" in exI)
  apply (simp add: domF_Index_parallel_ass_lm1)

  apply (intro ballI)
  apply (drule_tac x="i" in bspec, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I2)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)

 (* <= *)
 apply (rule)
 apply (simp add: Alpha_parallel_memF)
 apply (simp add: Index_parallel_memF)
 apply (elim exE conjE)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (simp add: Union_snd_Un)
 apply (rename_tac s X Y Z Yf1 Yf2)

 apply (rule_tac x="(%i. if i:I1 then Yf1 i else Yf2 i)" in exI)
       (* the necessity of the condition "I1 Int I2 = {}" *)
 apply (simp add: domF_Index_parallel_ass_lm2)

 apply (intro ballI)
 apply (simp)
 apply (erule disjE)

  apply (drule_tac x="i" in bspec, simp, simp)
  apply (subgoal_tac "(- Union (snd ` PXf ` I1)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)

  apply (drule_tac x="i" in bspec, simp)
  apply (case_tac "i ~: I1")
  apply (subgoal_tac "(- Union (snd ` PXf ` I2)) <= (- snd (PXf i))")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (force)
  apply (blast)
done

(*** domSF ***)

lemma domSF_Index_parallel_ass:
  "[| I1 Int I2 = {} ; finite I1 ; finite I2 |] ==>
   [[[||]:(I1 Un I2) PXf]]SF =
   [[[||]:I1 PXf |[Union (snd ` PXf ` I1), Union (snd ` PXf ` I2)]| [||]:I2 PXf]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Index_parallel_ass)
apply (simp add: domF_Index_parallel_ass)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Index_parallel_ass:
 "[| I1 Int I2 = {} ; finite I1 ; finite I2 |] ==>
  [||]:(I1 Un I2) PXf =SF
  [||]:I1 PXf |[Union (snd ` PXf ` I1), Union (snd ` PXf ` I2)]| [||]:I2 PXf"
by (simp add: eqSF_def domSF_Index_parallel_ass)

(************************************
 |             induct               |
 ************************************)

(*** domT ***)

lemma domT_Index_parallel_induct:
  "finite I ==>
   [[[||]:(insert i I) PXf]]T =
   [[ fst (PXf i) |[snd (PXf i), Union (snd ` PXf ` I)]| [||]:I PXf]]T"
apply (insert domT_Index_parallel_ass[of "{i}" I PXf])
apply (simp add: Index_parallel_one)
apply (insert domT_Alpha_parallel_ass
  [of "fst (PXf i)" "snd (PXf i)" "{}" "SKIP" "Union (snd ` PXf ` I)" "[||]:I PXf"])
apply (simp)

apply (rule Alpha_parallel_evalT_cong, simp_all)
apply (simp add: domT_Index_parallel_SKIP_left)
done

(*** domF ***)

lemma domF_Index_parallel_induct:
  "[| finite I ; i ~: I |] ==>
   [[[||]:(insert i I) PXf]]F =
   [[ fst (PXf i) |[snd (PXf i), Union (snd ` PXf ` I)]| [||]:I PXf]]F"
apply (insert domF_Index_parallel_ass[of "{i}" I PXf])
apply (simp)

apply (simp add: Index_parallel_one)
apply (insert domF_Alpha_parallel_ass
  [of "fst (PXf i)" "snd (PXf i)" "{}" "SKIP" "Union (snd ` PXf ` I)" "[||]:I PXf"])
apply (simp)

apply (rule Alpha_parallel_evalF_cong, simp_all)
apply (simp add: domF_Index_parallel_SKIP_left)
done

(*** domSF ***)

lemma domSF_Index_parallel_induct:
  "[| finite I ; i ~: I |] ==>
   [[[||]:(insert i I) PXf]]SF =
   [[ fst (PXf i) |[snd (PXf i), Union (snd ` PXf ` I)]| [||]:I PXf]]SF"
apply (simp add: proc_eqSF_decompo)
apply (simp add: domT_Index_parallel_induct)
apply (simp add: domF_Index_parallel_induct)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Index_parallel_induct:
 "[| finite I ; i ~: I |] ==>
  [||]:(insert i I) PXf =SF
  fst (PXf i) |[snd (PXf i), Union (snd ` PXf ` I)]| [||]:I PXf"
by (simp add: eqSF_def domSF_Index_parallel_induct)

(****************** to add them again ******************)

declare not_None_eq    [simp]

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end

