           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June     2005  (modified)   |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_law_ref = CSP_law_commut + CSP_law_decompo:

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. rules for refinement

 *****************************************************************)

(*-------------------------------------------------------*
 |              decompose Internal choice                |
 *-------------------------------------------------------*)

(*** or <= ***)                                         (* unsafe *)

lemma domSF_Int_choice_left1:
  "[[Q]]SF <= [[P1]]SF
   ==> [[Q]]SF <= [[P1 |~| P2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
by (rule, simp add: Int_choice_mem, force)+

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Int_choice_left1:
  "P1 <=SF Q ==> P1 |~| P2 <=SF Q"
by (simp add: refSF_def domSF_Int_choice_left1)

lemma csp_Int_choice_left2:
  "P2 <=SF Q ==> P1 |~| P2 <=SF Q"
apply (rule csp_rw_left)
apply (rule csp_commut)
by (simp add: csp_Int_choice_left1)

(*** <= and ***)                                          (* safe *)

lemma domSF_Int_choice_right1:
  "[| [[Q1]]SF <= [[P]]SF ; [[Q2]]SF <= [[P]]SF |]
   ==> [[Q1 |~| Q2]]SF <= [[P]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
by (rule, simp add: Int_choice_mem, force)+

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Int_choice_right:
  "[| P <=SF Q1 ; P <=SF Q2 |]
      ==> P <=SF Q1 |~| Q2"
by (simp add: refSF_def domSF_Int_choice_right1)

(*-------------------------------------------------------*
 |        decompose Replicated internal choice           |
 *-------------------------------------------------------*)

(*** EX <= ***)                                           (* unsafe *)

lemma domSF_Rep_int_choice_left:
  "(EX a. a:X & [[Q]]SF <= [[Pf a]]SF)
   ==> [[Q]]SF <= [[!! :X .. Pf]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
by (rule, simp add: Rep_int_choice_mem, force)+

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Rep_int_choice0_left:
  "(EX a. a:X & Pf a <=SF Q)
      ==> !! :X .. Pf <=SF Q"
by (simp add: refSF_def domSF_Rep_int_choice_left)

lemma csp_Rep_int_choice0_left_x:
  "[| x:X ; Pf x <=SF Q |]
  ==> !! :X .. Pf <=SF Q"
apply (rule csp_Rep_int_choice0_left)
by (fast)

lemma csp_Rep_int_choice1_left:
  "(EX a. a:X & Pf a <=SF Q)
      ==> ! :X .. Pf <=SF Q"
apply (simp add: Rep_int_choice_OR1_def)
apply (rule csp_Rep_int_choice0_left)
apply (elim conjE exE)
apply (rule_tac x="OR1 a" in exI)
by (simp)

lemma csp_Rep_int_choice1_left_x:
  "[| x:X ; Pf x <=SF Q |]
  ==> ! :X .. Pf <=SF Q"
apply (rule csp_Rep_int_choice1_left)
by (fast)

lemma csp_Rep_int_choice2_left:
  "(EX a. a:X & Pf a <=SF Q)
      ==> !nat :X .. Pf <=SF Q"
apply (simp add: Rep_int_choice_OR2_def)
apply (rule csp_Rep_int_choice0_left)
apply (elim conjE exE)
apply (rule_tac x="OR2 a" in exI)
by (simp)

lemma csp_Rep_int_choice2_left_x:
  "[| x:X ; Pf x <=SF Q |]
  ==> !nat :X .. Pf <=SF Q"
apply (rule csp_Rep_int_choice2_left)
by (fast)

lemmas csp_Rep_int_choice_left = csp_Rep_int_choice0_left
                                 csp_Rep_int_choice1_left
                                 csp_Rep_int_choice2_left

lemmas csp_Rep_int_choice_left_x = csp_Rep_int_choice0_left_x
                                   csp_Rep_int_choice1_left_x
                                   csp_Rep_int_choice2_left_x

(*** <= ALL ***)                                         (* safe *)

lemma domSF_Rep_int_choice_right:
  "[| !!a. a:X ==> [[Qf a]]SF <= [[P]]SF |]
        ==> [[!! :X .. Qf]]SF <= [[P]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
by (rule, simp add: Rep_int_choice_mem, force)+

lemma domT_Rep_int_choice_right_invI:
  "[| [[!! :X .. Qf]]T <= [[P]]T ; a:X |]
   ==> [[Qf a]]T <= [[P]]T"
apply (rule)
apply (simp add: subsetT_iff)
apply (simp add: Rep_int_choice_mem)
apply (drule_tac x="t" in spec)
apply (drule mp)
 apply (rule_tac x="a" in bexI)
 apply (simp)
 apply (simp)
apply (simp)
done

lemma domT_Rep_int_choice_right_inv:
  "[[!! :X .. Qf]]T <= [[P]]T
   ==> ALL a:X. [[Qf a]]T <= [[P]]T"
apply (rule ballI)
apply (simp add: domT_Rep_int_choice_right_invI)
done

lemma domF_Rep_int_choice_right_invI:
  "[| [[!! :X .. Qf]]F <= [[P]]F ; a:X |]
   ==> [[Qf a]]F <= [[P]]F"
apply (rule)
apply (simp add: subsetF_iff)
apply (simp add: Rep_int_choice_mem)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Xa" in spec)
apply (drule mp)
 apply (rule_tac x="a" in bexI)
 apply (simp)
 apply (simp)
apply (simp)
done

lemma domF_Rep_int_choice_right_inv:
  "[[!! :X .. Qf]]F <= [[P]]F
   ==> ALL a:X. [[Qf a]]F <= [[P]]F"
apply (intro ballI)
apply (simp add: domF_Rep_int_choice_right_invI)
done

lemma domSF_Rep_int_choice_right_invI:
  "[| [[!! :X .. Qf]]SF <= [[P]]SF ; a:X |]
   ==> [[Qf a]]SF <= [[P]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (erule conjE)
apply (simp add: domT_Rep_int_choice_right_invI)
apply (simp add: domF_Rep_int_choice_right_invI)
done

lemma domSF_Rep_int_choice_right_inv:
  "[[!! :X .. Qf]]SF <= [[P]]SF
   ==> ALL a:X. [[Qf a]]SF <= [[P]]SF"
apply (intro ballI)
apply (simp add: domSF_Rep_int_choice_right_invI)
done

(* E *)

lemma domT_Rep_int_choice_rightE:
  "[| [[!! :X .. Qf]]T <= [[P]]T ; 
      ALL a:X. [[Qf a]]T <= [[P]]T ==> R
   |] ==> R"
by (simp add: domT_Rep_int_choice_right_inv)

lemma domF_Rep_int_choice_rightE:
  "[| [[!! :X .. Qf]]F <= [[P]]F ; 
      ALL a:X. [[Qf a]]F <= [[P]]F ==> R
   |] ==> R"
by (simp add: domF_Rep_int_choice_right_inv)

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Rep_int_choice0_right:
  "[| !!a. a:X ==> P <=SF Qf a |]
               ==> P <=SF !! :X .. Qf"
by (simp add: refSF_def domSF_Rep_int_choice_right)

lemma csp_Rep_int_choice0_rightE:
  "[| P <=SF !! :X .. Qf ;
      ALL a:X. P <=SF Qf a ==> R
   |] ==> R"
apply (simp add: refSF_def)
apply (simp add: domSF_Rep_int_choice_right_inv)
done

lemma csp_Rep_int_choice1_right:
  "[| !!a. a:X ==> P <=SF Qf a |]
               ==> P <=SF ! :X .. Qf"
apply (simp add: Rep_int_choice_OR1_def)
apply (rule csp_Rep_int_choice0_right)
apply (simp add: image_iff)
apply (elim bexE)
by (simp)

lemma csp_Rep_int_choice1_rightE:
  "[| P <=SF ! :X .. Qf ;
      ALL a:X. P <=SF Qf a ==> R
   |] ==> R"
apply (simp add: Rep_int_choice_OR1_def)
apply (rule csp_Rep_int_choice0_rightE)
apply (simp)
apply (erule csp_Rep_int_choice0_rightE)
apply (subgoal_tac "ALL a:X. P <=SF Qf a")
by (auto)

lemma csp_Rep_int_choice2_right:
  "[| !!a. a:X ==> P <=SF Qf a |]
               ==> P <=SF !nat :X .. Qf"
apply (simp add: Rep_int_choice_OR2_def)
apply (rule csp_Rep_int_choice0_right)
apply (simp add: image_iff)
apply (elim bexE)
by (simp)

lemma csp_Rep_int_choice2_rightE:
  "[| P <=SF !nat :X .. Qf ;
      ALL a:X. P <=SF Qf a ==> R
   |] ==> R"
apply (simp add: Rep_int_choice_OR2_def)
apply (rule csp_Rep_int_choice0_rightE)
apply (simp)
apply (erule csp_Rep_int_choice0_rightE)
apply (subgoal_tac "ALL a:X. P <=SF Qf a")
by (auto)


lemmas csp_Rep_int_choice_right 
     = csp_Rep_int_choice0_right
       csp_Rep_int_choice1_right
       csp_Rep_int_choice2_right

lemmas csp_Rep_int_choice_rightE
     = csp_Rep_int_choice0_rightE
       csp_Rep_int_choice1_rightE
       csp_Rep_int_choice2_rightE

(*-------------------------------------------------------*
 |             decomposition with subset                 |
 *-------------------------------------------------------*)

(*** Rep_int_choice ***)                                        (* unsafe *)

lemma domSF_Rep_int_choice_subset:
  "[|  Y <= X  ; !!a. a:Y ==> [[Qf a]]SF <= [[Pf a]]SF |]
                    ==> [[!! :Y .. Qf]]SF <= [[!! :X .. Pf]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
by (rule, simp add: Rep_int_choice_mem, force)+

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Rep_int_choice0_subset:
  "[|  Y <= X  ; !!a. a:Y ==> Pf a <=SF Qf a |]
                    ==> !! :X .. Pf <=SF !! :Y .. Qf"
by (simp add: refSF_def domSF_Rep_int_choice_subset)

lemma csp_Rep_int_choice1_subset:
  "[|  Y <= X  ; !!a. a:Y ==> Pf a <=SF Qf a |]
                    ==> ! :X .. Pf <=SF ! :Y .. Qf"
apply (simp add: Rep_int_choice_OR1_def)
apply (rule csp_Rep_int_choice0_subset)
by (auto)

lemma csp_Rep_int_choice2_subset:
  "[|  Y <= X  ; !!a. a:Y ==> Pf a <=SF Qf a |]
                    ==> !nat :X .. Pf <=SF !nat :Y .. Qf"
apply (simp add: Rep_int_choice_OR2_def)
apply (rule csp_Rep_int_choice0_subset)
by (auto)

lemmas csp_Rep_int_choice_subset 
     = csp_Rep_int_choice0_subset
       csp_Rep_int_choice1_subset
       csp_Rep_int_choice2_subset

(*** ! x:X .. and ? -> ***)

lemma domSF_Int_Ext_pre_choice_subset:
  "[| Y ~={} ; Y <= X ; !!a. a:Y ==> [[Qf a]]SF <= [[Pf a]]SF |]
        ==> [[? :Y -> Qf]]SF <= [[! x:X .. (x -> Pf x)]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
 apply (rule)
 apply (simp add: Rep_int_choice_mem Ext_pre_choice_mem Act_prefix_mem)
 apply (elim disjE conjE exE, simp)
 apply (rule disjI2)
 apply (rule_tac x="a" in bexI)
 apply (simp)
 apply (rule_tac x="s" in exI)
 apply (simp)
 apply (force)
 apply (fast)

 apply (rule)
 apply (simp add: Rep_int_choice_mem Ext_pre_choice_mem Act_prefix_mem)
 apply (elim disjE conjE exE, simp)
  apply (subgoal_tac "EX a. a:Y")
  apply (elim exE)
  apply (rule_tac x="a" in bexI)
  apply (case_tac "Ev a ~: Xa", simp)
  apply (fast)
  apply (fast)
  apply (fast)

 apply (rule_tac x="a" in bexI)
 apply (simp)
 apply (rule_tac x="sa" in exI)
 apply (simp)
 apply (force)
 apply (fast)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Int_Ext_pre_choice_subset:
  "[| Y ~={} ; Y <= X ; !!a. a:Y ==> Pf a <=SF Qf a |]
        ==> ! x:X .. (x -> Pf x) <=SF
            ? :Y -> Qf"
by (simp add: refSF_def domSF_Int_Ext_pre_choice_subset)

lemmas csp_decompo_subset = csp_Rep_int_choice_subset
                            csp_Int_Ext_pre_choice_subset

(*-------------------------------------------------------*
 |               decompose external choice               |
 *-------------------------------------------------------*)

lemma domSF_Ext_choice_right:
  "[| [[Q1]]SF <= [[P]]SF ; [[Q2]]SF <= [[P]]SF |]
      ==> [[Q1 [+] Q2]]SF <= [[P]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
 apply (rule)
 apply (simp add: Ext_choice_mem)
 apply (force)

 apply (rule)
 apply (simp add: Ext_choice_mem)
 apply (simp add: memF_IntF memF_UnF memT_UnT)
 apply (elim disjE conjE exE, simp)
 apply (force, force, force)

 apply (simp)
 apply (subgoal_tac "([]t, X) :f [[Q1]]F")
 apply (force)
 apply (rule domSF_F2_F4[of "[[Q1]]T" "[[Q1]]F"])
 apply (simp_all)

 apply (subgoal_tac "([]t, X) :f [[Q2]]F")
 apply (force)
 apply (rule domSF_F2_F4[of "[[Q2]]T" "[[Q2]]F"])
 apply (simp_all)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Ext_choice_right:
  "[| P <=SF Q1 ;
      P <=SF Q2 |]
      ==> P <=SF Q1 [+] Q2"
by (simp add: refSF_def domSF_Ext_choice_right)

(*-------------------------------------------------------*
 |              Int_choice_Ext_choice                    |
 *-------------------------------------------------------*)

lemma domSF_Int_choice_Ext_choice:
  "[[P1 [+] P2]]SF <= [[P1 |~| P2]]SF"
apply (simp add: proc_subsetSF_decompo)
apply (rule)
 apply (rule)
 apply (simp add: Ext_choice_mem Int_choice_mem)
 apply (rule)
 apply (simp add: Ext_choice_mem Int_choice_mem)
 apply (simp add: memF_IntF memF_UnF memT_UnT)
 apply (elim disjE conjE exE)
  apply (simp_all)

  apply (rule disjI1)
  apply (rule domSF_F2_F4)
  apply (simp_all)

  apply (rule disjI2)
  apply (rule domSF_F2_F4)
  apply (simp_all)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_Int_choice_Ext_choice:
  "P1 |~| P2 <=SF P1 [+] P2"
by (simp add: refSF_def domSF_Int_choice_Ext_choice)

lemma csp_Int_choice_left3:
  "P1 [+] P2 <=SF Q
   ==> P1 |~| P2 <=SF Q"
apply (rule csp_trans)
apply (rule csp_Int_choice_Ext_choice)
by (simp)

(****************** to add them again ******************)

declare not_None_eq    [simp]

end
