           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_proc = SKIP + STOP + DIV + Act_prefix 
                + Ext_pre_choice + Ext_choice + Int_choice
                + Rep_int_choice + Parallel + Hiding
                + Conditional + Renaming + Seq_compo
                + CSP_semantics_exp + Proc_name: 

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. ([[P]]Tx X, [[P]]Fx X) : domSF
         1. ([[P]]T , [[P]]F) : domSF
         2. decompose
         3. mono
         4. Timeout eval

 *****************************************************************)

(*********************************************************
          !!!  ([[P]]Tx X, [[P]]Fx X) : domSF   !!!
 *********************************************************)

lemma procx_domSF[simp]: "P : Procx ==> ([[P]]Tx X, [[P]]Fx X) : domSF"
apply (rule Procx.induct[of P "(%P. ([[P]]Tx X, [[P]]Fx X) : domSF)"])
apply (simp_all add: evalTx_simp evalFx_simp)
apply (simp add: Proc_name_evalSF)
apply (simp add: STOP_domSF)
apply (simp add: SKIP_domSF)
apply (simp add: DIV_domSF)

apply (simp add: Act_prefix_domSF)
apply (simp add: Ext_pre_choice_domSF)
apply (simp add: Ext_choice_domSF)
apply (simp add: Int_choice_domSF)
apply (simp add: Rep_int_choice_domSF)
apply (simp add: Conditional_domSF)
apply (simp add: Parallel_domSF)
apply (simp add: Hiding_domSF)
apply (simp add: Renaming_domSF)
apply (simp add: Seq_compo_domSF)
done

(*********************************************************
          !!!  ([[P]]T , [[P]]F) : domSF    !!!
 *********************************************************)

lemma proc_domSF[simp]: "([[P]]T , [[P]]F) : domSF"
apply (insert procx_domSF[of "%x. P"])
apply (simp add: Procx_def evalTx_evalT_evalFx_evalF 
            del: procx_domSF)
done

(*********************************************************
                   decompose of eval
 *********************************************************)

(*** T ***)

lemma fstSF_evalSFx[simp]: 
  "P : Procx ==> fstSF ([[P]]SFx x) = [[P]]Tx x"
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (simp add: fstSF_def)
apply (simp add: Abs_domSF_inverse)
done

lemma fstSF_evalSFX[simp]: 
  "P : ProcX ==> fstSF ([[P]]SFX x pn) = [[%X. P X pn]]Tx x"
apply (simp add: evalSFX_def)
apply (simp add: ProcX_def)
done

lemma fstSF_evalSF[simp]: "fstSF ([[P]]SF) = [[P]]T"
apply (simp add: evalSF_def)
apply (simp add: pairSF_def)
apply (simp add: fstSF_def)
apply (simp add: Abs_domSF_inverse)
done

(*** F ***)

lemma sndSF_evalSFx[simp]: 
  "P : Procx ==> sndSF ([[P]]SFx x) = [[P]]Fx x"
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (simp add: sndSF_def)
apply (simp add: Abs_domSF_inverse)
done

lemma sndSF_evalSFX[simp]: 
  "P : ProcX ==> sndSF ([[P]]SFX x pn) = [[%X. P X pn]]Fx x"
apply (simp add: evalSFX_def)
apply (simp add: ProcX_def)
done

lemma sndSF_evalSF[simp]: "sndSF ([[P]]SF) = [[P]]F"
apply (simp add: evalSF_def)
apply (simp add: pairSF_def)
apply (simp add: sndSF_def)
apply (simp add: Abs_domSF_inverse)
done

(*** relation ***)

lemma procx_eqSF_decompo:
  "[| P : Procx ; Q : Procx |]
   ==> ([[P]]SFx = [[Q]]SFx)
     = (([[P]]Tx = [[Q]]Tx) & ([[P]]Fx = [[Q]]Fx))"
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (simp add: expand_fun_eq)
apply (simp add: Abs_domSF_inject)
by (auto)

lemma proc_eqSF_decompo:
  "([[P]]SF = [[Q]]SF)
 = (([[P]]T = [[Q]]T) & ([[P]]F = [[Q]]F))"
apply (simp add: evalSF_def)
apply (simp add: pairSF_def)
apply (simp add: Abs_domSF_inject)
done

(* subset *)

lemma procx_subsetSF_decompo:
  "[| P : Procx ; Q : Procx |]
   ==> ([[P]]SFx <= [[Q]]SFx)
     = (([[P]]Tx <= [[Q]]Tx) & ([[P]]Fx <= [[Q]]Fx))"
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (simp add: order_prod_def)
apply (simp add: subsetSF_def)
apply (simp add: Abs_domSF_inverse)
apply (simp add: order_pair_def)
by (rule, simp_all)

lemma proc_subsetSF_decompo:
  "([[P]]SF <= [[Q]]SF)
 = (([[P]]T <= [[Q]]T) & ([[P]]F <= [[Q]]F))"
apply (simp add: evalSF_def)
apply (simp add: pairSF_def)
apply (simp add: subsetSF_def)
apply (simp add: Abs_domSF_inverse)
apply (simp add: order_pair_def)
done

(*********************************************************
               !!!  mono evalSFx  !!!
 *********************************************************)

lemma evalSFx_mono: 
  "[| P : Procx ; x <= y |] ==> [[P]]SFx x <= [[P]]SFx y"
apply (simp add: subsetSF_decompo)

apply (rule Procx.induct 
   [of P "(%P. ([[P]]Tx x <= [[P]]Tx y & [[P]]Fx x <= [[P]]Fx y))"])

apply (simp_all add: evalTx_simp evalFx_simp)
apply (simp add: Proc_name_evalT_mono)
apply (simp add: Proc_name_evalF_mono)

(* Act_prefix *)
apply (simp add: Act_prefix_evalT_mono)
apply (simp add: Act_prefix_evalF_mono)

(* Ext_pre_choice *)
apply (simp add: Ext_pre_choice_evalT_mono)
apply (simp add: Ext_pre_choice_evalF_mono)

(* Ext_choice *)
apply (simp add: Ext_choice_evalT_mono)
apply (simp add: Ext_choice_evalF_mono)

(* Int_choice *)
apply (simp add: Int_choice_evalT_mono)
apply (simp add: Int_choice_evalF_mono)

(* Rep_int_choice *)
apply (simp add: Rep_int_choice_evalT_mono)
apply (simp add: Rep_int_choice_evalF_mono)

(* Conditional *)
apply (simp add: Conditional_evalT_mono)
apply (simp add: Conditional_evalF_mono)

(* Parallel *)
apply (simp add: Parallel_evalT_mono)
apply (simp add: Parallel_evalF_mono)

(* Hiding *)
apply (simp add: Hiding_evalT_mono)
apply (simp add: Hiding_evalF_mono)

(* Renaming *)
apply (simp add: Renaming_evalT_mono)
apply (simp add: Renaming_evalF_mono)

(* Seq_compo *)
apply (simp add: Seq_compo_evalT_mono)
apply (simp add: Seq_compo_evalF_mono)
done

lemma evalSFx_mono_fun[simp]: "P : Procx ==> mono [[P]]SFx"
by (simp add: mono_def evalSFx_mono)

(*********************************************************
               !!!  mono evalTx mono  !!!
 *********************************************************)

lemma evalTx_mono: 
  "[| P : Procx ; x <= y |] ==> [[P]]Tx x <= [[P]]Tx y"
apply (subgoal_tac "[[P]]SFx x <= [[P]]SFx y")
apply (simp add: subsetSF_decompo)
apply (simp add: evalSFx_mono)
done

lemma evalTx_mono_fun[simp]: "P : Procx ==> mono [[P]]Tx"
by (simp add: mono_def evalTx_mono)

(*********************************************************
               !!!   mono [[P]]F  !!!
 *********************************************************)

lemma evalFx_mono: 
  "[| P : Procx ; x <= y |] ==> [[P]]Fx x <= [[P]]Fx y"
apply (subgoal_tac "[[P]]SFx x <= [[P]]SFx y")
apply (simp add: subsetSF_decompo)
apply (simp add: evalSFx_mono)
done

lemma evalFx_mono_fun[simp]: "P : Procx ==> mono [[P]]Fx"
by (simp add: mono_def evalFx_mono)

(*********************************************************
               !!!   mono [[P]]SFX  !!!
 *********************************************************)

lemma evalSFX_mono: 
  "[| P : ProcX ; x <= y |] ==> [[P]]SFX x <= [[P]]SFX y"
apply (simp add: evalSFX_def)
apply (simp add: ProcX_def)
apply (simp (no_asm) add: order_prod_def)
apply (rule allI)
apply (drule_tac x="xa" in spec)
apply (simp add: evalSFx_mono)
done

lemma evalSFX_mono_fun: "P : ProcX ==> mono [[P]]SFX"
by (simp add: evalSFX_mono mono_def)

(*********************************************************
              Healthiness conditions
 *********************************************************)

lemma proc_T2_T3:
  "(s @t [Tick]t, X) :f [[P]]F ==> (s @t [Tick]t, Y) :f [[P]]F"
apply (insert domSF_T2[of "[[P]]T" "[[P]]F" "s @t [Tick]t" X])
apply (simp)
apply (insert domSF_T3[of "[[P]]T" "[[P]]F" "s" Y])
apply (simp)
apply (case_tac "~ notick s", simp_all)
done

lemma proc_F2:
  "[| (s, X) :f [[P]]F ; Y <= X |] ==> (s, Y) :f [[P]]F"
apply (insert memF_F2[of s X "[[P]]F" Y])
by (simp)

(*********************************************************
                     Timeout trace eval
 *********************************************************)

lemma Timeout_memT:
  "(t :t [[P [> Q]]T) = (t :t [[P]]T | t :t [[Q]]T)"
by (simp add: evalT_def evalt_def memT_UnT)

(*********************************************************
                     Timeout failure eval
 *********************************************************)

lemma Timeout_memF:
  "f :f [[P [> Q]]F 
   = (f :f [[Q]]F | 
     (EX s X. f = (s,X) & s ~= []t & (s,X) :f [[P]]F) |
     (EX   X. f = ([]t,X) & X <= Evset & [Tick]t :t [[P]]T))"
apply (simp add: Ext_choice_memF)
apply (simp add: memF_UnF memF_IntF)
apply (simp add: Int_choice_memF)
apply (simp add: memT_UnT)
apply (simp add: Int_choice_memT)
apply (simp add: STOP_memF STOP_memT)
apply (rule iffI)

 (* => *)
 apply (elim conjE exE disjE, simp_all)
 apply (rule disjI1)
 apply (rule domSF_F2_F4, simp_all)

 (* <= *)
 apply (elim conjE exE disjE, simp_all)
 apply (auto elim: memF_pairE)
done

lemmas Timeout_mem = Timeout_memT Timeout_memF

(*---------------------------------------------------*
 |          Procx, ProcX and composition             |
 *---------------------------------------------------*)

(*** SF ***)

lemma evalSF_Procx_compo: 
  "P : Procx ==> [[P x]]SF = [[P]]SFx (%pn. [[x pn]]SF)"
apply (simp add: eqSF_decompo)
apply (rule Procx.induct 
   [of P "(%P. ([[P x]]T = [[P]]Tx (%pn. [[x pn]]SF) &
                [[P x]]F = [[P]]Fx (%pn. [[x pn]]SF)))"])
apply (simp_all add: evalTx_simp evalFx_simp)
apply (simp_all add: evalT_def evalF_def)
done

lemma evalSF_ProcX_compo:
  "P : ProcX ==> [[P x pn]]SF= [[%x. P x pn]]SFx (%pn. [[x pn]]SF)"
apply (simp add: ProcX_def)
apply (drule_tac x="pn" in spec)
apply (insert evalSF_Procx_compo[of "%x. P x pn" x])
by (simp)

lemma evalSF_ProcX_compo_fun: 
  "P : ProcX ==> (%pn. [[P x pn]]SF) = [[P]]SFX (%pn. [[x pn]]SF)"
apply (simp add: evalSFX_def)
apply (simp add: expand_fun_eq)
apply (simp add: evalSF_ProcX_compo)
done

lemma evalSF_ProcX_compo_pn: 
  "P : ProcX ==> [[P x pn]]SF = [[P]]SFX (%pn. [[x pn]]SF) pn"
apply (simp add: evalSFX_def)
apply (simp add: evalSF_ProcX_compo)
done

(*** T ***)

lemma evalT_Procx_compo:
  "P : Procx ==> [[P x]]T = [[P]]Tx (%pn. [[x pn]]SF)"
apply (insert evalSF_Procx_compo[of P x])
apply (simp add: eqSF_decompo)
done

lemma evalT_ProcX_compo:
  "P : ProcX ==> [[P x pn]]T = [[%x. P x pn]]Tx (%pn. [[x pn]]SF)"
apply (simp add: ProcX_def)
apply (drule_tac x="pn" in spec)
apply (insert evalT_Procx_compo[of "%x. P x pn" x])
by (simp)

(*** F ***)

lemma evalF_Procx_compo:
  "P : Procx ==> [[P x]]F = [[P]]Fx (%pn. [[x pn]]SF)"
apply (insert evalSF_Procx_compo[of P x])
apply (simp add: eqSF_decompo)
done

lemma evalF_ProcX_compo:
  "P : ProcX ==> [[P x pn]]F = [[%x. P x pn]]Fx (%pn. [[x pn]]SF)"
apply (simp add: ProcX_def)
apply (drule_tac x="pn" in spec)
apply (insert evalF_Procx_compo[of "%x. P x pn" x])
by (simp)

(*** composition of process expressions ***)

lemma evalSFx_Procx_compo_ProcX:
   "[| P : Procx ; Q : ProcX |] ==>
    [[P o Q]]SFx x = [[P]]SFx ([[Q]]SFX x)"
apply (subgoal_tac "ALL P. P : Procx --> P o Q : Procx")
apply (simp add: ProcX_def)
apply (simp add: eqSF_decompo)
apply (rule Procx.induct[of P 
  "(%P. [[P o Q]]Tx x = [[P]]Tx ([[Q]]SFX x) &
        [[P o Q]]Fx x = [[P]]Fx ([[Q]]SFX x))"])
apply (simp_all add: comp_def evalTx_simp evalFx_simp)
apply (simp add: evalSFX_def)
apply (simp add: compo_in_Procx)
done

lemmas evalSFx_Procx_compo_ProcX_comp =
       evalSFx_Procx_compo_ProcX[simplified comp_def]

lemma evalSFx_Procx_compo_ProcX_ALL:
   "ALL P Q x. (P : Procx & Q : ProcX) -->
    [[P o Q]]SFx x = [[P]]SFx ([[Q]]SFX x)"
by (simp add: evalSFx_Procx_compo_ProcX)

lemma evalSFX_ProcX_compo_ProcX:
   "[| P : ProcX ; Q : ProcX |] ==>
    [[P o Q]]SFX x = [[P]]SFX ([[Q]]SFX x)"
apply (simp add: evalSFX_def)
apply (simp add: expand_fun_eq)
apply (rule allI)
apply (simp add: ProcX_def)
apply (drule_tac x="xa" in spec)
apply (insert evalSFx_Procx_compo_ProcX_ALL)
apply (drule_tac x="%X. P X xa" in spec)
apply (drule_tac x="Q" in spec)
apply (drule_tac x="x" in spec)
apply (simp add: ProcX_def)
apply (simp add: comp_def evalSFX_def)
done

lemmas evalSFX_ProcX_compo_ProcX_comp =
       evalSFX_ProcX_compo_ProcX[simplified comp_def]

lemma evalSFX_interation_ProcX:
   "P : ProcX ==> [[P ^ n]]SFX x = ([[P]]SFX ^ n) x"
apply (induct_tac n)
apply (simp add: evalSFX_def)
apply (simp add: iteration_in_ProcX evalSFX_ProcX_compo_ProcX_comp)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]
declare not_None_eq    [simp]

end
