           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_proc_cms = CSP_proc           + Constant_cms
                    + Act_prefix_cms     + Ext_choice_cms
                    + Ext_pre_choice_cms + Int_choice_cms
                    + Rep_int_choice_cms + Conditional_cms
                    + Parallel_cms       + Renaming_cms
                    + Seq_compo_cms      + Proc_name_cms :

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

   (* note: split_paired_All: (ALL x. ?P x) = (ALL a b. ?P (a, b)) *)
   (*       split_paired_Ex : (EX x. ?P x) = (EX a b. ?P (a, b))   *)

declare split_paired_All [simp del]
declare split_paired_Ex  [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. proc : nohide --> non-expanding
         2. proc : guarded & nohide --> contraction
         3. 
         4. 

 *****************************************************************)

(*********************************************************
             decompose subset of proc_rest
 *********************************************************)

lemma Procx_rest_subsetSF_decompo:
   "[| P : Procx ; Q : Procx |] ==>
     ([[P]]SFx x1 rest n <= [[Q]]SFx x2 rest n)
   = ([[P]]Tx x1 rest n <= [[Q]]Tx x2 rest n &
      [[P]]Fx x1 rest n <= [[Q]]Fx x2 rest n)"
apply (simp add: restSF_def)
apply (simp add: subsetSF_def)
apply (simp add: Abs_domSF_inverse)

apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (simp add: restSFpair_def)
apply (simp add: Abs_domSF_inverse)
apply (simp add: pair_restriction_def)
apply (simp add: order_pair_def)
done

lemma Procx_rest_subsetT_decompo:
   "[| P : Procx ; Q : Procx ;  ([[P]]SFx x1 rest n <= [[Q]]SFx x2 rest n) |]
    ==> [[P]]Tx x1 rest n <= [[Q]]Tx x2 rest n"
by (simp add: Procx_rest_subsetSF_decompo)

lemma Procx_rest_subsetF_decompo:
   "[| P : Procx ; Q : Procx ; ([[P]]SFx x1 rest n <= [[Q]]SFx x2 rest n) |]
    ==> [[P]]Fx x1 rest n <= [[Q]]Fx x2 rest n"
by (simp add: Procx_rest_subsetSF_decompo)

(* eq *)

lemma Procx_rest_eqSF_decompo:
   "[| P : Procx ; Q : Procx |] ==>
     ([[P]]SFx x1 rest n = [[Q]]SFx x2 rest n)
   = ([[P]]Tx x1 rest n = [[Q]]Tx x2 rest n &
      [[P]]Fx x1 rest n = [[Q]]Fx x2 rest n)"
apply (simp add: eq_iff)
apply (simp add: Procx_rest_subsetSF_decompo)
apply (rule iffI)
by (simp_all)

lemma Procx_rest_eqT_decompo:
   "[| P : Procx ; Q : Procx ; [[P]]SFx x1 rest n = [[Q]]SFx x2 rest n |]
   ==> [[P]]Tx x1 rest n = [[Q]]Tx x2 rest n"
by (simp add: Procx_rest_eqSF_decompo)

lemma Procx_rest_eqF_decompo:
   "[| P : Procx ; Q : Procx ; [[P]]SFx x1 rest n = [[Q]]SFx x2 rest n |]
   ==> [[P]]Fx x1 rest n = [[Q]]Fx x2 rest n"
by (simp add: Procx_rest_eqSF_decompo)

(*********************************************************
              map_alpha decomposition
 *********************************************************)

(* Abs_domSF *)

lemma map_alpha_Abs_domSFx:
  "P : Procx ==>
   map_alpha [[P]]SFx alpha = map_alpha ([[P]]Tx ** [[P]]Fx) alpha"
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (simp add: pair_fun_def)
apply (simp add: distance_Abs_domSF)
done

(* decompo *)

lemma map_alpha_evalSFx_decompo:
  "P : Procx ==>
   map_alpha [[P]]SFx alpha = 
   (map_alpha [[P]]Tx alpha & map_alpha [[P]]Fx alpha)"
apply (simp add: map_alpha_Abs_domSFx)
apply (simp add: pair_map_alpha_compo)
done

(* non_expanding *)

lemma non_expanding_evalSFx_decompo:
  "P : Procx ==>
   non_expanding [[P]]SFx =
   (non_expanding [[P]]Tx & non_expanding [[P]]Fx)"
apply (simp add: non_expanding_def)
by (simp add: map_alpha_evalSFx_decompo)

(* contraction_alpha *)

lemma contraction_alpha_evalSFx_decompo:
  "P : Procx ==>
    contraction_alpha [[P]]SFx alpha = 
   (contraction_alpha [[P]]Tx alpha & contraction_alpha [[P]]Fx alpha)"
apply (simp add: contraction_alpha_def)
by (auto simp add: map_alpha_evalSFx_decompo)

(*********************************************************
             !!!   proc non-expanding   !!!
 *********************************************************)

lemma evalSFx_non_expanding_lm:
  "P : Procx ==> nohide (P (%pn. DIV)) --> non_expanding [[P]]SFx"
apply (simp add: non_expanding_evalSFx_decompo)
apply (rule Procx.induct
       [of P "(%P. (nohide (P (%pn. DIV))
              --> non_expanding [[P]]Tx & non_expanding [[P]]Fx))"])
apply (simp)
apply (simp add: Procx_def non_expanding_evalSFx_decompo[THEN sym])
apply (simp add: Proc_name_evalSF_non_expanding)
apply (simp_all add: evalTx_simp evalFx_simp nohideX_def)

apply (simp add: Constant_non_expanding)
apply (simp add: Constant_non_expanding)
apply (simp add: Constant_non_expanding)
apply (simp add: Act_prefix_evalT_non_expanding 
                 Act_prefix_evalF_non_expanding)
apply (simp add: Ext_pre_choice_evalT_non_expanding
                 Ext_pre_choice_evalF_non_expanding)
apply (simp add: Ext_choice_evalT_non_expanding
                 Ext_choice_evalF_non_expanding)
apply (simp add: Int_choice_evalT_non_expanding
                 Int_choice_evalF_non_expanding)
apply (simp add: Rep_int_choice_evalT_non_expanding
                 Rep_int_choice_evalF_non_expanding)
apply (simp add: Conditional_evalT_non_expanding
                 Conditional_evalF_non_expanding)
apply (simp add: Parallel_evalT_non_expanding
                 Parallel_evalF_non_expanding)
apply (simp add: Renaming_evalT_non_expanding
                 Renaming_evalF_non_expanding)
apply (simp add: Seq_compo_evalT_non_expanding
                 Seq_compo_evalF_non_expanding)
done

lemma evalSFx_non_expanding:
  "[| P : Procx ; nohide (P (%pn. DIV)) |] ==> non_expanding [[P]]SFx"
by (simp add: evalSFx_non_expanding_lm)

lemma evalTx_and_evalFx_non_expanding:
  "[| P : Procx ; nohide (P (%pn. DIV)) |]
   ==> (non_expanding [[P]]Tx & non_expanding [[P]]Fx)"
apply (simp add: non_expanding_evalSFx_decompo[THEN sym])
by (simp add: evalSFx_non_expanding)

lemma evalTx_non_expanding:
  "[| P : Procx ; nohide (P (%pn. DIV)) |]
   ==> non_expanding [[P]]Tx"
by (simp add: evalTx_and_evalFx_non_expanding)

lemma evalFx_non_expanding:
  "[| P : Procx ; nohide (P (%pn. DIV)) |]
   ==> non_expanding [[P]]Fx"
by (simp add: evalTx_and_evalFx_non_expanding)

(*********************************************************
            !!!   proc contraction map   !!!
 *********************************************************)

lemma evalSFx_contraction_half_lm:
  "P : Procx ==> (nohide (P (%pn. DIV)) & guard (P (%pn. DIV)))
                  --> contraction_alpha [[P]]SFx (1/2)"
apply (simp add: contraction_alpha_evalSFx_decompo)
apply (rule Procx.induct
       [of P "(%P. (nohide (P (%pm. DIV)) & guard (P (%pm. DIV)) -->
        contraction_alpha [[P]]Tx (1 / 2) & contraction_alpha [[P]]Fx (1 / 2)))"])
apply (simp_all add: evalTx_simp evalFx_simp)

apply (simp add: Constant_contraction_alpha)
apply (simp add: Constant_contraction_alpha)

apply (intro impI conjI)
apply (rule Act_prefix_evalT_contraction_half)
apply (simp add: evalTx_non_expanding)
apply (rule Act_prefix_evalF_contraction_half)
apply (simp add: evalFx_non_expanding)

apply (intro impI conjI)
apply (rule Ext_pre_choice_evalT_contraction_half)
apply (rule allI)
apply (simp add: evalTx_non_expanding)
apply (rule Ext_pre_choice_evalF_contraction_half)
apply (rule allI)
apply (simp add: evalFx_non_expanding)

apply (simp add: Ext_choice_evalT_contraction_alpha
                 Ext_choice_evalF_contraction_alpha)
apply (simp add: Int_choice_evalT_contraction_alpha
                 Int_choice_evalF_contraction_alpha)
apply (simp add: Rep_int_choice_evalT_contraction_alpha
                 Rep_int_choice_evalF_contraction_alpha)
apply (simp add: Conditional_evalT_contraction_alpha
                 Conditional_evalF_contraction_alpha)
apply (simp add: Parallel_evalT_contraction_alpha
                 Parallel_evalF_contraction_alpha)
apply (simp add: Renaming_evalT_contraction_alpha
                 Renaming_evalF_contraction_alpha)
apply (simp add: Seq_compo_evalT_contraction_alpha
                 Seq_compo_evalF_contraction_alpha)
done

lemma evalSFx_contraction_half:
  "[| P : Procx ; nohide (P (%pn. DIV)) ; guard (P (%pn. DIV)) |]
 ==> contraction_alpha [[P]]SFx (1/2)"
by (simp add: evalSFx_contraction_half_lm)

lemma evalTx_and_evalFx_contraction_half:
  "[| P : Procx ; nohide (P (%pn. DIV)) ; guard (P (%pn. DIV)) |]
   ==> (contraction_alpha [[P]]Tx (1/2) & contraction_alpha [[P]]Fx (1/2))"
apply (simp add: contraction_alpha_evalSFx_decompo[THEN sym])
by (simp add: evalSFx_contraction_half)

lemma evalTx_contraction_half:
  "[| P : Procx ; nohide (P (%pn. DIV)) ; guard (P (%pn. DIV)) |]
   ==> contraction_alpha [[P]]Tx (1/2)"
by (simp add: evalTx_and_evalFx_contraction_half)

lemma evalFx_contraction_half:
  "[| P : Procx ; nohide (P (%pn. DIV)) ; guard (P (%pn. DIV)) |]
   ==> contraction_alpha [[P]]Fx (1/2)"
by (simp add: evalTx_and_evalFx_contraction_half)

(*------------------------------------------------------*
 |             Banach theorem for procDF                |
 *------------------------------------------------------*)

(************************
   non-expanding evalSFX
 ************************)

lemma evalSFX_non_expanding:
  "[| P : ProcX ; nohideX P |] ==> non_expanding [[P]]SFX"
apply (simp add: evalSFX_def)
apply (simp add: ProcX_def)
apply (simp add: nohideX_def)
apply (simp add: prod_non_expand)
apply (simp add: proj_fun_def)
apply (simp add: comp_def)
apply (rule allI)
apply (drule_tac x="i" in spec)
apply (drule_tac x="i" in spec)
apply (simp add: evalSFx_non_expanding)
done

(***************************
   contraction_half evalSFX
 ***************************)

lemma evalSFX_contraction_half:
  "[| P : ProcX ; nohideX P ; guardX P |]
    ==> contraction_alpha [[P]]SFX (1/2)"
apply (simp add: evalSFX_def)
apply (simp add: ProcX_def)
apply (simp add: nohideX_def)
apply (simp add: guardX_def)
apply (simp add: prod_contra_alpha)
apply (simp add: proj_fun_def)
apply (simp add: comp_def)
apply (rule allI)
apply (drule_tac x="i" in spec)
apply (drule_tac x="i" in spec)
apply (simp add: evalSFx_contraction_half)
done

(***************************
     contraction evalSFX
 ***************************)

lemma evalSFX_contraction:
  "[| P : ProcX ; nohideX P ; guardX P |]
    ==> contraction [[P]]SFX"
apply (simp add: contraction_def)
apply (rule_tac x="1/2" in exI)
apply (simp add: evalSFX_contraction_half)
done

(************************
      Banach evalSFX
 ************************)

lemma evalSFX_Banach_thm:
   "[| P : ProcX ; nohideX P ; guardX P |]
     ==> [[P]]SFX hasUFP & (%n. ([[P]]SFX^n) x0) convergeTo UFP [[P]]SFX"
apply (rule Banach_thm)
apply (simp add: evalSFX_contraction)
done

(************************
     constructive_rs
 ************************)

lemma evalSFX_constructive_rs:
  "[| P : ProcX ; nohideX P ; guardX P |] ==> constructive_rs [[P]]SFX"
by (simp add: contra_alpha_to_contst evalSFX_contraction_half)

(****************** to add them again ******************)

declare Union_image_eq   [simp]
declare Inter_image_eq   [simp]

declare split_paired_All [simp]
declare split_paired_Ex  [simp]

declare not_None_eq      [simp]

end
