           (*-------------------------------------------*
            |                CSP-Prover                 |
            |              December 2004                |
            |              February 2005  (modified)    |
            |              June     2005  (modified)    |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_proc_cpo = CSP_proc           + Constant_cpo
                    + Act_prefix_cpo     + Ext_choice_cpo
                    + Ext_pre_choice_cpo + Int_choice_cpo
                    + Rep_int_choice_cpo + Conditional_cpo
                    + Parallel_cpo       + Hiding_cpo
                    + Renaming_cpo       + Seq_compo_cpo
                    + Proc_name_cpo :

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*****************************************************************

         1. [[P]]SF : continuous
         2.
         3. 
         4. 

 *****************************************************************)

(*********************************************************
              continuous decomposition
 *********************************************************)

(* evalSFx --> evalTx and evalFx *)

lemma continuous_evalSFx_decompo_lm_only_if:
 "[| P : Procx ; continuous [[P]]SFx |]
  ==> continuous (%x. ([[P]]Tx x, [[P]]Fx x))"
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (insert compo_continuous
       [of "(%x. Abs_domSF ([[P]]Tx x, [[P]]Fx x))" "Rep_domSF" ])
apply (simp add: comp_def)
apply (simp add: cont_Rep_domSF)
apply (simp add: Abs_domSF_inverse)
done

lemma continuous_evalSFx_decompo_lm_if:
  "[| P : Procx ; continuous (%x. ([[P]]Tx x, [[P]]Fx x)) |]
   ==> continuous [[P]]SFx"
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (insert continuous_Abs_domSF[of "(%x. ([[P]]Tx x, [[P]]Fx x))"])
by (simp add: comp_def)

(* iff *)

lemma continuous_evalSFx_decompo_lm:
  "P : Procx ==> continuous [[P]]SFx =  continuous (%x. ([[P]]Tx x, [[P]]Fx x))"
apply (rule iffI)
apply (simp add: continuous_evalSFx_decompo_lm_only_if)
apply (simp add: continuous_evalSFx_decompo_lm_if)
done

(* decompo *)

lemma continuous_evalSFx_decompo:
  "P : Procx ==> continuous [[P]]SFx = (continuous [[P]]Tx & continuous [[P]]Fx)"
apply (simp add: continuous_evalSFx_decompo_lm)
apply (simp add: pair_continuous)
apply (subgoal_tac "(fst o (%x. ([[P]]Tx x, [[P]]Fx x))) = [[P]]Tx")
apply (subgoal_tac "(snd o (%x. ([[P]]Tx x, [[P]]Fx x))) = [[P]]Fx")
apply (simp)
apply (simp_all add: expand_fun_eq)
done

(*** lemmas for pcocess names ***)

lemma Proc_name_evalT_continuous: "continuous (%x. fstSF (x pn))"
apply (subgoal_tac "continuous [[%x. x pn]]SFx")
apply (simp only: Procx_def continuous_evalSFx_decompo)
apply (simp add: evalTx_simp, fast)
apply (simp add: Proc_name_evalSF_continuous)
done

lemma Proc_name_evalF_continuous: "continuous (%x. sndSF (x pn))"
apply (subgoal_tac "continuous [[%x. x pn]]SFx")
apply (simp only: Procx_def continuous_evalSFx_decompo)
apply (simp add: evalFx_simp, fast)
apply (simp add: Proc_name_evalSF_continuous)
done

(*****************************************************************
                !!!   [[P]]SFx continuous   !!!
 *****************************************************************)

lemma evalSFx_continuous: "P : Procx ==> continuous [[P]]SFx"
apply (simp add: continuous_evalSFx_decompo)
apply (rule Procx.induct[of P "(%P. continuous [[P]]Tx & continuous [[P]]Fx)"])
apply (simp_all add: evalTx_simp evalFx_simp)

apply (simp add: Proc_name_evalT_continuous
                 Proc_name_evalF_continuous)
apply (simp add: Constant_continuous)
apply (simp add: Constant_continuous)
apply (simp add: Constant_continuous)
apply (simp add: Act_prefix_evalT_continuous
                 Act_prefix_evalF_continuous)
apply (simp add: Ext_pre_choice_evalT_continuous
                 Ext_pre_choice_evalF_continuous)
apply (simp add: Ext_choice_evalT_continuous
                 Ext_choice_evalF_continuous)
apply (simp add: Int_choice_evalT_continuous
                 Int_choice_evalF_continuous)
apply (simp add: Rep_int_choice_evalT_continuous
                 Rep_int_choice_evalF_continuous)
apply (simp add: Conditional_evalT_continuous
                 Conditional_evalF_continuous)
apply (simp add: Parallel_evalT_continuous
                 Parallel_evalF_continuous)
apply (simp add: Hiding_evalT_continuous
                 Hiding_evalF_continuous)
apply (simp add: Renaming_evalT_continuous
                 Renaming_evalF_continuous)
apply (simp add: Seq_compo_evalT_continuous
                 Seq_compo_evalF_continuous)
done

(*****************************************************************
               !!!   [[P]]SFX continuous   !!!
 *****************************************************************)

lemma evalSFX_continuous: "P : ProcX ==> continuous [[P]]SFX"
apply (simp add: prod_continuous)
apply (simp add: evalSFX_def)
apply (simp add: proj_fun_def)
apply (intro allI)
apply (simp add: ProcX_def)
apply (rename_tac pn)
apply (drule_tac x="pn" in spec)
apply (subgoal_tac
  "(%x. x pn) o (%X pn. [[%X. P X pn]]SFx X) = [[%X. P X pn]]SFx")
apply (simp add: evalSFx_continuous)
apply (simp add: expand_fun_eq)
done

(*****************************************************************
                      Tarski on evalSFX
 *****************************************************************)

lemma evalSFX_Tarski_thm:
   "P : ProcX ==>
   [[P]]SFX hasLFP & LFP [[P]]SFX isLUB {x. EX n. x = ([[P]]SFX^n) Bot}"
apply (rule Tarski_thm)
apply (simp add: evalSFX_continuous)
done

lemma evalSFX_Tarski_thm_LUB:
   "P : ProcX ==> LFP [[P]]SFX = LUB {x. EX n. x = ([[P]]SFX^n) Bot}"
apply (rule sym)
apply (rule isLUB_LUB)
apply (simp add: evalSFX_Tarski_thm)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
