           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_semantics = CSP_syntax + Domain_SF_prod
                     + Trace_par + Trace_hide + Trace_ren + Trace_seq:

(*****************************************************************

         1. semantic clause
         2. 
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                 semantic clause (parts)
 *********************************************************)

(*** T ***)

consts 
  STOPt :: "'a domT"
  SKIPt :: "'a domT"
  DIVt  :: "'a domT"
  Act_prefixt  :: "'a => 'a domT => 'a domT"   ("(1_ /->t _)"    [150,80] 80)
  Ext_pre_choicet :: "'a set => ('a => 'a domT) => 'a domT"
                                               ("(1? :_ /->t _)" [90,80] 80)
  Ext_choicet     :: "'a domT => 'a domT => 'a domT"
                                               ("(1_ /[+]t _)"   [72,73] 72)
  Int_choicet     :: "'a domT => 'a domT => 'a domT"
                                               ("(1_ /|~|t _)"   [64,65] 64)
  Rep_int_choicet :: "(('a,nat) or_type) set => ((('a,nat) or_type) => 'a domT) => 'a domT"
                                               ("(1!! :_ ..t /_)" [90,68] 68) 
  Conditionalt    :: "bool => 'a domT => 'a domT => 'a domT"
                                 ("(0IFt _ /THENt _ /ELSEt _)"   [900,60,60] 88)
  Parallelt       :: "'a domT => 'a set => 'a domT => 'a domT"
                                               ("(1_ /|[_]|t _)" [76,0,77] 76)
  Hidingt         :: "'a domT => 'a set => 'a domT" 
                                               ("(1_ /--t _)"    [84,85] 84)
  Renamingt       :: "'a domT => ('a * 'a) set => 'a domT"
                                               ("(1_ /[[_]]t)"   [84,0] 84)
  Seq_compot      :: "'a domT => 'a domT => 'a domT"  
                                               ("(1_ /;;t _)"    [79,78] 78)

defs
 STOPt_def : "STOPt == {[]t}t"
 SKIPt_def : "SKIPt == {[]t, [Tick]t}t"
 DIVt_def  : "DIVt  == {[]t}t"
 Act_prefixt_def : 
   "a ->t T     == {t. t = []t | (EX s. t = [Ev a]t @t s & s :t T) }t"
 Ext_pre_choicet_def : 
   "? :X ->t Tf == {t. t = []t | (EX a s. t = [Ev a]t @t s & s :t Tf a & a : X) }t"
 Ext_choicet_def :
   "S [+]t T    == S UnT T"
 Int_choicet_def :
   "S |~|t T    == S UnT T"
 Rep_int_choicet_def :
   "!! :X ..t Tf == {t. t = []t | (EX a:X. t :t Tf a) }t"
 Conditionalt_def :
   "IFt b THENt S ELSEt T == (if b then S else T)"
 Parallelt_def :
   "S |[X]|t T  == {u. EX s t. u : s |[X]|tr t & s :t S & t :t T }t"
 Hidingt_def :
   "T --t X     == {t. EX s. t = s --tr X & s :t T }t"
 Renamingt_def :
   "T [[r]]t    == {t. EX s. s [[r]]tr t & s :t T }t"
 Seq_compot_def :
  "S ;;t T      == {u. (EX s. u = rmtick s & s :t S ) |
                       (EX s t. u = s @t t & s @t [Tick]t :t S & t :t T ) }t"

lemmas evalt_def
  = STOPt_def SKIPt_def DIVt_def Act_prefixt_def Ext_pre_choicet_def
    Ext_choicet_def Int_choicet_def Rep_int_choicet_def Conditionalt_def
    Parallelt_def Hidingt_def Renamingt_def Seq_compot_def

(*** F ***)

consts 
  STOPf :: "'a domF"
  SKIPf :: "'a domF"
  DIVf  :: "'a domF"
  Act_prefixf  :: "'a => 'a domF => 'a domF"   ("(1_ /->f _)"    [150,80] 80)
  Ext_pre_choicef :: "'a set => ('a => 'a domF) => 'a domF"
                                               ("(1? :_ /->f _)" [90,80] 80)
  Ext_choicef     :: "('a domT * 'a domF) => ('a domT * 'a domF) => 'a domF"
                                               ("(1_ /[+]f _)"   [72,73] 72)
  Int_choicef     :: "'a domF => 'a domF => 'a domF"
                                               ("(1_ /|~|f _)"   [64,65] 64)
  Rep_int_choicef :: "(('a,nat) or_type) set => ((('a,nat) or_type) => 'a domF) => 'a domF"
                                               ("(1!! :_ ..f /_)" [90,68] 68) 
  Conditionalf    :: "bool => 'a domF => 'a domF => 'a domF"
                                 ("(0IFf _ /THENf _ /ELSEf _)"   [900,60,60] 88)
  Parallelf       :: "'a domF => 'a set => 'a domF => 'a domF"
                                               ("(1_ /|[_]|f _)" [76,0,77] 76)
  Hidingf         :: "'a domF => 'a set => 'a domF" 
                                               ("(1_ /--f _)"    [84,85] 84)
  Renamingf       :: "'a domF => ('a * 'a) set => 'a domF"
                                               ("(1_ /[[_]]f)"   [84,0] 84)
  Seq_compof      :: "('a domT * 'a domF) => 'a domF => 'a domF"  
                                               ("(1_ /;;f _)"    [79,78] 78)

defs
 STOPf_def : "STOPf == {f. EX X. f = ([]t, X) }f"
 SKIPf_def : "SKIPf == {f. (EX X. f = ([]t, X) & X <= Evset) |
                           (EX X. f = ([Tick]t, X)) }f"
 DIVf_def  : "DIVf  == {}f"
 Act_prefixf_def : 
   "a ->f F     == {f. (EX X.   f = ([]t,X) & Ev a ~: X) |
                       (EX s X. f = ([Ev a]t @t s, X) & (s,X) :f F ) }f"
 Ext_pre_choicef_def : 
   "? :X ->f Ff == {f. (EX Y.     f = ([]t,Y) & Ev`X Int Y = {}) |
                       (EX a s Y. f = ([Ev a]t @t s, Y) & 
                                   (s,Y) :f Ff a & a : X) }f"
 Ext_choicef_def :
   "SE [+]f TF  == {f. (EX   X. f = ([]t,X) & f :f snd SE IntF snd TF) |
                       (EX s X. f = (s,X)   & f :f snd SE UnF snd TF & 
                                  s ~= []t) |
                       (EX   X. f = ([]t,X) & [Tick]t :t fst SE UnT fst TF &
                                  X <= Evset) }f"
 Int_choicef_def :
   "E |~|f F    == E UnF F"
 Rep_int_choicef_def :
   "!! :X ..f Ff == {f. EX a:X. f :f Ff a}f "
 Conditionalf_def :
   "IFf b THENf E ELSEf F == (if b then E else F)"
 Parallelf_def :
   "E |[X]|f F  == {f. EX u Y Z. f = (u, Y Un Z) &
                         Y-((Ev`X) Un {Tick})= Z-((Ev`X) Un {Tick}) &
                        (EX s t. u : s |[X]|tr t & (s,Y) :f E & (t,Z) :f F) }f"
 Hidingf_def :
   "F --f X     == {f. EX s Y. f = (s --tr X, Y) & (s,(Ev`X) Un Y) :f F }f"
 Renamingf_def :
   "F [[r]]f    == {f. EX s t X. f = (t,X) & s [[r]]tr t & (s, [[r]]inv X) :f F }f"
 Seq_compof_def :
  "SE ;;f F      == {f. (EX t X. f = (t,X) & (t, X Un {Tick}) :f snd SE &
                                  notick t) |
                        (EX s t X. f = (s @t t,X) & s @t [Tick]t :t fst SE & 
                                   (t, X) :f F ) }f"
lemmas evalf_def
  = STOPf_def SKIPf_def DIVf_def Act_prefixf_def Ext_pre_choicef_def
    Ext_choicef_def Int_choicef_def Rep_int_choicef_def Conditionalf_def
    Parallelf_def Hidingf_def Renamingf_def Seq_compof_def

(*********************************************************
                    semantic clause
 *********************************************************)

consts
  evalT  :: "'a proc => 'a domT"  ("[[_]]T")
  evalF  :: "'a proc => 'a domF"  ("[[_]]F")
  evalSF :: "'a proc => 'a domSF" ("[[_]]SF")

primrec
  "[[STOP]]T       = STOPt"
  "[[SKIP]]T       = SKIPt"
  "[[DIV]]T        = DIVt"
  "[[a -> P]]T     = a ->t [[P]]T"
  "[[? :X -> Pf]]T = ? :X ->t (%a. [[Pf a]]T)"
  "[[P [+] Q]]T    = [[P]]T [+]t [[Q]]T"
  "[[P |~| Q]]T    = [[P]]T |~|t [[Q]]T"
  "[[!! :X .. Pf]]T = !! :X ..t (%a. [[Pf a]]T)"
  "[[IF b THEN P ELSE Q]]T = IFt b THENt [[P]]T ELSEt [[Q]]T"
  "[[P |[X]| Q]]T  = [[P]]T |[X]|t [[Q]]T"
  "[[P -- X]]T     = [[P]]T --t X"
  "[[P [[r]]]]T    = [[P]]T [[r]]t"
  "[[P ;; Q]]T     = [[P]]T ;;t [[Q]]T"

primrec
  "[[STOP]]F       = STOPf"
  "[[SKIP]]F       = SKIPf"
  "[[DIV]]F        = DIVf"
  "[[a -> P]]F     = a ->f [[P]]F"
  "[[? :X -> Pf]]F = ? :X ->f (%a. [[Pf a]]F)"
  "[[P [+] Q]]F    = ([[P]]T,[[P]]F) [+]f ([[Q]]T,[[Q]]F)"
  "[[P |~| Q]]F    = [[P]]F |~|f [[Q]]F"
  "[[!! :X .. Pf]]F = !! :X ..f (%a. [[Pf a]]F)"
  "[[IF b THEN P ELSE Q]]F = IFf b THENf [[P]]F ELSEf [[Q]]F"
  "[[P |[X]| Q]]F  = [[P]]F |[X]|f [[Q]]F"
  "[[P -- X]]F     = [[P]]F --f X"
  "[[P [[r]]]]F    = [[P]]F [[r]]f"
  "[[P ;; Q]]F     = ([[P]]T,[[P]]F) ;;f [[Q]]F"

declare evalT.simps [simp del]
declare evalF.simps [simp del]

lemma evalT_Rep_int_choice_OR1: 
  "[[! :X .. Pf]]T = !! :OR1 ` X ..t (%a. [[Pf (inv OR1 a)]]T)"
apply (simp add: Rep_int_choice_OR1_def)
apply (simp add: evalT.simps)
done

lemma evalT_Rep_int_choice_OR2: 
  "[[!nat :X .. Pf]]T = !! :OR2 ` X ..t (%a. [[Pf (inv OR2 a)]]T)"
apply (simp add: Rep_int_choice_OR2_def)
apply (simp add: evalT.simps)
done

lemma evalF_Rep_int_choice_OR1: 
  "[[! :X .. Pf]]F = !! :OR1 ` X ..f (%a. [[Pf (inv OR1 a)]]F)"
apply (simp add: Rep_int_choice_OR1_def)
apply (simp add: evalF.simps)
done

lemma evalF_Rep_int_choice_OR2: 
  "[[!nat :X .. Pf]]F = !! :OR2 ` X ..f (%a. [[Pf (inv OR2 a)]]F)"
apply (simp add: Rep_int_choice_OR2_def)
apply (simp add: evalF.simps)
done

lemmas evalT_def = evalT.simps
                   evalT_Rep_int_choice_OR1 
                   evalT_Rep_int_choice_OR2

lemmas evalF_def = evalF.simps
                   evalF_Rep_int_choice_OR1 
                   evalF_Rep_int_choice_OR2

(*** test ***)
(*
lemma test1: "[[! :X .. Pf]]T = {t. t = []t | (EX a:X. t :t [[Pf a]]T) }t"
by (simp add: evalT_def add: evalt_def)

lemma test2: "[[! :X .. Pf]]F = {f. (EX a:X. f :f [[Pf a]]F) }f"
by (simp add: evalF_def add: evalf_def)
*)

defs
  evalSF_def : "[[P]]SF == ([[P]]T ,, [[P]]F)"

(*********************************************************
              relations over processes
 *********************************************************)

consts
  refSF :: "'a proc => 'a proc => bool"
                                           ("(0_ /<=SF /_)" [50,50] 50)
  eqSF  :: "'a proc => 'a proc => bool"
                                           ("(0_ /=SF /_)" [50,50] 50)
  refSF_prod :: "('n => 'a proc) => ('n => 'a proc) => bool"
                                           ("(0_ /<==SF /_)" [50,50] 50)
  eqSF_prod  :: "('n => 'a proc) => ('n => 'a proc) => bool"
                                           ("(0_ /==SF /_)" [50,50] 50)

defs
  refSF_def : "P1 <=SF P2 == [[P2]]SF <= [[P1]]SF"
  eqSF_def  : "P1  =SF P2 == [[P1]]SF  = [[P2]]SF"
  refSF_prod_def : "Pp1 <==SF Pp2 == (ALL n. (Pp1 n) <=SF (Pp2 n))"
  eqSF_prod_def  : "Pp1  ==SF Pp2 == (ALL n. (Pp1 n)  =SF (Pp2 n))"

(*------------------------------------*
 |              X-Symbols             |
 *------------------------------------*)

syntax (output)
  "_refSFX"    :: "'a proc => 'a proc => bool"
                   ("(0_ /<=SF /_)" [50,50] 50)

syntax (xsymbols)
  "_refSFX"    :: "'a proc => 'a proc => bool"
                   ("(0_ /\<sqsubseteq>SF /_)" [50,50] 50)

translations
  "P1 \<sqsubseteq>SF P2" == "P1 <=SF P2"

(*------------------*
 |      csp law     |
 *------------------*)

(*** eq and ref ***)

lemma csp_eq_ref_iff:
  "(P1 =SF P2) = (P1 <=SF P2 & P2 <=SF P1)"
by (auto simp add: refSF_def eqSF_def)

lemma csp_eq_ref:
  "P1 =SF P2 ==> P1 <=SF P2"
by (simp add: csp_eq_ref_iff)

lemma csp_ref_eq:
  "[| P1 <=SF P2 ; P2 <=SF P1 |] ==> P1 =SF P2"
by (simp add: csp_eq_ref_iff)

(*** reflexivity ***)

lemma csp_reflex_eq[simp]: "P =SF P"
by (simp add: eqSF_def)

lemma csp_reflex_ref[simp]: "P <=SF P"
by (simp add: refSF_def)

lemmas csp_reflex = csp_reflex_eq csp_reflex_ref

(*** symmetry ***)

lemma csp_sym: "P1 =SF P2 ==> P2 =SF P1"
by (simp add: eqSF_def)

lemma csp_symE:
  "[| P1 =SF P2 ; P2 =SF P1 ==> Z |] ==> Z"
by (simp add: eqSF_def)

(*** transitivity ***)

lemma csp_trans_left_eq: "[| P1 =SF P2 ; P2 =SF P3 |] ==> P1 =SF P3"
by (simp add: eqSF_def)

lemma csp_trans_left_ref: "[| P1 <=SF P2 ; P2 <=SF P3 |] ==> P1 <=SF P3"
by (simp add: refSF_def)

lemmas csp_trans_left = csp_trans_left_eq csp_trans_left_ref
lemmas csp_trans = csp_trans_left

lemma csp_trans_right_eq: "[| P2 =SF P3 ; P1 =SF P2 |] ==> P1 =SF P3"
by (simp add: eqSF_def)

lemma csp_trans_right_ref: "[| P2 <=SF P3 ;  P1 <=SF P2 |] ==> P1 <=SF P3"
by (simp add: refSF_def)

lemmas csp_trans_rught = csp_trans_right_eq csp_trans_right_ref

(*** rewrite (eq) ***)

lemma csp_rw_left_eq:
  "[| P1 =SF P2 ; P2 =SF P3 |] ==> P1 =SF P3"
by (simp add: eqSF_def)

lemma csp_rw_left_ref:
  "[| P1 =SF P2 ; P2 <=SF P3 |] ==> P1 <=SF P3"
by (simp add: refSF_def eqSF_def)

lemmas csp_rw_left = csp_rw_left_eq csp_rw_left_ref

lemma csp_rw_right_eq:
  "[| P3 =SF P2 ; P1 =SF P2 |] ==> P1 =SF P3"
by (simp add: eqSF_def)

lemma csp_rw_right_ref:
  "[| P3 =SF P2 ; P1 <=SF P2 |] ==> P1 <=SF P3"
by (simp add: refSF_def eqSF_def)

lemmas csp_rw_right = csp_rw_right_eq csp_rw_right_ref

(*** rewrite (ref) ***)

lemma csp_tr_left_eq: "[| P1 =SF P2 ; P2 =SF P3 |] ==> P1 =SF P3"
by (simp add: eqSF_def)

lemma csp_tr_left_ref: "[| P1 <=SF P2 ; P2 <=SF P3 |] ==> P1 <=SF P3"
by (simp add: refSF_def eqSF_def)

lemmas csp_tr_left = csp_tr_left_eq csp_tr_left_ref

lemma csp_tr_right_eq: "[| P2 =SF P3 ; P1 =SF P2 |] ==> P1 =SF P3"
by (simp add: eqSF_def)

lemma csp_tr_right_ref: "[| P2 <=SF P3 ; P1 <=SF P2 |] ==> P1 <=SF P3"
by (simp add: refSF_def eqSF_def)

lemmas csp_tr_right = csp_tr_right_eq csp_tr_right_ref

end
