           (*-------------------------------------------*
            |                CSP-Prover                 |
            |                June 2005                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_semantics_exp = CSP_semantics:

(*****************************************************************

         1. semantic clause for process expressions
         2. 
         3. 
         4. 

 *****************************************************************)

(*********************************************************
             semantics of process expressions
 *********************************************************)

consts
  evalTFx :: 
    "(('pn,'a) procx * 
      (('pn,'a) domSF_prod => 'a domT) *
      (('pn,'a) domSF_prod => 'a domF)) set"

inductive "evalTFx"
intros
evalTFx_pn:
  "((%x. x pn), (%x. fstSF (x pn)), (%x. sndSF (x pn))) : evalTFx"

evalTFx_STOP: 
  "((%x. STOP), (%x. STOPt), (%x. STOPf)) : evalTFx"

evalTFx_SKIP: 
  "((%x. SKIP), (%x. SKIPt), (%x. SKIPf)) : evalTFx"

evalTFx_DIV: 
  "((%x. DIV), (%x. DIVt), (%x. DIVf)) : evalTFx"

evalTFx_Act_prefix:
  "(P,T,F) : evalTFx ==> 
   ((%x. a -> (P x)), (%x. a ->t (T x)), (%x. a ->f (F x))) : evalTFx"

evalTFx_Ext_pre_choice:
  "[| ALL a. (Pf a, Tf a, Ff a) : evalTFx |] ==> 
   ((%x. ? :X -> (%a. (Pf a x))), 
    (%x. ? :X ->t (%a. (Tf a x))), (%x. ? :X ->f (%a. (Ff a x)))) : evalTFx"

evalTFx_Ext_choice:
  "[| (P, T, F) : evalTFx ; (Q, S, E) : evalTFx |] ==>
   ((%x. (P x) [+] (Q x)), (%x. (T x) [+]t (S x)),
    (%x. (T x,F x) [+]f (S x,E x))) : evalTFx"

evalTFx_Int_choice:
  "[| (P, T, F) : evalTFx ; (Q, S, E) : evalTFx |] ==>
   ((%x. (P x) |~| (Q x)), (%x. (T x) |~|t (S x)), (%x. (F x) |~|f (E x)))
   : evalTFx"

evalTFx_Rep_int_choice:
  "[| ALL a. (Pf a, Tf a, Ff a) : evalTFx |] ==> 
   ((%x. !! :Y .. (%a. (Pf a x))), 
    (%x. !! :Y ..t (%a. (Tf a x))), (%x. !! :Y ..f (%a. (Ff a x)))) : evalTFx"

evalTFx_Conditional:
  "[| (P, T, F) : evalTFx ; (Q, S, E) : evalTFx |] ==>
   ((%x. IF b THEN (P x) ELSE (Q x)),
    (%x. IFt b THENt (T x) ELSEt (S x)),
    (%x. IFf b THENf (F x) ELSEf (E x))) : evalTFx"

evalTFx_Parallel:
  "[| (P, T, F) : evalTFx ; (Q, S, E) : evalTFx |] ==>
   ((%x. (P x) |[X]| (Q x)), (%x. (T x) |[X]|t (S x)), (%x. (F x) |[X]|f (E x)))
   : evalTFx"

evalTFx_Hiding:
  "[| (P, T, F) : evalTFx |] ==>
   ((%x. (P x) -- X), (%x. (T x) --t X), (%x. (F x) --f X))
   : evalTFx"

evalTFx_Renaming:
  "[| (P, T, F) : evalTFx |] ==>
   ((%x. (P x) [[r]]), (%x. (T x) [[r]]t), (%x. (F x) [[r]]f))
   : evalTFx"

evalTFx_Seq_compo:
  "[| (P, T, F) : evalTFx ; (Q, S, E) : evalTFx |] ==>
   ((%x. (P x) ;; (Q x)), (%x. (T x) ;;t (S x)),
    (%x. (T x,F x) ;;f (E x))) : evalTFx"

lemmas evalTFx_def = 
  evalTFx_pn evalTFx_STOP  evalTFx_SKIP        evalTFx_DIV
  evalTFx_Act_prefix   evalTFx_Ext_pre_choice   evalTFx_Ext_choice 
  evalTFx_Int_choice   evalTFx_Rep_int_choice   evalTFx_Conditional
  evalTFx_Parallel     evalTFx_Hiding           evalTFx_Renaming
  evalTFx_Seq_compo

(*** process is in procx ***)

lemma proc_in_procx[simp]: "(%x. P) : Procx"
apply (induct_tac P)
by (simp_all add: Procx_def)

(**************************
   evaluation of proc exp
 **************************)

consts
  evalTHEx ::
     "('pn,'a) procx =>
      ((('pn,'a) domSF_prod => 'a domT) * (('pn,'a) domSF_prod => 'a domF))"
                                                         ("[[_]]THEx")
  evalTx  :: "('pn,'a) procx =>
                      (('pn,'a) domSF_prod => 'a domT)"  ("[[_]]Tx")
  evalFx  :: "('pn,'a) procx =>
                      (('pn,'a) domSF_prod => 'a domF)"  ("[[_]]Fx")
  evalSFx :: "('pn,'a) procx =>
                      (('pn,'a) domSF_prod => 'a domSF)" ("[[_]]SFx")

  evalSFX :: "('pn,'a) procX =>
              (('pn,'a) domSF_prod => ('pn,'a) domSF_prod)"    ("[[_]]SFX")

defs
  evalTHEx_def : 
    "[[P]]THEx == (THE TF. (P,TF) : evalTFx)"
  evalTx_def : 
    "[[P]]Tx == (%x. (fst [[P]]THEx) x)"
  evalFx_def : 
    "[[P]]Fx == (%x. (snd [[P]]THEx) x)"
  evalSFx_def : 
    "[[P]]SFx == (%x. ([[P]]Tx x ,, [[P]]Fx x))"
  evalSFX_def : 
    "[[P]]SFX == (%X. (%pn. ([[(%X. (P X) pn)]]SFx X)))"

(**************************************************************
                           lemmas
 **************************************************************)

(*---------------*
 |  in proc_exp  |
 *---------------*)

lemma evalTFx_proc_exp:
  "(P,T,F) : evalTFx ==> P : Procx"
apply (erule evalTFx.induct)
by (simp_all add: Procx_def)

(*---------------*
 |   existency   |
 *---------------*)

lemma evalTFx_ex:
  "P : Procx ==> EX T F. (P,T,F) : evalTFx"
apply (erule Procx.induct)

apply (rule_tac x="(%x. fstSF (x pn))" in exI)
apply (rule_tac x="(%x. sndSF (x pn))" in exI)
apply (simp add: evalTFx_pn)

apply (rule_tac x="(%x. STOPt)" in exI)
apply (rule_tac x="(%x. STOPf)" in exI)
apply (simp add: evalTFx_STOP)

apply (rule_tac x="(%x. SKIPt)" in exI)
apply (rule_tac x="(%x. SKIPf)" in exI)
apply (simp add: evalTFx_SKIP)

apply (rule_tac x="(%x. DIVt)" in exI)
apply (rule_tac x="(%x. DIVf)" in exI)
apply (simp add: evalTFx_DIV)

apply (elim conjE exE)
apply (rule_tac x="(%x. a ->t T x)" in exI)
apply (rule_tac x="(%x. a ->f F x)" in exI)
apply (simp add: evalTFx_Act_prefix)

apply (subgoal_tac 
  "EX TFf. ALL a. (Pf a, (fst (TFf a)), (snd (TFf a))) : evalTFx")
apply (elim conjE exE)
apply (rule_tac x="(%x. ? :X ->t (%a. (fst (TFf a) x)))" in exI)
apply (rule_tac x="(%x. ? :X ->f (%a. (snd (TFf a) x)))" in exI)
apply (simp add: evalTFx_Ext_pre_choice)
apply (rule_tac x="(%a. SOME TF. (Pf a, fst TF, snd TF) : evalTFx)" in exI)
apply (intro allI)
apply (drule_tac x="a" in spec)
apply (elim conjE exE)
apply (simp)
apply (rule someI2)
apply (simp)
apply (simp)

apply (elim conjE exE)
apply (rule_tac x="(%x. (T x) [+]t (Ta x))" in exI)
apply (rule_tac x="(%x. ((T x),(F x)) [+]f ((Ta x),(Fa x)))" in exI)
apply (simp add: evalTFx_Ext_choice)

apply (elim conjE exE)
apply (rule_tac x="(%x. (T x) |~|t (Ta x))" in exI)
apply (rule_tac x="(%x. (F x) |~|f (Fa x))" in exI)
apply (simp add: evalTFx_Int_choice)

apply (subgoal_tac 
  "EX TFf. ALL a. (Pf a, (fst (TFf a)), (snd (TFf a))) : evalTFx")
apply (elim conjE exE)
apply (rule_tac x="(%x. !! :X ..t (%a. (fst (TFf a) x)))" in exI)
apply (rule_tac x="(%x. !! :X ..f (%a. (snd (TFf a) x)))" in exI)
apply (simp add: evalTFx_Rep_int_choice)
apply (rule_tac x="(%a. SOME TF. (Pf a, fst TF, snd TF) : evalTFx)" in exI)
apply (intro allI)
apply (drule_tac x="a" in spec)
apply (elim conjE exE)
apply (simp)
apply (rule someI2)
apply (simp)
apply (simp)

apply (elim conjE exE)
apply (rule_tac x="(%x. IFt b THENt (T x) ELSEt (Ta x))" in exI)
apply (rule_tac x="(%x. IFf b THENf (F x) ELSEf (Fa x))" in exI)
apply (simp add: evalTFx_Conditional)

apply (elim conjE exE)
apply (rule_tac x="(%x. (T x) |[X]|t (Ta x))" in exI)
apply (rule_tac x="(%x. (F x) |[X]|f (Fa x))" in exI)
apply (simp add: evalTFx_Parallel)

apply (elim conjE exE)
apply (rule_tac x="(%x. (T x) --t X)" in exI)
apply (rule_tac x="(%x. (F x) --f X)" in exI)
apply (simp add: evalTFx_Hiding)

apply (elim conjE exE)
apply (rule_tac x="(%x. (T x) [[r]]t)" in exI)
apply (rule_tac x="(%x. (F x) [[r]]f)" in exI)
apply (simp add: evalTFx_Renaming)

apply (elim conjE exE)
apply (rule_tac x="(%x. (T x) ;;t (Ta x))" in exI)
apply (rule_tac x="(%x. ((T x),(F x)) ;;f (Fa x))" in exI)
apply (simp add: evalTFx_Seq_compo)
done

(*---------------*
 |     eval      |
 *---------------*)

(* part *)

(*** pn ***)

lemma evalTFx_pn_simp_only_if:
  "(%x. x pn, T, F) : evalTFx
     ==> (T = (%x. fstSF (x pn))) & (F = (%x. sndSF (x pn)))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)

apply (elim conjE)
apply (case_tac "pna = pn", simp)
apply (drule_tac x="(%x. if x = pn then STOP else SKIP)" in spec)
apply (simp)
apply (elim conjE exE, drule_tac x="(%x. SKIP)" in spec, simp)
by (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)+

lemma evalTFx_pn_simp:
  "(%x. x pn, T, F) : evalTFx
     = ((T = (%x. fstSF (x pn))) & (F = (%x. sndSF (x pn))))"
apply (rule)
apply (simp add: evalTFx_pn_simp_only_if)
apply (simp add: evalTFx_pn)
done

(*** STOP ***)

lemma evalTFx_STOP_simp_only_if:
  "(%x. STOP, T, F) : evalTFx ==> (T = (%x. STOPt)) & (F = (%x. STOPf))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
by (elim conjE exE, drule_tac x="(%x. SKIP)" in spec, simp)

lemma evalTFx_STOP_simp:
  "(%x. STOP, T, F) : evalTFx = ((T = (%x. STOPt)) & (F = (%x. STOPf)))"
apply (rule)
apply (simp add: evalTFx_STOP_simp_only_if)
apply (simp add: evalTFx_STOP)
done

(*** SKIP ***)

lemma evalTFx_SKIP_simp_only_if:
  "(%x. SKIP, T, F) : evalTFx ==> (T = (%x. SKIPt)) & (F = (%x. SKIPf))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
by (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)

lemma evalTFx_SKIP_simp:
  "(%x. SKIP, T, F) : evalTFx = ((T = (%x. SKIPt)) & (F = (%x. SKIPf)))"
apply (rule)
apply (simp add: evalTFx_SKIP_simp_only_if)
apply (simp add: evalTFx_SKIP)
done

(*** DIV ***)

lemma evalTFx_DIV_simp_only_if:
  "(%x. DIV, T, F) : evalTFx ==> (T = (%x. DIVt)) & (F = (%x. DIVf))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
by (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)

lemma evalTFx_DIV_simp:
  "(%x. DIV, T, F) : evalTFx = ((T = (%x. DIVt)) & (F = (%x. DIVf)))"
apply (rule)
apply (simp add: evalTFx_DIV_simp_only_if)
apply (simp add: evalTFx_DIV)
done

(*** Act_prefix ***)

lemma evalTFx_Act_prefix_simp_only_if:
  "(%x. a -> P x, T, F) : evalTFx
    ==> EX T' F'. ((P, T', F') : evalTFx &
                  (T = (%x. a ->t (T' x))) & (F = (%x. a ->f (F' x))))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (simp add: expand_fun_eq)
done

lemma evalTFx_Act_prefix_simp:
  "(%x. a -> P x, T, F) : evalTFx
    = (EX T' F'. ((P, T', F') : evalTFx &
                 (T = (%x. a ->t (T' x))) & (F = (%x. a ->f (F' x)))))"
apply (rule)
apply (simp add: evalTFx_Act_prefix_simp_only_if)
apply (auto simp add: evalTFx_Act_prefix)
done

(*** Ext_pre_choice ***)

lemma evalTFx_Ext_pre_choice_simp_only_if:
  "((%x. ? a:X -> (Pf a x)), T, F) : evalTFx
    ==> EX Tf Ff. ((ALL a. (Pf a, Tf a, Ff a) : evalTFx) &
                   T = (%x. ? :X ->t (%a. (Tf a x))) & 
                   F = (%x. ? :X ->f (%a. (Ff a x))))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Tf" in exI)
apply (rule_tac x="Ff" in exI)
apply (subgoal_tac "Pf = Pfa", simp)
apply (simp add: expand_fun_eq)
done

lemma evalTFx_Ext_pre_choice_simp:
  "((%x. ? a:X -> (Pf a x)), T, F) : evalTFx
    = (EX Tf Ff. ((ALL a. (Pf a, Tf a, Ff a) : evalTFx) &
                 T = (%x. ? :X ->t (%a. (Tf a x))) & 
                 F = (%x. ? :X ->f (%a. (Ff a x)))))"
apply (rule)
apply (simp add: evalTFx_Ext_pre_choice_simp_only_if)
apply (auto simp add: evalTFx_Ext_pre_choice)
done

(*** Ext_choice ***)

lemma evalTFx_Ext_choice_simp_only_if:
  "((%x. (P x) [+] (Q x)), T, F) : evalTFx
    ==> EX T1 F1. (P, T1, F1) : evalTFx & 
       (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) [+]t (T2 x)) & 
                         F = (%x. ((T1 x),(F1 x)) [+]f ((T2 x),(F2 x))))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (rule_tac x="S" in exI)
apply (rule_tac x="E" in exI)
apply (subgoal_tac "Q = Qa", simp)
apply (simp_all add: expand_fun_eq)
done

lemma evalTFx_Ext_choice_simp:
  "((%x. (P x) [+] (Q x)), T, F) : evalTFx
    = ( EX T1 F1. (P, T1, F1) : evalTFx & 
       (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) [+]t (T2 x)) & 
                         F = (%x. ((T1 x),(F1 x)) [+]f ((T2 x),(F2 x)))))"
apply (rule)
apply (simp add: evalTFx_Ext_choice_simp_only_if)
apply (auto simp add: evalTFx_Ext_choice)
done

(*** Int_choice ***)

lemma evalTFx_Int_choice_simp_only_if:
  "((%x. (P x) |~| (Q x)), T, F) : evalTFx
    ==> EX T1 F1. (P, T1, F1) : evalTFx & 
       (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) |~|t (T2 x)) & 
                         F = (%x. (F1 x) |~|f (F2 x)))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (rule_tac x="S" in exI)
apply (rule_tac x="E" in exI)
apply (subgoal_tac "Q = Qa", simp)
apply (simp_all add: expand_fun_eq)
done

lemma evalTFx_Int_choice_simp:
  "((%x. (P x) |~| (Q x)), T, F) : evalTFx
     = (EX T1 F1. (P, T1, F1) : evalTFx & 
       (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) |~|t (T2 x)) & 
                         F = (%x. (F1 x) |~|f (F2 x))))"
apply (rule)
apply (simp add: evalTFx_Int_choice_simp_only_if)
apply (auto simp add: evalTFx_Int_choice)
done

(*** Rep_int_choice ***)

lemma evalTFx_Rep_int_choice_simp_only_if:
  "((%x. !! :X .. (%a. (Pf a x))), T, F) : evalTFx
    ==> EX Tf Ff. ((ALL a. (Pf a, Tf a, Ff a) : evalTFx) &
                   T = (%x. !! :X ..t (%a. (Tf a x))) & 
                   F = (%x. !! :X ..f (%a. (Ff a x))))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Tf" in exI)
apply (rule_tac x="Ff" in exI)
apply (subgoal_tac "Pf = Pfa", simp)
apply (simp add: expand_fun_eq)
done

lemma evalTFx_Rep_int_choice_simp:
  "((%x. !! :X .. (%a. (Pf a x))), T, F) : evalTFx
    = (EX Tf Ff. ((ALL a. (Pf a, Tf a, Ff a) : evalTFx) &
                  T = (%x. !! :X ..t (%a. (Tf a x))) & 
                  F = (%x. !! :X ..f (%a. (Ff a x)))))"
apply (rule)
apply (simp add: evalTFx_Rep_int_choice_simp_only_if)
apply (auto simp add: evalTFx_Rep_int_choice)
done

(*** Conditional ***)

lemma evalTFx_Conditional_simp_only_if:
  "((%x. IF b THEN (P x) ELSE (Q x)), T, F) : evalTFx
    ==> EX T1 F1. (P, T1, F1) : evalTFx &
       (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. IFt b THENt (T1 x) ELSEt (T2 x)) & 
                         F = (%x. IFf b THENf (F1 x) ELSEf (F2 x)))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (rule_tac x="S" in exI)
apply (rule_tac x="E" in exI)
apply (subgoal_tac "Q = Qa", simp)
apply (simp_all add: expand_fun_eq)
done

lemma evalTFx_Conditional_simp:
  "((%x. IF b THEN (P x) ELSE (Q x)), T, F) : evalTFx
    = (EX T1 F1. (P, T1, F1) : evalTFx &
      (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. IFt b THENt (T1 x) ELSEt (T2 x)) & 
                         F = (%x. IFf b THENf (F1 x) ELSEf (F2 x))))"
apply (rule)
apply (simp add: evalTFx_Conditional_simp_only_if)
apply (auto simp add: evalTFx_Conditional)
done

(*** Parallel ***)

lemma evalTFx_Parallel_simp_only_if:
  "((%x. (P x) |[X]| (Q x)), T, F) : evalTFx
    ==> EX T1 F1. (P, T1, F1) : evalTFx & 
       (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) |[X]|t (T2 x)) & 
                         F = (%x. (F1 x) |[X]|f (F2 x)))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (rule_tac x="S" in exI)
apply (rule_tac x="E" in exI)
apply (subgoal_tac "Q = Qa", simp)
apply (simp_all add: expand_fun_eq)
done

lemma evalTFx_Parallel_simp:
  "((%x. (P x) |[X]| (Q x)), T, F) : evalTFx
    = (EX T1 F1. (P, T1, F1) : evalTFx & 
      (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) |[X]|t (T2 x)) & 
                         F = (%x. (F1 x) |[X]|f (F2 x))))"
apply (rule)
apply (simp add: evalTFx_Parallel_simp_only_if)
apply (auto simp add: evalTFx_Parallel)
done

(*** Hiding ***)

lemma evalTFx_Hiding_simp_only_if:
  "((%x. (P x) -- X), T, F) : evalTFx
    ==> EX T' F'. ((P, T', F') : evalTFx &
                   T = (%x. (T' x) --t X) & F = (%x. (F' x) --f X))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (simp add: expand_fun_eq)
done

lemma evalTFx_Hiding_simp:
  "((%x. (P x) -- X), T, F) : evalTFx
    = (EX T' F'. ((P, T', F') : evalTFx &
                   T = (%x. (T' x) --t X) & F = (%x. (F' x) --f X)))"
apply (rule)
apply (simp add: evalTFx_Hiding_simp_only_if)
apply (auto simp add: evalTFx_Hiding)
done

(*** Renaming ***)

lemma evalTFx_Renaming_simp_only_if:
  "((%x. (P x) [[r]]), T, F) : evalTFx
    ==> EX T' F'. ((P, T', F') : evalTFx &
                   T = (%x. (T' x) [[r]]t) & F = (%x. (F' x) [[r]]f))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (simp add: expand_fun_eq)
done

lemma evalTFx_Renaming_simp:
  "((%x. (P x) [[r]]), T, F) : evalTFx
    = (EX T' F'. ((P, T', F') : evalTFx &
                 T = (%x. (T' x) [[r]]t) & F = (%x. (F' x) [[r]]f)))"
apply (rule)
apply (simp add: evalTFx_Renaming_simp_only_if)
apply (auto simp add: evalTFx_Renaming)
done

(*** Seq_compo ***)

lemma evalTFx_Seq_compo_simp_only_if:
  "((%x. (P x) ;; (Q x)), T, F) : evalTFx
    ==> EX T1 F1. (P, T1, F1) : evalTFx & 
       (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) ;;t (T2 x)) & 
                         F = (%x. ((T1 x),(F1 x)) ;;f (F2 x)))"
apply (erule evalTFx.cases)
apply (simp_all add: expand_fun_eq)
apply (elim conjE exE, drule_tac x="(%x. STOP)" in spec, simp)
apply (elim conjE)
apply (rule_tac x="Ta" in exI)
apply (rule_tac x="Fa" in exI)
apply (subgoal_tac "P = Pa", simp)
apply (rule_tac x="S" in exI)
apply (rule_tac x="E" in exI)
apply (subgoal_tac "Q = Qa", simp)
apply (simp_all add: expand_fun_eq)
done

lemma evalTFx_Seq_compo_simp:
  "((%x. (P x) ;; (Q x)), T, F) : evalTFx
    = (EX T1 F1. (P, T1, F1) : evalTFx & 
      (EX T2 F2. (Q, T2, F2) : evalTFx & 
                         T = (%x. (T1 x) ;;t (T2 x)) & 
                         F = (%x. ((T1 x),(F1 x)) ;;f (F2 x))))"
apply (rule)
apply (simp add: evalTFx_Seq_compo_simp_only_if)
apply (auto simp add: evalTFx_Seq_compo)
done

lemmas evalTFx_simp = 
  evalTFx_pn_simp   evalTFx_STOP_simp   evalTFx_SKIP_simp
  evalTFx_DIV_simp  evalTFx_Act_prefix_simp 
  evalTFx_Ext_pre_choice_simp evalTFx_Ext_choice_simp
  evalTFx_Int_choice_simp     evalTFx_Rep_int_choice_simp
  evalTFx_Conditional_simp    evalTFx_Parallel_simp
  evalTFx_Hiding_simp         evalTFx_Renaming_simp
  evalTFx_Seq_compo_simp

(*** process expression unique ***)

lemma evalTFx_unique_induct:
  "P : Procx ==>
   ALL S E T F. (P,S,E) : evalTFx & (P,T,F) : evalTFx
                 --> (S = T & E = F)"
apply (rule Procx.induct[of P
   "(%P. (ALL S E T F. (P,S,E) : evalTFx & (P,T,F) : evalTFx
         --> (S = T & E = F)))"])
apply (simp_all add: evalTFx_simp)

(* Act_prefix *)
apply (fast)

(* Ext_pre_choice *)
apply (intro allI impI)
apply (elim conjE exE)
apply (simp)
apply (subgoal_tac "ALL a. Tf a = Tfa a", simp)
apply (subgoal_tac "ALL a. Ff a = Ffa a", simp)
apply (intro allI)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (force)
apply (intro allI)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (force)

(* Ext_choice *)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="T1" in spec)
apply (drule_tac x="F1" in spec)
apply (drule_tac x="T2" in spec)
apply (drule_tac x="F2" in spec)
apply (drule_tac x="T1a" in spec)
apply (drule_tac x="F1a" in spec)
apply (drule_tac x="T2a" in spec)
apply (drule_tac x="F2a" in spec)
apply (simp)

(* Int_choice *)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="T1" in spec)
apply (drule_tac x="F1" in spec)
apply (drule_tac x="T2" in spec)
apply (drule_tac x="F2" in spec)
apply (drule_tac x="T1a" in spec)
apply (drule_tac x="F1a" in spec)
apply (drule_tac x="T2a" in spec)
apply (drule_tac x="F2a" in spec)
apply (simp)

(* Rep_int_choice *)
apply (intro allI impI)
apply (elim conjE exE)
apply (simp)
apply (subgoal_tac "ALL a. Tf a = Tfa a", simp)
apply (subgoal_tac "ALL a. Ff a = Ffa a", simp)
apply (intro allI)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (force)
apply (intro allI)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (force)

(* Conditional *)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="T1" in spec)
apply (drule_tac x="F1" in spec)
apply (drule_tac x="T2" in spec)
apply (drule_tac x="F2" in spec)
apply (drule_tac x="T1a" in spec)
apply (drule_tac x="F1a" in spec)
apply (drule_tac x="T2a" in spec)
apply (drule_tac x="F2a" in spec)
apply (simp)

(* Parallel *)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="T1" in spec)
apply (drule_tac x="F1" in spec)
apply (drule_tac x="T2" in spec)
apply (drule_tac x="F2" in spec)
apply (drule_tac x="T1a" in spec)
apply (drule_tac x="F1a" in spec)
apply (drule_tac x="T2a" in spec)
apply (drule_tac x="F2a" in spec)
apply (simp)

(* Hiding *)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="T'" in spec)
apply (drule_tac x="F'" in spec)
apply (drule_tac x="T'a" in spec)
apply (drule_tac x="F'a" in spec)
apply (simp)

(* Renaming *)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="T'" in spec)
apply (drule_tac x="F'" in spec)
apply (drule_tac x="T'a" in spec)
apply (drule_tac x="F'a" in spec)
apply (simp)

(* Seq_compo *)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="T1" in spec)
apply (drule_tac x="F1" in spec)
apply (drule_tac x="T2" in spec)
apply (drule_tac x="F2" in spec)
apply (drule_tac x="T1a" in spec)
apply (drule_tac x="F1a" in spec)
apply (drule_tac x="T2a" in spec)
apply (drule_tac x="F2a" in spec)
apply (simp)
done

(*** proc_exp unique ***)

lemma evalTFx_unique:
  "[| (P,S,E) : evalTFx ; (P,T,F) : evalTFx |]
   ==> (S = T & E = F)"
apply (subgoal_tac "P : Procx")
apply (insert evalTFx_unique_induct[of P], simp)
apply (drule_tac x="S" in spec)
apply (drule_tac x="E" in spec)
apply (drule_tac x="T" in spec)
apply (drule_tac x="F" in spec)
apply (simp)
apply (simp add: evalTFx_proc_exp)
done

(*********************************************************************
       lemmas for  [[ ]]THEx, [[ ]]Tx, [[ ]]Fx, [[ ]]SFx
 *********************************************************************)

lemma evalTFx_evalTHEx : 
    "(P,TF) : evalTFx = ([[P]]THEx = TF & P : Procx)" 
apply (rule iffI)

apply (subgoal_tac "EX T F. TF = (T,F)")
apply (elim exE)
apply (simp add: evalTHEx_def)
apply (simp add: evalTFx_proc_exp)
apply (rule the_equality)
apply (simp add: evalTFx_ex)
apply (simp add: pair_eq_decompo)
apply (rule evalTFx_unique, simp, simp)
apply (force)

apply (elim conjE)
apply (drule sym)
apply (insert evalTFx_ex[of P], simp)
apply (elim conjE exE)
apply (simp add: evalTHEx_def)
apply (rule theI)
apply (simp)
apply (simp add: pair_eq_decompo)
apply (rule evalTFx_unique, simp, simp)
done

lemma evalTHEx_evalTFx1:
  "P : Procx ==> ([[P]]THEx = TF) = ((P,TF) : evalTFx)"
by (simp add: evalTFx_evalTHEx)

lemma evalTHEx_evalTFx2:
  "P : Procx ==> (TF = [[P]]THEx) = ((P,TF) : evalTFx)"
by (auto simp add: evalTFx_evalTHEx)

lemmas evalTHEx_evalTFx = evalTHEx_evalTFx1 evalTHEx_evalTFx2

(*********************************************************
                [[ ]]Tx and [[ ]]Fx
 *********************************************************)

lemma evalTx_evalTFx1:
  "P : Procx ==> ([[P]]Tx = T) = (EX F. ((P,T,F) : evalTFx))"
apply (simp add: evalTHEx_evalTFx1[THEN sym])
apply (simp add: evalTx_def)
apply (simp add: pair_eq_decompo)
done

lemma  evalTx_evalTFx2:
  "P : Procx ==> (T = [[P]]Tx) = (EX F. ((P,T,F) : evalTFx))"
by (auto simp add: evalTx_evalTFx1[THEN sym])

lemmas evalTx_evalTFx = evalTx_evalTFx1 evalTx_evalTFx2

lemma evalTx_evalTFxI1:
  "[| P : Procx; EX F. (P,T,F) : evalTFx |]
   ==> [[P]]Tx= T"
by (simp add: evalTx_evalTFx)

lemma evalTx_evalTFxI2:
  "[| P : Procx; EX F. (P,T,F) : evalTFx |]
   ==> T = [[P]]Tx"
by (simp add: evalTx_evalTFx)

lemmas evalTx_evalTFxI = evalTx_evalTFxI1 evalTx_evalTFxI2

lemma evalFx_evalTFx1:
  "P : Procx ==> ([[P]]Fx = F) = (EX T. ((P,T,F) : evalTFx))"
apply (simp add: evalTHEx_evalTFx1[THEN sym])
apply (simp add: evalFx_def)
apply (simp add: pair_eq_decompo)
done

lemma evalFx_evalTFx2:
  "P : Procx ==> (F = [[P]]Fx) = (EX T. ((P,T,F) : evalTFx))"
by (auto simp add: evalFx_evalTFx1[THEN sym])

lemmas evalFx_evalTFx = evalFx_evalTFx1 evalFx_evalTFx2

lemma evalFx_evalTFxI1:
  "[| P : Procx; EX T. (P,T,F) : evalTFx |]
   ==> [[P]]Fx = F"
by (simp add: evalFx_evalTFx)

lemma evalFx_evalTFxI2:
  "[| P : Procx; EX T. (P,T,F) : evalTFx |]
   ==> F = [[P]]Fx"
by (simp add: evalFx_evalTFx)

lemmas evalFx_evalTFxI = evalFx_evalTFxI1 evalFx_evalTFxI2

(*------------------*
 |  TF and T and F  |
 *------------------*)

lemma  Procx_evalTFx_lm: 
  "[| P : Procx ; [[P]]Tx= T ; [[P]]Fx = F |]
   ==> (P, T, F) : evalTFx"
apply (simp add: evalTx_evalTFx)
apply (simp add: evalFx_evalTFx)
apply (elim exE)
apply (insert evalTFx_unique_induct[of P], simp)
apply (drule_tac x="T" in spec)
apply (drule_tac x="Fa" in spec)
apply (drule_tac x="Ta" in spec)
apply (drule_tac x="F" in spec)
by (simp)

lemma Procx_evalTFx[simp]:
  "P : Procx ==> (P, [[P]]Tx, [[P]]Fx) : evalTFx"
by (simp add: Procx_evalTFx_lm)

(*-----------*
 |  T cases  |
 *-----------*)

lemma evalTx_pn_simp: 
  "[[%x. x pn]]Tx= (%x. fstSF (x pn))"
by (rule evalTx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalTx_STOP_simp: 
  "[[%x. STOP]]Tx= (%x. STOPt)"
by (rule evalTx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalTx_SKIP_simp: 
  "[[%x. SKIP]]Tx= (%x. SKIPt)"
by (rule evalTx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalTx_DIV_simp: 
  "[[%x. DIV]]Tx= (%x. DIVt)"
by (rule evalTx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalTx_Act_prefix_simp:
  "P : Procx ==>
   [[%x. a -> (P x)]]Tx= (%x. a ->t ([[P]]Tx x))"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x="(%x. a ->f ([[P]]Fx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
done

lemma evalTx_Ext_pre_choice_simp:
  "ALL a. Pf a : Procx ==>
   [[%x. ? :X -> (%a. (Pf a x))]]Tx= (%x. ? :X ->t (%a. ([[Pf a]]Tx x)))"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x="(%x. ? :X ->f (%a. ([[Pf a]]Fx x)))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="(%a. [[Pf a]]Tx)" in exI)
apply (rule_tac x="(%a. [[Pf a]]Fx)" in exI)
apply (simp)
done

lemma evalTx_Ext_choice_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) [+] (Q x)]]Tx= (%x. [[P]]Tx x [+]t [[Q]]Tx x)"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x=
  "(%x. (([[P]]Tx x),([[P]]Fx x)) [+]f (([[Q]]Tx x),([[Q]]Fx x)))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalTx_Int_choice_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) |~| (Q x)]]Tx= (%x. [[P]]Tx x |~|t [[Q]]Tx x)"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. ([[P]]Fx x) |~|f ([[Q]]Fx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalTx_Rep_int_choice_simp:
  "ALL a. Pf a : Procx ==>
   [[%x. !! :X .. (%a. (Pf a x))]]Tx= (%x. !! :X ..t (%a. ([[Pf a]]Tx x)))"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x="(%x. !! :X ..f (%a. ([[Pf a]]Fx x)))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="(%a. [[Pf a]]Tx)" in exI)
apply (rule_tac x="(%a. [[Pf a]]Fx)" in exI)
apply (simp)
done

lemma evalTx_Conditional_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. IF b THEN (P x) ELSE (Q x)]]Tx= 
    (%x. IFt b THENt ([[P]]Tx x) ELSEt ([[Q]]Tx x))"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. IFf b THENf ([[P]]Fx x) ELSEf ([[Q]]Fx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalTx_Parallel_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) |[X]| (Q x)]]Tx= (%x. [[P]]Tx x |[X]|t [[Q]]Tx x)"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. ([[P]]Fx x) |[X]|f ([[Q]]Fx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalTx_Hiding_simp:
  "P : Procx ==> 
   [[%x. (P x) -- X]]Tx= (%x. [[P]]Tx x --t X)"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. [[P]]Fx x --f X)" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
done

lemma evalTx_Renaming_simp:
  "P : Procx ==> 
   [[%x. (P x) [[r]]]]Tx= (%x. [[P]]Tx x [[r]]t)"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. [[P]]Fx x [[r]]f)" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
done

lemma evalTx_Seq_compo_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) ;; (Q x)]]Tx= (%x. [[P]]Tx x ;;t [[Q]]Tx x)"
apply (rule evalTx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. (([[P]]Tx x),([[P]]Fx x)) ;;f ([[Q]]Fx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemmas evalTx_simp =
  evalTx_pn_simp evalTx_STOP_simp evalTx_SKIP_simp evalTx_DIV_simp
  evalTx_Act_prefix_simp evalTx_Ext_pre_choice_simp 
  evalTx_Ext_choice_simp evalTx_Int_choice_simp
  evalTx_Rep_int_choice_simp evalTx_Conditional_simp
  evalTx_Parallel_simp evalTx_Hiding_simp
  evalTx_Renaming_simp evalTx_Seq_compo_simp

(*-----------*
 |  F cases  |
 *-----------*)

lemma evalFx_pn_simp: 
  "[[%x. x pn]]Fx = (%x. sndSF (x pn))"
by (rule evalFx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalFx_STOP_simp: 
  "[[%x. STOP]]Fx = (%x. STOPf)"
by (rule evalFx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalFx_SKIP_simp: 
  "[[%x. SKIP]]Fx = (%x. SKIPf)"
by (rule evalFx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalFx_DIV_simp: 
  "[[%x. DIV]]Fx = (%x. DIVf)"
by (rule evalFx_evalTFxI, simp add: Procx_def, simp add: evalTFx_simp)

lemma evalFx_Act_prefix_simp:
  "P : Procx ==>
   [[%x. a -> (P x)]]Fx = (%x. a ->f ([[P]]Fx x))"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x="(%x. a ->t ([[P]]Tx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
done

lemma evalFx_Ext_pre_choice_simp:
  "ALL a. Pf a : Procx ==>
   [[%x. ? :X -> (%a. (Pf a x))]]Fx = (%x. ? :X ->f (%a. ([[Pf a]]Fx x)))"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x="(%x. ? :X ->t (%a. ([[Pf a]]Tx x)))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="(%a. [[Pf a]]Tx)" in exI)
apply (rule_tac x="(%a. [[Pf a]]Fx)" in exI)
apply (simp)
done

lemma evalFx_Ext_choice_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) [+] (Q x)]]Fx = 
    (%x. ([[P]]Tx x, [[P]]Fx x) [+]f ([[Q]]Tx x, [[Q]]Fx x))"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. ([[P]]Tx x [+]t [[Q]]Tx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalFx_Int_choice_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) |~| (Q x)]]Fx = (%x. [[P]]Fx x |~|f [[Q]]Fx x)"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. ([[P]]Tx x) |~|t ([[Q]]Tx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalFx_Rep_int_choice_simp:
  "ALL a. Pf a : Procx ==>
   [[%x. !! :X .. (%a. (Pf a x))]]Fx = (%x. !! :X ..f (%a. ([[Pf a]]Fx x)))"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x="(%x. !! :X ..t (%a. ([[Pf a]]Tx x)))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="(%a. [[Pf a]]Tx)" in exI)
apply (rule_tac x="(%a. [[Pf a]]Fx)" in exI)
apply (simp)
done

lemma evalFx_Conditional_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. IF b THEN (P x) ELSE (Q x)]]Fx = 
    (%x. IFf b THENf ([[P]]Fx x) ELSEf ([[Q]]Fx x))"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. IFt b THENt ([[P]]Tx x) ELSEt ([[Q]]Tx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalFx_Parallel_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) |[X]| (Q x)]]Fx = (%x. [[P]]Fx x |[X]|f [[Q]]Fx x)"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. ([[P]]Tx x) |[X]|t ([[Q]]Tx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemma evalFx_Hiding_simp:
  "P : Procx ==> 
   [[%x. (P x) -- X]]Fx = (%x. [[P]]Fx x --f X)"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. [[P]]Tx x --t X)" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
done

lemma evalFx_Renaming_simp:
  "P : Procx ==> 
   [[%x. (P x) [[r]]]]Fx = (%x. [[P]]Fx x [[r]]f)"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. [[P]]Tx x [[r]]t)" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
done

lemma evalFx_Seq_compo_simp:
  "[| P : Procx ; Q : Procx |] ==> 
   [[%x. (P x) ;; (Q x)]]Fx = (%x. (([[P]]Tx x),([[P]]Fx x)) ;;f ([[Q]]Fx x))"
apply (rule evalFx_evalTFxI)
apply (simp add: Procx_def)
apply (rule_tac x= "(%x. ([[P]]Tx x ;;t [[Q]]Tx x))" in exI)
apply (simp add: evalTFx_simp)
apply (rule_tac x="[[P]]Tx" in exI)
apply (rule_tac x="[[P]]Fx" in exI)
apply (simp)
apply (rule_tac x="[[Q]]Tx" in exI)
apply (rule_tac x="[[Q]]Fx" in exI)
apply (simp)
done

lemmas evalFx_simp =
  evalFx_pn_simp evalFx_STOP_simp evalFx_SKIP_simp evalFx_DIV_simp
  evalFx_Act_prefix_simp evalFx_Ext_pre_choice_simp 
  evalFx_Ext_choice_simp evalFx_Int_choice_simp
  evalFx_Rep_int_choice_simp evalFx_Conditional_simp
  evalFx_Parallel_simp evalFx_Hiding_simp
  evalFx_Renaming_simp evalFx_Seq_compo_simp

(*----------------*
 |   (%x x pn)    |
 *----------------*)

lemma evalSFx_pn[simp]: "[[%x. x pn]]SFx = (%x. x pn)"
apply (simp add: expand_fun_eq)
apply (simp add: evalSFx_def)
apply (simp add: pairSF_def)
apply (simp add: evalTx_simp)
apply (simp add: evalFx_simp)
apply (simp add: fstSF_def sndSF_def)
apply (simp add: Rep_domSF_inverse)
done

(*-----------------------------*
 |   [[P]]SF  <-->  [[P]]SFx   |
 *-----------------------------*)

(*** T and F ***)

lemma evalTx_evalT_evalFx_evalF: 
  "([[%y. P]]Tx x = [[P]]T) & ([[%y. P]]Fx x = [[P]]F)"
apply (induct_tac P)
apply (simp_all add: evalT_def evalTx_simp)
apply (simp_all add: evalF_def evalFx_simp)
done

lemma evalT_evalTx:
  "[[P]]T = [[%y. P]]Tx (%pn. ({[]t}t,, {}f))"
by (simp add: evalTx_evalT_evalFx_evalF)

lemma evalF_evalFx:
  "[[P]]F = [[%y. P]]Fx (%pn. ({[]t}t,, {}f))"
by (simp add: evalTx_evalT_evalFx_evalF)

(*** SF ***)

lemma evalSFx_evalSF: "[[%y. P]]SFx x = [[P]]SF"
apply (simp add: evalSF_def evalSFx_def)
apply (simp add: evalTx_evalT_evalFx_evalF)
done

lemma evalSF_evalSFx: "[[P]]SF = [[%y. P]]SFx (%pn. ({[]t}t,, {}f))"
by (simp add: evalSFx_evalSF)

end
