           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_syntax = Infra:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*-----------------------------------------------------------*
 |                                                           |
 |    Process Type Definitions                               |
 |                                                           |
 |             'a proc : type of process expressions         |
 |                       'a : event                          |
 |                                                           |
 *-----------------------------------------------------------*)

(*********************************************************************
                       process expression
 *********************************************************************)

datatype 
 'a proc
    = STOP

    | SKIP

    | DIV

    | Act_prefix     "'a" "'a proc"      ("(1_ /-> _)" [150,80] 80)

    | Ext_pre_choice "'a set" "'a => 'a proc"
                                               ("(1? :_ /-> _)" [90,80] 80)
    | Ext_choice     "'a proc" "'a proc"  
                                               ("(1_ /[+] _)" [72,73] 72)
    | Int_choice     "'a proc" "'a proc"  
                                               ("(1_ /|~| _)" [64,65] 64)
    | Rep_int_choice "(('a, nat) or_type) set" "(('a, nat) or_type) => 'a proc"
                                               ("(1!! :_ .. /_)" [90,68] 68) 
    | Conditional    "bool" "'a proc" "'a proc"
                                 ("(0IF _ /THEN _ /ELSE _)" [900,60,60] 88)
    | Parallel       "'a proc" "'a set" "'a proc"  
                                               ("(1_ /|[_]| _)" [76,0,77] 76)
    | Hiding         "'a proc" "'a set"   ("(1_ /-- _)" [84,85] 84)

    | Renaming       "'a proc" "('a * 'a) set"
                                               ("(1_ /[[_]])" [84,0] 84)
    | Seq_compo      "'a proc" "'a proc"  
                                               ("(1_ /;; _)" [79,78] 78)

(*--------------------------------------------------------------------*
 |                                                                    |
 | (1) binding power:                                                 |
 |      Proc_name > Conditional > {Hiding, Renaming} >                |
 |      {Act_prefix, Ext_pre_choice} > Seq_compo > Parallel >         |
 |      Ext_choice > Rep_int_choice > int_choice                      |
 |                                                                    |
 |   The binding power 90 is higher than ones of processes ops.       |
 |                                                                    |
 | (2) DIV is a short notation as follows: DIV = ! {} .. (%x. STOP)   |
 |                                                                    |
 | (3) About ("(1_ /-> _)" [150,80] 80)                               |
 |      The nth operator ! has the power 101.                         |
 |      So, 150 > 101 is needed.                                      |
 |                                                                    |
 | (4) natural number is defaultly available in replicated internal   |
 |     choice. it is necessary for defining fixed points.             |
 |                                                                    |
 *--------------------------------------------------------------------*)

(*** external prefix ***)

syntax
  "@Ext_pre_choice"  :: "pttrn => 'a set => 'a proc 
                => 'a proc"  ("(1? _:_ /-> _)" [900,900,80] 80)

translations
  "? x:X -> P"  == "? :X -> (%x. P)"

(*** internal choice ***)

consts
  Rep_int_choice_OR1 :: "'a set => ('a => 'a proc) => 'a proc"
                                      ("(1! :_ .. /_)" [90,68] 68)
  Rep_int_choice_OR2 :: "nat set => (nat => 'a proc) => 'a proc"
                                      ("(1!nat :_ .. /_)" [90,68] 68)

defs
  Rep_int_choice_OR1_def : 
        "! :X .. Pf == !! :(OR1 ` X) .. (%x. (Pf ((inv OR1) x)))"

  Rep_int_choice_OR2_def : 
        "!nat :X .. Pf == !! :(OR2 ` X) .. (%x. (Pf ((inv OR2) x)))"

(*** internal prefix choice ***)

consts
  Int_pre_choice :: "'a set => ('a => 'a proc) => 'a proc"  
                                      ("(1! :_ /-> _)" [900,80] 80)

defs
  Int_pre_choice_def : "! :X -> Pf == ! :X .. (%x. x -> (Pf x))"

syntax
  "@Int_pre_choice"  :: "pttrn => 'a set => 'a proc 
                         => 'a proc"  ("(1! _:_ /-> _)" [900,900,80] 80)

  "@Int_pre_choice_UNIV" ::  "pttrn => 'a proc => 'a proc"
                                           ("(1! _ /-> _)" [900,80] 80)

translations
  "! x:X -> P"  == "! :X -> (%x. P)"
  "! x -> P"    == "! x:UNIV -> P"

(*** sending and receiving prefixes ***)

consts
  Send_prefix :: "('x => 'a) => 'x => 'a proc => 'a proc" 
                                      ("(1_ ! _ /-> _)" [900,1000,80] 80)
  Nondet_send_prefix :: "('x => 'a) => 'x set => 
                         ('x => 'a proc) => 'a proc" 
  Rec_prefix  :: "('x => 'a) => 'x set => 
                  ('x => 'a proc) => 'a proc" 

defs
  Send_prefix_def : 
    "a ! x -> P    == a x -> P"
  Nondet_send_prefix_def : 
    "Nondet_send_prefix f X Pf == ! :(f ` X) -> (%x. (Pf ((inv f) x)))"
  Rec_prefix_def : 
    "Rec_prefix f X Pf == ? :(f ` X) -> (%x. (Pf ((inv f) x)))"

syntax
  "@Nondet_send_prefix" :: 
     "('x => 'a) => pttrn => 'x set => 'a proc => 'a proc"
                               ("(1_ !? _:_ /-> _)" [900,900,1000,80] 80)

  "@Nondet_send_prefix_UNIV" :: 
     "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ !? _ /-> _)" [900,900,80] 80)

  "@Rec_prefix" :: 
     "('x => 'a) => pttrn => 'x set => 'a proc => 'a proc"
                               ("(1_ ? _:_ /-> _)" [900,900,1000,80] 80)
  "@Rec_Prefix_UNIV" :: 
     "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ ? _ /-> _)" [900,900,80] 80)

translations
  "a !? x:X -> P" == "Nondet_send_prefix a X (%x. P)"
  "a !? x -> P"   == "a !? x:UNIV -> P"

  "a ? x:X -> P"  == "Rec_prefix a X (%x. P)"
  "a ? x -> P"    == "a? x:UNIV -> P"

(*** internal choice ***)

consts
  Rep_int_choice_fun :: 
    "('x => 'a) => 'x set => ('x => 'a proc) => 'a proc"
defs
  Rep_int_choice_fun_def :
    "Rep_int_choice_fun f X Pf == ! :(f ` X) .. (%x. (Pf ((inv f) x)))"

syntax
  "@Rep_int_choice":: 
      "pttrn => (('a,nat) or_type) set => 'a proc => 'a proc"  
                               ("(1!! _:_ .. /_)" [900,900,68] 68)
  "@Rep_int_choice_UNIV":: 
      "pttrn => 'a proc => 'a proc"  
                               ("(1!! _ .. /_)" [900,68] 68)
  "@Rep_int_choice_OR1":: 
      "pttrn => 'a set => 'a proc => 'a proc"  
                               ("(1! _:_ .. /_)" [900,900,68] 68)
  "@Rep_int_choice_OR1_UNIV":: 
      "pttrn => 'a proc => 'a proc"  
                               ("(1! _ .. /_)" [900,68] 68)
  "@Rep_int_choice_OR2":: 
      "pttrn => nat set => 'a proc => 'a proc"  
                               ("(1!nat _:_ .. /_)" [900,900,68] 68)
  "@Rep_int_choice_OR2_UNIV":: 
      "pttrn => 'a proc => 'a proc"  
                               ("(1!nat _ .. /_)" [900,68] 68)
  "@Rep_int_choice_OR1_fun" :: 
      "('x => 'a) => pttrn => 'x set => 'a proc => 'a proc"
                               ("(1! _ _:_ .. /_)" [900,900,1000,80] 80)
  "@Rep_int_choice_OR1_fun_UNIV" ::
      "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1! _ _ .. /_)" [900,900,80] 80)

translations
  "!! x:X .. P"   == "!! :X .. (%x. P)"
  "!! x .. P"     == "!! x:UNIV .. P"
  "! x:X .. P"    == "! :X .. (%x. P)"
  "! x .. P"      == "! x:UNIV .. P"
  "!nat n:N .. P" == "!nat :N .. (%n. P)"
  "!nat n .. P"   == "!nat n:UNIV .. P"
  "! f x:X .. P"  == "Rep_int_choice_fun f X (%x. P)"
  "! f x .. P"    == "! f x:UNIV .. P"

(*** parallel ***)

syntax
  "_Interleave"    :: "'a proc => 'a proc
                    => 'a proc"  ("(1_ /||| _)" [76,77] 76)
  "_Synchro"       :: "'a proc => 'a proc
                    => 'a proc"  ("(1_ /|| _)" [76,77] 76)

translations
  "P ||| Q"     == "P |[{}]| Q"
  "P || Q"      == "P |[UNIV]| Q"

consts
  Alpha_parallel :: "'a proc => 'a set => 'a set => 'a proc
                      => 'a proc"  ("(1_ /(2|[_,/_]|)/ _)" [76,0,0,77] 76)
defs
  Alpha_parallel_def :
   "P |[X,Y]| Q  == (P |[- X]| SKIP) |[X Int Y]| (Q |[- Y]| SKIP)"

consts
  Rep_parallel :: "('a proc * 'a set) list
                   => 'a proc"  ("(1[||] _)" [77] 77)

primrec
  "[||] [] = SKIP"
  "[||] (PX # PXs) = (fst PX) |[snd PX, Union (snd ` set PXs)]| ([||] PXs)"

consts
  Index_parallel :: "'i set => ('i => ('a proc * 'a set))
                      => 'a proc"  ("(1[||]:_ /_)" [90,77] 77)

defs
  Index_parallel_def :
   "[||]:I PXf == [||] (map PXf (SOME Is. Is isListOf I))"

syntax
  "@Index_parallel" :: 
     "pttrn => 'i set => ('a proc * 'a set)
                      => 'a proc"  ("(1[||] _:_ .. /_)" [900,90,77] 77)

translations
  "[||] i:I .. PX"  == "[||]:I (%i. PX)"

(*** timeout ***)

syntax
  "_Timeout"  :: "'a proc => 'a proc 
               => 'a proc"  ("(1_ /[> _)" [73,74] 73)
translations
  "P [> Q"      == "(P |~| STOP) [+] Q"

(*------------------------------------*
 |                                    |
 |              X-Symbols             |
 |                                    |
 *------------------------------------*)

syntax (output)
  "_PrefixX"         :: "'a => 'a proc => 'a proc"
                                   ("(1_ /-> _)" [150,80] 80)
  "_Ext_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                                   ("(1? _:_ /-> _)" [900,900,80] 80)
  "_Ext_choiceX"     :: "'a proc => 'a proc => 'a proc"
                                   ("(1_ /[+] _)" [72,73] 72)
  "_Int_choiceX"     :: "'a proc => 'a proc => 'a proc"
                                   ("(1_ /|~| _)" [64,65] 64)
  "_Rep_int_choice_OR1X" :: "pttrn => 'a set => 'a proc 
                     => 'a proc"  ("(1! _:_ .. /_)" [900,900,68] 68)

  "_Int_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                                   ("(1! _:_ /-> _)" [900,900,80] 80)

  "_Rep_int_choice_OR1_funX" :: 
      "('x => 'a) => pttrn => 'x set => 'a proc => 'a proc"
                               ("(1! _ _:_ .. /_)" [900,900,1000,80] 80)
  "_Rep_int_choice_OR1_fun_UNIVX" ::
      "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1! _ _ .. /_)" [900,900,80] 80)

syntax (xsymbols)
  "_PrefixX"         :: "'a => 'a proc => 'a proc"
                          ("(1_ /\<rightarrow> _)" [150,80] 80)
  "_Ext_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                          ("(1? _:_ /\<rightarrow> _)" [900,900,80] 80)
  "_Ext_choiceX"     :: "'a proc => 'a proc => 'a proc"
                          ("(1_ /\<box> _)" [72,73] 72)
  "_Int_choiceX"     :: "'a proc => 'a proc => 'a proc"
                          ("(1_ /\<sqinter> _)" [64,65] 64)
  "_Rep_int_choice_OR1X" :: "pttrn => 'a set => 'a proc 
                     => 'a proc"  ("(1! _:_ \<bullet> /_)" [900,900,68] 68)
  "_Int_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                          ("(1! _:_ /\<rightarrow> _)" [900,900,80] 80)

  "_Rep_int_choice_OR1_funX" :: 
      "('x => 'a) => pttrn => 'x set => 'a proc => 'a proc"
                               ("(1! _ _:_ \<bullet> /_)" [900,900,1000,80] 80)
  "_Rep_int_choice_OR1_fun_UNIVX" ::
      "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1! _ _ \<bullet> /_)" [900,900,80] 80)

translations
  "a \<rightarrow> P"     == "a -> P"
  "? x:X \<rightarrow> P" == "? x:X -> P"
  "P \<box> Q"            == "P [+] Q"
  "P \<sqinter> Q"        == "P |~| Q"
  "! x:X \<bullet> P"     == "! x:X .. P"
  "! x:X \<rightarrow> P" == "! x:X -> P"
  "! f x:X \<bullet> P"   == "! f x:X .. P"
  "! f x \<bullet> P"     == "! f x:UNIV \<bullet> P"

syntax (output)
  "_Nondet_send_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                                  'a proc => 'a proc"
                                  ("(1_ !? _:_ /-> _)" [900,900,1000,80] 80)
  "_Nondet_send_prefix_UNIVX" :: "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ !? _ /-> _)" [900,900,80] 80)

  "_Send_prefixX"     :: "('x => 'a) => 'x => 'a proc => 'a proc" 
                          ("(1_ ! _ /-> _)" [900,1000,80] 80)

  "_Rec_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                          'a proc => 'a proc"
                          ("(1_ ? _:_ /-> _)" [900,900,1000,80] 80)
  "_Rec_prefix_UNIVX" :: "('x => 'a) => pttrn =>
                          'a proc => 'a proc"
                          ("(1_ ? _ /-> _)" [900,900,80] 80)

syntax (xsymbols)
  "_Nondet_send_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                                  'a proc => 'a proc"
                               ("(1_ !? _:_ /\<rightarrow> _)" [900,900,1000,80] 80)
  "_Nondet_send_prefix_UNIVX" :: "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ !? _ /\<rightarrow> _)" [900,900,80] 80)

  "_Send_prefixX"     :: "('x => 'a) => 'x => 'a proc => 'a proc" 
                          ("(1_ ! _ /\<rightarrow> _)" [900,1000,80] 80)

  "_Rec_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                          'a proc => 'a proc"
                          ("(1_ ? _:_ /\<rightarrow> _)" [900,900,1000,80] 80)
  "_Rec_prefix_UNIVX" :: "('x => 'a) => pttrn =>
                          'a proc => 'a proc"
                          ("(1_ ? _ /\<rightarrow> _)" [900,900,80] 80)

translations
  "a !? x:X \<rightarrow> P" == "a !? x:X -> P"
  "a !? x \<rightarrow> P"   == "a !? x:UNIV \<rightarrow> P"

  "a ! x \<rightarrow> P"    == "a ! x -> P"
  "a ? x:X \<rightarrow> P"  == "a ? x:X -> P"
  "a ? x \<rightarrow> P"    == "a ? x:UNIV \<rightarrow> P"

syntax (output)
  "_Proc_nameX" :: "'n => 'a proc" ("<_>" [0] 90)
  "_TimeoutX"   :: "'a proc => 'a proc 
                 => 'a proc"  ("(1_ /[> _)" [73,74] 73)
  "_HidingX"    :: "'a proc => 'a set => 'a proc"  
                                        ("(1_ /-- _)" [84,85] 84)
  "_RenamingX"  :: "'a proc => ('a * 'a) set => 'a proc"  
                                        ("(1_ /[[_]])" [84,0] 84)

syntax (xsymbols)
  "_Proc_nameX" :: "'n => 'a proc" ("\<langle>_\<rangle>" [0] 90)
  "_TimeoutX"   :: "'a proc => 'a proc 
                 => 'a proc"  ("(1_ /\<rhd> _)" [73,74] 73)
  "_HidingX"    :: "'a proc => 'a set => 'a proc"  
                                   ("(1_ /\<midarrow> _)" [84,85] 84)
  "_RenamingX"  :: "'a proc => ('a * 'a) set => 'a proc"  
                                        ("(1_ /\<lbrakk>_\<rbrakk>)" [84,0] 84)

translations
  "P \<rhd> Q"            == "(P |~| STOP) [+] Q"
  "P \<midarrow> X"       == "P -- X"
  "P \<lbrakk>r\<rbrakk>" == "P [[r]]"

(*********************************************************************
             recursive process (fixed point) and
         process expression (process with variable)
 *********************************************************************)

types ('pn,'a) procx = "('pn => 'a proc) => 'a proc"
types ('pn,'a) procX = "('pn => 'a proc) => ('pn => 'a proc)"

consts
 FIXn :: "nat => ('pn,'a) procX => ('pn => 'a proc)"  ("FIX[_] _" [0,1000] 55)
 FIX  :: "('pn,'a) procX => ('pn => 'a proc)"  ("FIX _" [1000] 55)

defs
 FIXn_def :  "FIX[n] PX == (PX^n) (%pm. DIV)"
 FIX_def  :     "FIX PX == (%pn. (!nat n .. ((FIX[n] PX) pn)))"

(*** Procx ***)

consts
  Procx :: "('pn,'a) procx set"
  ProcX :: "('pn,'a) procX set"

defs
  ProcX_def: "ProcX == {P. (ALL pn. (%X. P X pn) : Procx)}"

inductive "Procx"
intros
Procx_pn:
  "(%x. x pn) : Procx"
Procx_STOP: 
  "(%x. STOP) : Procx"
Procx_SKIP: 
  "(%x. SKIP) : Procx"
Procx_DIV:
  "(%x. DIV) : Procx"
Procx_Act_prefix:
  "P : Procx ==> (%x. a -> (P x)) : Procx"
Procx_Ext_pre_choice:
  "[| ALL a. Pf a : Procx |] ==> (%x. ? a:X -> (Pf a x)) : Procx"
Procx_Ext_choice:
  "[| P : Procx ; Q : Procx |] ==> (%x. (P x) [+] (Q x)) : Procx"
Procx_Int_choice:
  "[| P : Procx ; Q : Procx |] ==> (%x. (P x) |~| (Q x)) : Procx"
Procx_Rep_int_choice_OR:
  "[| ALL a. Pf a : Procx |] ==> (%x. !! a:X .. (Pf a x)) : Procx"
Procx_Conditional:
  "[| P : Procx ; Q : Procx |] ==> (%x. IF b THEN (P x) ELSE (Q x)) : Procx"
Procx_Parallel:
  "[| P : Procx ; Q : Procx |] ==> (%x. (P x) |[X]| (Q x)) : Procx"
Procx_Hiding:
  "P : Procx ==> (%x. (P x) -- X) : Procx"
Procx_Renaming:
  "P : Procx ==> (%x. (P x) [[r]]) : Procx"
Procx_Seq_compo:
  "[| P : Procx ; Q : Procx |] ==> (%x. (P x) ;; (Q x)) : Procx"

lemma Procx_Rep_int_choice_OR1:
  "[| ALL a. Pf a : Procx |] ==> (%x. ! a:X .. (Pf a x)) : Procx"
by (simp add: Procx_Rep_int_choice_OR Rep_int_choice_OR1_def)

lemma Procx_Rep_int_choice_OR2:
  "[| ALL a. Pf a : Procx |] ==> (%x. !nat a:X .. (Pf a x)) : Procx"
by (simp add: Procx_Rep_int_choice_OR Rep_int_choice_OR2_def)

lemmas Procx_Rep_int_choice = Procx_Rep_int_choice_OR
                              Procx_Rep_int_choice_OR1
                              Procx_Rep_int_choice_OR2

lemmas Procx_def = 
  Procx_pn Procx_STOP  Procx_SKIP        Procx_DIV
  Procx_Act_prefix   Procx_Ext_pre_choice   Procx_Ext_choice 
  Procx_Int_choice   Procx_Rep_int_choice   Procx_Conditional
  Procx_Parallel     Procx_Hiding           Procx_Renaming
  Procx_Seq_compo

(*------------------------------------------------------*
 |  "Procx.induct" is available.                        |
 |                                                      |
 |      apply (rule Procx.induct[of P PR])              |
 |                                                      |
 *------------------------------------------------------*)

lemma compo_in_Procx:
  "[| P : Procx ; Q : ProcX |] ==> (%x. P (Q x)) : Procx"
apply (rule Procx.induct[of P "(%P. (%x. P (Q x)) : Procx)"])
apply (simp)
apply (simp add: ProcX_def)
apply (simp_all add: Procx_def)
done

lemma compo_in_Procx_comp:
  "[| P : Procx ; Q : ProcX |] ==> P o Q : Procx"
by (simp add: comp_def compo_in_Procx)

lemma compo_in_Procx_ALL:
  "ALL P Q. (P : Procx & Q : ProcX ) --> (%x. P (Q x)) : Procx"
by (simp add: compo_in_Procx)

lemma compo_in_ProcX:
  "[| P : ProcX ; Q : ProcX |] ==> (%x. P (Q x)): ProcX"
apply (simp add: ProcX_def)
apply (intro allI)
apply (drule_tac x="pn" in spec)
apply (insert compo_in_Procx_ALL)
apply (drule_tac x="(%X. P X pn)" in spec)
apply (drule_tac x="Q" in spec)
apply (simp add: ProcX_def)
done

lemma compo_in_ProcX_comp:
  "[| P : ProcX ; Q : ProcX |] ==> P o Q : ProcX"
by (simp add: comp_def compo_in_ProcX)

lemma iteration_in_ProcX:
  "P : ProcX ==> P ^ n : ProcX"
apply (induct_tac n)
apply (simp add: ProcX_def)
apply (simp add: Procx_def)
apply (simp add: compo_in_ProcX)
done

lemma iteration_in_Procx:
  "P : ProcX ==> (%X. (P ^ n) X pn) : Procx"
apply (insert iteration_in_ProcX[of P n])
apply (simp)
apply (simp add: ProcX_def)
done

(*************************************************************
                   Syntactical functions
 *************************************************************)

consts
  guard :: "'a proc => bool"

primrec
  "guard STOP                 = True"
  "guard SKIP                 = True"
  "guard DIV                  = False"
  "guard (a -> P)             = True"
  "guard (? :X -> Pf)         = True"
  "guard (P [+] Q)            = (guard P & guard Q)"
  "guard (P |~| Q)            = (guard P & guard Q)"
  "guard (!! :X .. Pf)        = (ALL x. guard (Pf x))"
  "guard (IF b THEN P ELSE Q) = (guard P & guard Q)"
  "guard (P |[X]| Q)          = (guard P & guard Q)"
  "guard (P -- X)             = guard P"
  "guard (P [[r]])            = guard P"
  "guard (P ;; Q)             = (guard P & guard Q)"

lemma guard_Rep_int_choice_OR1[simp]:
  "guard (! :X .. Pf)         = (ALL x. guard (Pf x))"
apply (auto simp add: Rep_int_choice_OR1_def)
apply (drule_tac x="OR1 x" in spec)
by (simp)

lemma guard_Rep_int_choice_OR2[simp]:
  "guard (!nat :X .. Pf)         = (ALL x. guard (Pf x))"
apply (auto simp add: Rep_int_choice_OR2_def)
apply (drule_tac x="OR2 x" in spec)
by (simp)

(*** nohide (proc) ***)

consts
  nohide :: "'a proc => bool"

primrec
  "nohide STOP                 = True"
  "nohide SKIP                 = True"
  "nohide DIV                  = True"
  "nohide (a -> P)             = nohide P"
  "nohide (? :X -> Pf)         = (ALL x. nohide (Pf x))"
  "nohide (P [+] Q)            = (nohide P & nohide Q)"
  "nohide (P |~| Q)            = (nohide P & nohide Q)"
  "nohide (!! :X .. Pf)        = (ALL x. nohide (Pf x))"
  "nohide (IF b THEN P ELSE Q) = (nohide P & nohide Q)"
  "nohide (P |[X]| Q)          = (nohide P & nohide Q)"
  "nohide (P -- X)             = False"
  "nohide (P [[r]])            = nohide P"
  "nohide (P ;; Q)             = (nohide P & nohide Q)"

lemma nohide_Rep_int_choice_OR1[simp]:
  "nohide (! :X .. Pf)         = (ALL x. nohide (Pf x))"
apply (auto simp add: Rep_int_choice_OR1_def)
apply (drule_tac x="OR1 x" in spec)
by (simp)

lemma nohide_Rep_int_choice_OR2[simp]:
  "nohide (!nat :X .. Pf)         = (ALL x. nohide (Pf x))"
apply (auto simp add: Rep_int_choice_OR2_def)
apply (drule_tac x="OR2 x" in spec)
by (simp)

(*** guarded (rec_proc_fun) ***)

consts
  guardX  :: "('pn,'a) procX => bool"
  nohideX :: "('pn,'a) procX => bool"

defs 
  guardX_def : 
    "guardX f == (ALL pn. guard (f (%pm. DIV) pn))"

  nohideX_def : 
    "nohideX f == (ALL pn. nohide (f (%pm. DIV) pn))"

(*********************************************************
             termination relation for proc
 *********************************************************)

consts
  procterm :: "('a proc * 'a proc) set"

inductive procterm
intros
  "(P, a -> P)             : procterm"
  "(Pf a, ? :X -> Pf)      : procterm"
  "(P, P [+] Q)            : procterm"
  "(Q, P [+] Q)            : procterm"
  "(P, P |~| Q)            : procterm"
  "(Q, P |~| Q)            : procterm"
  "(Pf a, !! :X .. Pf)     : procterm"
  "(P, IF b THEN P ELSE Q) : procterm"
  "(Q, IF b THEN P ELSE Q) : procterm"
  "(P, P |[X]| Q)          : procterm"
  "(Q, P |[X]| Q)          : procterm"
  "(P, P -- X)             : procterm"
  "(P, P [[r]])            : procterm"
  "(P, P ;; Q)             : procterm"
  "(Q, P ;; Q)             : procterm"

lemma wf_procterm: "wf procterm"
  apply (unfold wf_def)
  apply (intro strip)
  apply (induct_tac x)
  apply (erule allE, erule mp, intro strip,
         erule procterm.elims,
         simp, simp, simp, simp, simp, simp,
         simp, simp, simp, simp, simp, simp,
         simp, simp, simp)+
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
