           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Conditional = CSP_semantics:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*********************************************************
                        Dom_T
 *********************************************************)

(*** Conditional_memT ***)

lemma Conditionalt_memT: 
  "(t :t IFt b THENt T ELSEt S)
 = (if b then t :t T else t :t S)"
by (simp add: evalt_def)

lemma Conditional_memT: 
  "(t :t [[IF b THEN P ELSE Q]]T)
 = (if b then t :t [[P]]T else t :t [[Q]]T)"
by (simp add: evalT_def evalt_def)

(*********************************************************
                        Dom_F
 *********************************************************)

(*** Conditional_memF ***)

lemma Conditionalf_memF: 
  "(f :f IFf b THENf F ELSEf E)
 = (if b then f :f F else f :f E)"
by (simp add: evalf_def)

lemma Conditional_memF: 
  "(f :f [[IF b THEN P ELSE Q]]F)
 = (if b then f :f [[P]]F else f :f [[Q]]F)"
by (simp add: evalF_def evalf_def)

lemmas Conditional_mem = Conditional_memT Conditional_memF
                         Conditionalt_memT Conditionalf_memF

(*******************************
             domSF
 *******************************)

(* T2 *)

lemma Conditional_T2 :
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_T2 (IFt b THENt T ELSEt S, IFf b THENf F ELSEf E)"
by (simp add: HC_T2_def Conditional_mem domSF_def)

(* F3 *)

lemma Conditional_F3 :
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_F3 (IFt b THENt T ELSEt S, IFf b THENf F ELSEf E)"
by (simp add: HC_F3_def Conditional_mem domSF_def)

(* T3_F4 *)

lemma Conditional_T3_F4 : 
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_T3_F4 (IFt b THENt T ELSEt S, IFf b THENf F ELSEf E)"
by (simp add: HC_T3_F4_def Conditional_mem domSF_iff)

(*** Conditional_domSF ***)

lemma Conditional_domSF : 
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> (IFt b THENt T ELSEt S, IFf b THENf F ELSEf E) : domSF"
apply (simp (no_asm) add: domSF_iff)
apply (simp add: Conditional_T2)
apply (simp add: Conditional_F3)
apply (simp add: Conditional_T3_F4)
done

(*********************************************************
                      mono
 *********************************************************)

(*** T ***)

lemma Conditional_evalT_mono:
  "[| T1 <= T2 ; S1 <= S2 |]
   ==> IFt b THENt T1 ELSEt S1 <= IFt b THENt T2 ELSEt S2"
apply (simp add: subsetT_iff)
apply (intro allI impI)
by (simp add: Conditional_mem)

(*** F ***)

lemma Conditional_evalF_mono:
  "[| F1 <= F2 ; E1 <= E2 |]
   ==> IFf b THENf F1 ELSEf E1 <= IFf b THENf F2 ELSEf E2"
apply (simp add: subsetF_iff)
apply (intro allI impI)
by (simp add: Conditional_mem)

(* cong *)

lemma Conditional_evalT_cong:
  "[| T1 = T2 ; S1 = S2 |]
   ==> IFt b THENt T1 ELSEt S1 = IFt b THENt T2 ELSEt S2"
by (simp)

lemma Conditional_evalF_cong:
  "[| F1 = F2 ; E1 = E2 |]
   ==> IFf b THENf F1 ELSEf E1 = IFf b THENf F2 ELSEf E2"
by (simp)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
