           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Conditional_cms = Conditional + Domain_SF_prod_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. IFt b THENt (T x) ELSEt (S x)) : map
         2. (%x. IFf b THENf (F x) ELSEf (E x)) : map
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Conditional T
 *********************************************************)

(*** restT (subset) ***)

lemma Conditional_restT_sub:
   "[| T x1 rest n <= T x2 rest n ;
       S x1 rest n <= S x2 rest n |]
    ==> (IFt b THENt (T x1) ELSEt (S x1)) rest n <= 
        (IFt b THENt (T x2) ELSEt (S x2)) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Conditional_mem)
done

(*** restT (equal) ***)

lemma Conditional_restT:
   "[| T x1 rest n = T x2 rest n ;
       S x1 rest n = S x2 rest n |]
    ==> (IFt b THENt (T x1) ELSEt (S x1)) rest n = 
        (IFt b THENt (T x2) ELSEt (S x2)) rest n"
apply (rule order_antisym)
by (simp_all add: Conditional_restT_sub)

(*** distT lemma ***)

lemma Conditional_distT:
  "TTs = {(T x1, T x2), (S x1, S x2)}
    ==> (EX TT. TT:TTs & 
             distance(IFt b THENt (T x1) ELSEt (S x1), 
                      IFt b THENt (T x2) ELSEt (S x2))
          <= distance((fst TT), (snd TT)))"
apply (rule rest_to_dist_pair)
by (auto intro: Conditional_restT)

(*** map_alpha T lemma ***)

lemma Conditional_evalT_map_alpha_lm:
  "[| distance (T x1, T x2) <= alpha * distance (x1, x2) ;
      distance (S x1, S x2) <= alpha * distance (x1, x2) |]
    ==> distance(IFt b THENt (T x1) ELSEt (S x1), 
                 IFt b THENt (T x2) ELSEt (S x2))
     <= alpha * distance (x1, x2)"
apply (insert Conditional_distT
       [of "{(T x1, T x2), (S x1, S x2)}" T x1 x2 S b])
by (auto)

(*** Conditional_evalT_non_expanding ***)

lemma Conditional_evalT_map_alpha:
 "[| map_alpha T alpha ; map_alpha S alpha |]
  ==> map_alpha (%x. IFt b THENt (T x) ELSEt (S x)) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Conditional_evalT_map_alpha_lm)

(*** Conditional_evalT_non_expanding ***)

lemma Conditional_evalT_non_expanding:
 "[| non_expanding T ; non_expanding S |]
  ==> non_expanding (%x. IFt b THENt (T x) ELSEt (S x))"
by (simp add: non_expanding_def Conditional_evalT_map_alpha)

(*** Conditional_evalT_contraction_alpha ***)

lemma Conditional_evalT_contraction_alpha:
 "[| contraction_alpha T alpha; contraction_alpha S alpha|]
  ==> contraction_alpha (%x. IFt b THENt (T x) ELSEt (S x)) alpha"
by (simp add: contraction_alpha_def Conditional_evalT_map_alpha)

(*********************************************************
                   map Conditional F
 *********************************************************)

(*** restF (subset) ***)

lemma Conditional_restF_sub:
   "[| F x1 rest n <= F x2 rest n ;
       E x1 rest n <= E x2 rest n |]
    ==> (IFf b THENf (F x1) ELSEf (E x1)) rest n <= 
        (IFf b THENf (F x2) ELSEf (E x2)) rest n"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_restF)
apply (simp add: Conditional_mem)
done

(*** restF (equal) ***)

lemma Conditional_restF:
   "[| F x1 rest n = F x2 rest n ;
       E x1 rest n = E x2 rest n |]
    ==> (IFf b THENf (F x1) ELSEf (E x1)) rest n = 
        (IFf b THENf (F x2) ELSEf (E x2)) rest n"
apply (rule order_antisym)
by (simp_all add: Conditional_restF_sub)

(*** distF lemma ***)

lemma Conditional_distF:
  "FFs = {(F x1, F x2), (E x1, E x2)}
    ==> (EX FF. FF:FFs & 
             distance(IFf b THENf (F x1) ELSEf (E x1), 
                      IFf b THENf (F x2) ELSEf (E x2))
          <= distance((fst FF), (snd FF)))"
apply (rule rest_to_dist_pair)
by (auto intro: Conditional_restF)

(*** map_alpha F lemma ***)

lemma Conditional_evalF_map_alpha_lm:
  "[| distance (F x1, F x2) <= alpha * distance (x1, x2) ;
      distance (E x1, E x2) <= alpha * distance (x1, x2) |]
    ==> distance(IFf b THENf (F x1) ELSEf (E x1), 
                 IFf b THENf (F x2) ELSEf (E x2))
     <= alpha * distance (x1, x2)"
apply (insert Conditional_distF
       [of "{(F x1, F x2), (E x1, E x2)}" F x1 x2 E b])
by (auto)

(*** Conditional_evalF_non_expanding ***)

lemma Conditional_evalF_map_alpha:
 "[| map_alpha F alpha ; map_alpha E alpha |]
  ==> map_alpha (%x. IFf b THENf (F x) ELSEf (E x)) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Conditional_evalF_map_alpha_lm)

(*** Conditional_evalF_non_expanding ***)

lemma Conditional_evalF_non_expanding:
 "[| non_expanding F ; non_expanding E |]
  ==> non_expanding (%x. IFf b THENf (F x) ELSEf (E x))"
by (simp add: non_expanding_def Conditional_evalF_map_alpha)

(*** Conditional_evalF_contraction_alpha ***)

lemma Conditional_evalF_contraction_alpha:
 "[| contraction_alpha F alpha; contraction_alpha E alpha|]
  ==> contraction_alpha (%x. IFf b THENf (F x) ELSEf (E x)) alpha"
by (simp add: contraction_alpha_def Conditional_evalF_map_alpha)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
