           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Ext_choice = CSP_semantics:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*********************************************************
                        Dom_T
 *********************************************************)

(*** Ext_choice_memT ***)

lemma Ext_choicet_memT: 
  "(t :t T [+]t S) = (t :t T | t :t S)"
by (simp add: evalt_def evalT_def memT_UnT)

lemma Ext_choice_memT: 
  "(t :t [[P [+] Q]]T) = (t :t [[P]]T | t :t [[Q]]T)"
by (simp add: Ext_choicet_memT evalT_def)

(*********************************************************
                        Dom_F
 *********************************************************)

(*** Ext_choice_domF ***)

lemma Ext_choice_domF: 
  "{f. (EX   X. f = ([]t,X) & f :f F IntF E) |
       (EX s X. f = (s,X)   & f :f F UnF  E & 
                s ~= []t) |
       (EX   X. f = ([]t,X) & [Tick]t :t T UnT S &
                X <= Evset)} : domF"
apply (simp add: domF_def HC_F2_def)
apply (intro allI impI)
apply (elim conjE disjE)

(* part1 *)
apply (simp add: memF_IntF)
apply (rule disjI1)
apply (rule conjI)
apply (rule memF_F2, simp_all)
apply (rule memF_F2, simp_all)

(* part2 *)
apply (simp add: memF_UnF)
apply (erule disjE)
apply (rule disjI1)
apply (rule memF_F2, simp_all)
apply (rule disjI2)
apply (rule memF_F2, simp_all)

(* part3 *)
apply (fast)
done

(*** Ext_choice_memT ***)

lemma Ext_choicef_memF: 
  "(f :f (T,F) [+]f (S,E)) =
   ((EX   X. f = ([]t,X) & f :f F IntF E) |
    (EX s X. f = (s,X)   & f :f F UnF  E & s ~= []t) |
    (EX   X. f = ([]t,X) & [Tick]t :t T UnT S & X <= Evset))"
apply (simp only: evalf_def evalF_def)
apply (simp only: memF_def Abs_domF_inverse Ext_choice_domF[simplified memF_def])
by (simp)

lemma Ext_choice_memF: 
  "(f :f [[P [+] Q]]F) =
   ((EX   X. f = ([]t,X) & f :f [[P]]F IntF [[Q]]F) |
    (EX s X. f = (s,X)   & f :f [[P]]F UnF [[Q]]F & s ~= []t) |
    (EX   X. f = ([]t,X) & [Tick]t :t [[P]]T UnT [[Q]]T & X <= Evset))"
by (simp add: evalF_def Ext_choicef_memF)

lemmas Ext_choice_mem = Ext_choicet_memT Ext_choicef_memF
                        Ext_choice_memT  Ext_choice_memF

(*******************************
             domSF
 *******************************)

(* T2 *)

lemma Ext_choice_T2 :
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_T2 (T [+]t S, (T,F) [+]f (S,E))"
apply (simp add: HC_T2_def Ext_choice_mem)
apply (intro allI impI)
apply (elim conjE exE)
apply (simp add: memF_UnF)

apply (simp add: domSF_def HC_T2_def)
apply (elim conjE)
apply (drule_tac x="s" in spec)
apply (drule_tac x="s" in spec)
by (fast)

(* F3 *)

lemma Ext_choice_F3 :
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_F3 (T [+]t S, (T,F) [+]f (S,E))"
apply (simp add: HC_F3_def Ext_choice_mem)
apply (intro allI impI)

apply (elim conjE disjE)

(* part1 *)
apply (simp add: memF_IntF)
apply (simp add: domSF_def HC_F3_def)

(* part2 *)
apply (simp add: memF_UnF)
apply (simp add: domSF_def HC_F3_def)
apply (force)

(* part3 *)
apply (simp add: memT_UnT memF_IntF)
apply (case_tac "Tick ~: Y")
 apply (simp add: Evset_def)
 apply (force)

 (* Tick : Y *)
 apply (drule_tac x="Tick" in spec)
 apply (simp)
done

(* T3_F4 *)

lemma Ext_choice_T3_F4 : 
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_T3_F4 (T [+]t S, (T,F) [+]f (S,E))"
apply (simp add: HC_T3_F4_def Ext_choice_mem)
apply (intro allI impI)
apply (elim conjE exE)

apply (simp add: domSF_iff HC_T3_F4_def)
apply (elim conjE)
apply (drule_tac x="s" in spec)
apply (drule_tac x="s" in spec)
apply (simp add: memF_UnF memT_UnT)
by (auto)

(*** Ext_choice_domSF ***)

lemma Ext_choice_domSF : 
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> (T [+]t S, (T,F) [+]f (S,E)) : domSF"
apply (simp (no_asm) add: domSF_iff)
apply (simp add: Ext_choice_T2)
apply (simp add: Ext_choice_F3)
apply (simp add: Ext_choice_T3_F4)
done

(*********************************************************
                      mono
 *********************************************************)

(*** T ***)

lemma Ext_choice_evalT_mono:
  "[| T1 <= T2 ; S1 <= S2 |]
   ==> T1 [+]t S1 <= T2 [+]t S2"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: Ext_choice_mem)
apply (erule disjE)
by (simp_all)

(*** F ***)

lemma Ext_choice_evalF_mono:
  "[| T1 <= T2 ; S1 <= S2 ; 
      F1 <= F2 ; E1 <= E2 |]
   ==> (T1,F1) [+]f (S1,E1) <= (T2,F2) [+]f (S2,E2)"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: Ext_choice_mem)

 apply (erule disjE)
 apply (simp add: memF_IntF)
 apply (force)

 apply (erule disjE)
 apply (simp add: memF_UnF)
 apply (force)

 apply (simp add: memT_UnT memF_IntF)
 apply (simp add: subsetT_iff)
 apply (elim conjE disjE)
 apply (auto)
done

(*** cong ***)

lemma Ext_choice_evalT_cong:
  "[| T1 = T2 ; S1 = S2 |]
   ==> T1 [+]t S1 = T2 [+]t S2"
by (simp)

lemma Ext_choice_evalF_cong:
  "[| T1 = T2 ; S1 = S2 ; F1 = F2 ; E1 = E2 |]
   ==> (T1,F1) [+]f (S1,E1) = (T2,F2) [+]f (S2,E2)"
by (simp)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
