           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Ext_choice_cms = Ext_choice + Domain_SF_prod_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. T x [+]t S x) : non expanding
         2. (%x. (T x,F x) [+]f (S x,E x) : non expanding
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Ext_choice T
 *********************************************************)

(*** restT (subset) ***)

lemma Ext_choice_restT_sub:
   "[| T x1 rest n <= T x2 rest n ;
       S x1 rest n <= S x2 rest n |]
    ==> (T x1 [+]t S x1) rest n <= (T x2 [+]t S x2) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Ext_choice_mem)
by (auto)

(*** restT (equal) ***)

lemma Ext_choice_restT:
   "[| T x1 rest n = T x2 rest n ;
       S x1 rest n = S x2 rest n |]
    ==> (T x1 [+]t S x1) rest n = (T x2 [+]t S x2) rest n"
apply (rule order_antisym)
by (simp_all add: Ext_choice_restT_sub)

(*** distT lemma ***)

lemma Ext_choice_distT:
"TTs = {(T x1, T x2), (S x1, S x2)}
 ==> (EX TT. TT:TTs & 
             distance((T x1 [+]t S x1), (T x2 [+]t S x2))
          <= distance((fst TT), (snd TT)))"
apply (rule rest_to_dist_pair)
by (auto intro: Ext_choice_restT)

(*** map_alpha T lemma ***)

lemma Ext_choice_evalT_map_alpha_lm:
  "[| distance (T x1, T x2) <= alpha * distance (x1, x2) ;
      distance (S x1, S x2) <= alpha * distance (x1, x2) |]
    ==> distance (T x1 [+]t S x1, T x2 [+]t S x2)
     <= alpha * distance (x1, x2)"
apply (insert Ext_choice_distT
       [of "{(T x1, T x2), (S x1, S x2)}" T x1 x2 S])
by (auto)

(*** Ext_choice_evalT_map_alpha ***)

lemma Ext_choice_evalT_map_alpha:
 "[| map_alpha T alpha ; map_alpha S alpha |]
  ==> map_alpha (%x. T x [+]t S x) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Ext_choice_evalT_map_alpha_lm)

(*** Ext_choice_evalT_non_expanding ***)

lemma Ext_choice_evalT_non_expanding:
 "[| non_expanding T ; non_expanding S |]
  ==> non_expanding (%x. T x [+]t S x)"
by (simp add: non_expanding_def Ext_choice_evalT_map_alpha)

(*** Ext_choice_evalT_contraction_alpha ***)

lemma Ext_choice_evalT_contraction_alpha:
 "[| contraction_alpha T alpha; contraction_alpha S alpha|]
  ==> contraction_alpha (%x. T x [+]t S x) alpha"
by (simp add: contraction_alpha_def Ext_choice_evalT_map_alpha)

(*********************************************************
                   map Ext_choice F
 *********************************************************)

(*** restF (subset) ***)

lemma Ext_choice_restF_sub:
   "[| T x1 rest n <= T x2 rest n ;
       S x1 rest n <= S x2 rest n ;
       F x1 rest n <= F x2 rest n ;
       E x1 rest n <= E x2 rest n |]
    ==> ((T x1, F x1) [+]f (S x1, E x1)) rest n 
     <= ((T x2, F x2) [+]f (S x2, E x2)) rest n"
apply (simp add: subsetF_iff subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restF in_restT)
apply (simp add: Ext_choice_mem)
apply (elim conjE exE disjE, simp_all)

 apply (simp add: memF_IntF)
 apply (simp add: memF_UnF)
 apply (erule disjE, simp_all)
 apply (simp add: memT_UnT)
 apply (erule disjE, simp_all)

 apply (simp add: memF_UnF)
 apply (rotate_tac 2)
 apply (drule_tac x="s' @t [Tick]t" in spec)
 apply (drule_tac x="s' @t [Tick]t" in spec)
 apply (drule_tac x="X" in spec)
 apply (drule_tac x="X" in spec)
 apply (fast)
done

(*** restF (equal) ***)

lemma Ext_choice_restF:
   "[| T x1 rest n = T x2 rest n ;
       S x1 rest n = S x2 rest n ;
       F x1 rest n = F x2 rest n ;
       E x1 rest n = E x2 rest n |]
    ==> ((T x1, F x1) [+]f (S x1, E x1)) rest n 
      = ((T x2, F x2) [+]f (S x2, E x2)) rest n"
apply (rule order_antisym)
by (simp_all add: Ext_choice_restF_sub)

(*** distF lemma ***)

lemma Ext_choice_distF:
  "[| TTs = {(T x1, T x2), (S x1, S x2)} ;
      FFs = {(F x1, F x2), (E x1, E x2)} |]
   ==> (EX TT. TT:TTs & 
               distance((T x1, F x1) [+]f (S x1, E x1), (T x2, F x2) [+]f (S x2, E x2))
            <= distance((fst TT), (snd TT))) |
       (EX FF. FF:FFs & 
               distance((T x1, F x1) [+]f (S x1, E x1), (T x2, F x2) [+]f (S x2, E x2))
            <= distance((fst FF), (snd FF)))"
apply (rule rest_to_dist_pair_two)
apply (simp_all)
by (auto intro: Ext_choice_restF)

(*** map_alpha F lemma ***)

lemma Ext_choice_evalF_map_alpha_lm:
  "[| distance (T x1, T x2) <= alpha * distance (x1, x2) ;
      distance (S x1, S x2) <= alpha * distance (x1, x2) ;
      distance (F x1, F x2) <= alpha * distance (x1, x2) ;
      distance (E x1, E x2) <= alpha * distance (x1, x2) |]
    ==> distance((T x1, F x1) [+]f (S x1, E x1), (T x2, F x2) [+]f (S x2, E x2))
     <= alpha * distance (x1, x2)"
apply (insert Ext_choice_distF
       [of "{(T x1, T x2), (S x1, S x2)}" T x1 x2 S
           "{(F x1, F x2), (E x1, E x2)}" F E])
by (auto)

(*** Ext_choice_evalF_map_alpha ***)

lemma Ext_choice_evalF_map_alpha:
 "[| map_alpha T alpha ; map_alpha S alpha ;
     map_alpha F alpha ; map_alpha E alpha |]
  ==> map_alpha (%x. (T x, F x) [+]f (S x, E x)) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Ext_choice_evalF_map_alpha_lm)

lemma Ext_choice_evalF_non_expanding:
 "[| non_expanding T ; non_expanding S ;
     non_expanding F ; non_expanding E |]
  ==> non_expanding (%x. (T x, F x) [+]f (S x, E x))"
by (simp add: non_expanding_def Ext_choice_evalF_map_alpha)

(*** Ext_choice_evalF_contraction_alpha ***)

lemma Ext_choice_evalF_contraction_alpha:
 "[| contraction_alpha T alpha; contraction_alpha S alpha ;
     contraction_alpha F alpha; contraction_alpha E alpha|]
  ==> contraction_alpha (%x. (T x, F x) [+]f (S x, E x)) alpha"
by (simp add: contraction_alpha_def Ext_choice_evalF_map_alpha)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
