           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Ext_pre_choice_cms = Ext_pre_choice + Act_prefix_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. ? a:X ->t Tf a x) : contraction map
         2. (%x. ? a:X ->f Ff a x) : contraction map
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Ext_pre_choice T
 *********************************************************)

(*** restT (subset) ***)

lemma Ext_pre_choice_Act_prefix_restT_sub:
   "[| ALL a : X.
         a ->t (Tf a x1) rest n <= a ->t (Tf a x2) rest n |]
    ==> (? :X ->t (%a. (Tf a x1))) rest n <=
        (? :X ->t (%a. (Tf a x2))) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Ext_pre_choice_mem)
apply (elim conjE exE disjE)
apply (simp_all)

apply (drule_tac x="a" in bspec, simp)
apply (drule_tac x="t" in spec)
apply (simp add: Act_prefix_mem)
apply (drule mp)
 apply (rule_tac x="s" in exI)
 apply (simp)
apply (elim conjE exE)
apply (rule_tac x="a" in exI)
apply (rule_tac x="s" in exI)
apply (simp add: not_None_T)
done

(*** restT (equal) ***)

lemma Ext_pre_choice_Act_prefix_restT:
   "[| ALL a : X.
         a ->t (Tf a x1) rest n = a ->t (Tf a x2) rest n |]
    ==> (? :X ->t (%a. (Tf a x1))) rest n =
        (? :X ->t (%a. (Tf a x2))) rest n"
apply (rule order_antisym)
by (simp_all add: Ext_pre_choice_Act_prefix_restT_sub)

(*** distT lemma ***)

lemma Ext_pre_choice_Act_prefix_distT_nonempty:
"[| X ~= {} ;
    TTs = {(a ->t (Tf a x1), a ->t (Tf a x2))|a. a : X} |]
 ==> (EX TT. TT:TTs & 
             distance((? :X ->t (%a. (Tf a x1))), (? :X ->t (%a. (Tf a x2))))
          <= distance((fst TT), (snd TT)))"
apply (rule rest_to_dist_pair)
apply (force)

apply (intro allI impI)
apply (rule Ext_pre_choice_Act_prefix_restT)
apply (rule ballI)
apply (simp)
apply (drule_tac x="a ->t (Tf a x1)" in spec)
apply (drule_tac x="a ->t (Tf a x2)" in spec)
by (auto)

(*** contractionT lemma ***)

lemma Ext_pre_choice_evalT_contraction_half_nonempty_lm:
  "[| X ~= {}; ALL a. distance (Tf a x1, Tf a x2)
                   <= distance (x1, x2) |]
    ==> distance (? :X ->t (%a. (Tf a x1)), ? :X ->t (%a. (Tf a x2))) * 2 
     <= distance (x1, x2)"
apply (insert Ext_pre_choice_Act_prefix_distT_nonempty
       [of X "{(a ->t (Tf a x1), a ->t (Tf a x2)) |a. a : X}" Tf x1 x2])
apply (simp)
apply (elim conjE exE)
apply (simp)
apply (drule_tac x="aa" in spec)
apply (subgoal_tac 
    "distance (aa ->t (Tf aa x1) , aa ->t (Tf aa x2)) * 2
   = distance (Tf aa x1, Tf aa x2)")
apply (simp)
by (simp add: Act_prefix_evalT_contraction_half_lm)

(*** Ext_pre_choice_evalT_contraction_half ***)

lemma Ext_pre_choice_evalT_contraction_half: 
 "ALL a. non_expanding (Tf a)
  ==> contraction_alpha (%x. (? :X ->t (%a. (Tf a x)))) (1 / 2)"
apply (simp add: contraction_alpha_def non_expanding_def map_alpha_def)
apply (case_tac "X = {}")
apply (simp add: evalt_def)
by (simp add: Ext_pre_choice_evalT_contraction_half_nonempty_lm)

(*** Ext_pre_choice_evalT_contraction ***)

lemma Ext_pre_choice_evalT_contraction: 
 "ALL a. non_expanding (Tf a)
  ==> contraction (%x. (? :X ->t (%a. (Tf a x))))"
apply (simp add: contraction_def)
apply (rule_tac x="1/2" in exI)
by (simp add: Ext_pre_choice_evalT_contraction_half)

(*** Ext_pre_choice_evalT_non_expanding ***)

lemma Ext_pre_choice_evalT_non_expanding:
 "ALL a. non_expanding (Tf a) ==> non_expanding (%x. (? :X ->t (%a. (Tf a x))))"
apply (rule contraction_non_expanding)
by (simp add: Ext_pre_choice_evalT_contraction)

(*********************************************************
                   map Ext_pre_choice F
 *********************************************************)

(*** restF (subset) ***)

lemma Ext_pre_choice_Act_prefix_restF_sub:
   "[| ALL a : X.
         a ->f (Ff a x1) rest n <= a ->f (Ff a x2) rest n |]
    ==> (? :X ->f (%a. (Ff a x1))) rest n <=
        (? :X ->f (%a. (Ff a x2))) rest n"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_restF)
apply (simp add: Ext_pre_choice_mem)
apply (elim conjE exE disjE)
apply (simp_all)

apply (drule_tac x="a" in bspec, simp)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Xa" in spec)
apply (simp add: Act_prefix_mem)
apply (drule mp)
 apply (rule_tac x="sa" in exI)
 apply (simp)
apply (elim conjE exE)
apply (rule_tac x="a" in exI)
apply (rule_tac x="sa" in exI)
apply (simp add: not_None_F)

apply (drule_tac x="a" in bspec, simp)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Xa" in spec)
apply (simp add: Act_prefix_mem)
apply (drule mp)
 apply (rule conjI)
 apply (rule_tac x="sa" in exI, simp)
 apply (rule_tac x="s'" in exI, simp)
apply (elim conjE exE)
apply (rule_tac x="a" in exI)
apply (rule_tac x="sa" in exI)
apply (simp add: not_None_F)
done

(*** restF (equal) ***)

lemma Ext_pre_choice_Act_prefix_restF:
   "[| ALL a : X.
         a ->f (Ff a x1) rest n = a ->f (Ff a x2) rest n |]
    ==> (? :X ->f (%a. (Ff a x1))) rest n =
        (? :X ->f (%a. (Ff a x2))) rest n"
apply (rule order_antisym)
by (simp_all add: Ext_pre_choice_Act_prefix_restF_sub)

(*** distF lemma ***)

lemma Ext_pre_choice_Act_prefix_distF_nonempty:
"[| X ~= {} ;
    FFs = {(a ->f (Ff a x1), a ->f (Ff a x2))|a. a : X} |]
 ==> (EX FF. FF:FFs & 
             distance((? :X ->f (%a. (Ff a x1))), (? :X ->f (%a. (Ff a x2))))
          <= distance((fst FF), (snd FF)))"
apply (rule rest_to_dist_pair)
apply (force)

apply (intro allI impI)
apply (rule Ext_pre_choice_Act_prefix_restF)
apply (rule ballI)
apply (simp)
apply (drule_tac x="a ->f (Ff a x1)" in spec)
apply (drule_tac x="a ->f (Ff a x2)" in spec)
by (auto)

(*** contractionF lemma ***)

lemma Ext_pre_choice_evalF_contraction_half_nonempty_lm:
  "[| X ~= {}; ALL a. distance (Ff a x1, Ff a x2)
                   <= distance (x1, x2) |]
    ==> distance (? :X ->f (%a. (Ff a x1)), ? :X ->f (%a. (Ff a x2))) * 2 
     <= distance (x1, x2)"
apply (insert Ext_pre_choice_Act_prefix_distF_nonempty
       [of X "{(a ->f (Ff a x1), a ->f (Ff a x2)) |a. a : X}" Ff x1 x2])
apply (simp)
apply (elim conjE exE, simp)
apply (drule_tac x="aa" in spec)
apply (subgoal_tac 
    "distance (aa ->f (Ff aa x1) , aa ->f (Ff aa x2)) * 2
   = distance (Ff aa x1, Ff aa x2)")
apply (simp)
by (simp add: Act_prefix_evalF_contraction_half_lm)

(*** Ext_pre_choice_evalF_contraction_half ***)

lemma Ext_pre_choice_evalF_contraction_half: 
 "ALL a. non_expanding (Ff a)
  ==> contraction_alpha (%x. (? :X ->f (%a. (Ff a x)))) (1 / 2)"
apply (simp add: contraction_alpha_def non_expanding_def map_alpha_def)
apply (case_tac "X = {}")
apply (simp add: evalf_def)
by (simp add: Ext_pre_choice_evalF_contraction_half_nonempty_lm)

(*** Ext_pre_choice_evalF_contraction ***)

lemma Ext_pre_choice_evalF_contraction: 
 "ALL a. non_expanding (Ff a)
  ==> contraction (%x. (? :X ->f (%a. (Ff a x))))"
apply (simp add: contraction_def)
apply (rule_tac x="1/2" in exI)
by (simp add: Ext_pre_choice_evalF_contraction_half)

(*** Ext_pre_choice_evalF_non_expanding ***)

lemma Ext_pre_choice_evalF_non_expanding:
 "ALL a. non_expanding (Ff a) ==> non_expanding (%x. (? :X ->f (%a. (Ff a x))))"
apply (rule contraction_non_expanding)
by (simp add: Ext_pre_choice_evalF_contraction)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
