           (*-------------------------------------------*
            |                CSP-Prover                 |
            |                 May 2005                  |
            |                June 2005  (modified)      |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Index_parallel = Rep_parallel:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(**************************************************************
                         lemmas
 **************************************************************)

(*------------------------*
 |           T            |
 *------------------------*)

(*** mem_index_style ***)

lemma to_index_style_T:
   "(ALL P X. (P,X):(PXf ` I) --> (u --tr (- X)) :t [[P]]T)
  = (ALL i:I. u --tr (- snd (PXf i)) :t [[fst (PXf i)]]T)"
apply (auto)
apply (simp add: pair_eq_decompo)
done

(*** Index_parallel_memT (pre) ***)

lemma Index_parallel_memT_pre:
  "[| I ~= {} ; finite I |]
   ==> (u :t [[ [||]:I PXf ]]T) = 
       (sett(u) <= insert Tick (Ev ` (Union (snd ` PXf ` I))) & 
        (ALL P X. (P,X):PXf ` I --> (u --tr (- X)) :t [[P]]T))"
apply (simp add: Index_parallel_def)
apply (subgoal_tac "EX Is. Is isListOf I")
apply (elim conjE exE)
apply (subgoal_tac "(map PXf (SOME Is. Is isListOf I)) ~= []")
apply (simp add: Rep_parallel_memT)
apply (rule someI2)
 apply (simp)
 apply (simp add: isListOf_set_eq)

apply (rule someI2)
 apply (simp)
 apply (simp add: isListOf_nonemptyset)

apply (simp add: isListOf_EX)
done

(*** Index_parallel_memT ***)

lemma Index_parallel_memT:
  "[| I ~= {} ; finite I |]
   ==> (u :t [[ [||]:I PXf ]]T) = 
       (sett(u) <= insert Tick (Ev ` (Union (snd ` PXf ` I))) & 
        (ALL i:I. (u --tr (- (snd (PXf i)))) :t [[fst (PXf i)]]T))"
apply (simp add: Index_parallel_memT_pre)
apply (simp add: to_index_style_T)
done

(*** Semantics for indexed alphabetized parallel on T ***)

lemma Index_parallel_evalT:
  "[| I ~= {} ; finite I |]
   ==> [[ [||]:I PXf ]]T = 
             {u. (sett(u) <= insert Tick (Ev ` (Union (snd ` PXf ` I))) & 
              (ALL i:I. (u --tr (- (snd (PXf i)))) :t [[fst (PXf i)]]T))}t"
apply (simp add: Index_parallel_memT[THEN sym])
apply (simp add: CollectT_open)
done

(*------------------------*
 |           F            |
 *------------------------*)

(*** lemmas for memF ***)

lemma Index_parallel_memF_lm1:
  "[| Is isListOf I ; length Ys = length Is |] ==>
    Union {Ys ! i Int insert Tick (Ev ` snd (map PXf Is ! i)) |i. i < length Ys} =
    Union {Ys ! (THE n. Is ! n = i & n < length Is) Int
           insert Tick (Ev ` snd (PXf i)) | i. i : I}"
apply (rule)

 (* <= *)
 apply (rule)
 apply (simp)
 apply (elim conjE exE)
 apply (rule_tac x=
   "Ys ! (THE n. Is ! n = (Is ! i) & n < length Is) Int
    insert Tick (Ev ` snd (PXf (Is ! i)))" in exI)
 apply (simp add: isListOf_THE_nth)
 apply (rule_tac x="(Is ! i)" in exI)
 apply (simp add: isListOf_THE_nth)
 apply (simp add: isListOf_nth_in_index)

 (* => *)
 apply (rule)
 apply (simp)
 apply (elim conjE exE)
 apply (simp)
 apply (erule isListOf_index_to_nthE)
 apply (drule_tac x="i" in bspec, simp)
 apply (elim conjE exE, simp)
 apply (simp add: isListOf_THE_nth)
 apply (rule_tac x=
   "Ys ! n Int insert Tick (Ev ` snd (PXf (Is ! n)))" in exI)
 apply (simp)
 apply (rule_tac x="n" in exI)
 apply (simp)
done

lemma Index_parallel_memF_lm2:
  "Is isListOf I ==>
   Union {Yf i Int insert Tick (Ev ` snd (PXf i)) |i. i : I} =
   Union {map Yf Is ! i Int insert Tick (Ev ` snd (map PXf Is ! i)) |i. i < length Is}"
apply (rule)

 (* <= *)
 apply (rule)
 apply (simp)
 apply (elim conjE exE)
 apply (erule isListOf_index_to_nthE)
 apply (drule_tac x="i" in bspec, simp)
 apply (elim conjE exE, simp)
 apply (rule_tac x=
   "map Yf Is ! n Int insert Tick (Ev ` snd (map PXf Is ! n))" in exI)
 apply (simp)
 apply (rule_tac x="n" in exI)
 apply (simp)

 (* => *)
 apply (rule)
 apply (simp)
 apply (elim conjE exE)
 apply (rule_tac x= "Yf (Is!i) Int insert Tick (Ev ` snd (PXf (Is!i)))" in exI)
 apply (simp)
 apply (rule_tac x="(Is!i)" in exI)
 apply (simp add: isListOf_nth_in_index)
done

(*** Index_parallel_memF ***)

lemma Index_parallel_memF:
  "[| I ~= {} ; finite I |]
   ==> (f :f [[ [||]:I PXf ]]F) = 
         (EX u. (sett(u) <= insert Tick (Ev ` Union (snd ` PXf ` I)) & 
          (EX Z. f = (u, Z) &
          (EX Yf. 
          (Z Int insert Tick (Ev ` Union (snd ` PXf ` I))
           = Union {((Yf i) Int insert Tick (Ev ` (snd (PXf i))))|i. i : I} &
          (ALL i:I. ((u --tr (- (snd (PXf i)))), Yf i) :f [[fst (PXf i)]]F))))))"
apply (simp add: Index_parallel_def)
apply (subgoal_tac "EX Is. Is isListOf I")
apply (elim conjE exE)
apply (subgoal_tac "(map PXf (SOME Is. Is isListOf I)) ~= []")
apply (simp add: Rep_parallel_memF_nth)
apply (rule someI2)
 apply (simp)
 apply (rule iffI)

  (* => *)
  apply (elim conjE exE)
  apply (rename_tac Is' Is u Z Ys)
  apply (rule_tac x="u" in exI)
  apply (simp add: isListOf_set_eq)
  apply (rule_tac x= "(%i. (Ys!(THE n. (Is!n) = i & n<length Is)))" in exI)
  apply (rule conjI)
  apply (simp add: Index_parallel_memF_lm1)
  apply (rule ballI)

  apply (rotate_tac 4)
  apply (erule isListOf_index_to_nthE)
  apply (drule_tac x="i" in bspec, simp)
  apply (elim conjE exE, simp)

  apply (rotate_tac 2)
  apply (drule sym)
  apply (simp add: isListOf_THE_nth)

  (* <= *)
  apply (elim conjE exE)
  apply (rename_tac Is' Is u Z Yf)
  apply (rule_tac x="u" in exI)
  apply (simp add: isListOf_set_eq)
  apply (rule_tac x= "map Yf Is" in exI)
  apply (simp)
  apply (rule conjI)
  apply (rule Index_parallel_memF_lm2, simp)
  apply (intro allI impI)
  apply (drule_tac x="Is ! i" in bspec)
  apply (simp add: isListOf_nth_in_index)
  apply (simp)

 apply (rule someI2)
 apply (simp)
 apply (simp add: isListOf_nonemptyset)

apply (simp add: isListOf_EX)
done

(*** Semantics for indexed alphabetized parallel on F ***)

lemma Index_parallel_evalF:
  "[| I ~= {} ; finite I |]
   ==> [[ [||]:I PXf ]]F = 
      {f. (EX u. (sett(u) <= insert Tick (Ev ` Union (snd ` PXf ` I)) & 
          (EX Z. f = (u, Z) &
          (EX Yf. 
          (Z Int insert Tick (Ev ` Union (snd ` PXf ` I))
           = Union {((Yf i) Int insert Tick (Ev ` (snd (PXf i))))|i. i : I} &
          (ALL i:I. ((u --tr (- (snd (PXf i)))), Yf i) :f [[fst (PXf i)]]F))))))}f"
apply (simp add: Index_parallel_memF[THEN sym])
apply (simp add: CollectF_open)
done

(************************************
 |           empty Index            |
 ************************************)

lemma Index_parallel_empty[simp]: "I = {} ==> [||]:I PXf = SKIP"
by (simp add: Index_parallel_def)

(************************************
 |            one Index             |
 ************************************)

lemma Index_parallel_one: 
  "I = {i} ==> [||]:I PXf = (fst (PXf i)) |[snd (PXf i), {}]| SKIP"
by (simp add: Index_parallel_def)

(************************************
 |              traces              |
 ************************************)

lemma domT_Index_parallel_trace:
  "[| I ~= {} ; finite I ; t :t [[[||]:I PXf]]T |] 
   ==> sett t <= insert Tick (Ev ` Union (snd ` PXf ` I))"
by (simp add: Index_parallel_memT)

lemma domF_Index_parallel_trace:
  "[| I ~= {} ; finite I ; (t,X) :f [[[||]:I PXf]]F |] 
   ==> sett t <= insert Tick (Ev ` Union (snd ` PXf ` I))"
by (simp add: Index_parallel_memF)

(****************** to add it again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
