           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Int_choice = CSP_semantics:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*********************************************************
                        Dom_T
 *********************************************************)

(*** Int_choice_memT ***)

lemma Int_choicet_memT: 
  "(t :t T |~|t S) = (t :t T | t :t S)"
by (simp add: evalt_def evalT_def memT_UnT)

lemma Int_choice_memT: 
  "(t :t [[P |~| Q]]T) = (t :t [[P]]T | t :t [[Q]]T)"
by (simp add: Int_choicet_memT evalT_def)

(*********************************************************
                        Dom_F
 *********************************************************)

(*** Int_choice_memF ***)

lemma Int_choicef_memF: 
  "(f :f F |~|f E) = (f :f F | f :f E)"
by (simp add: evalf_def evalF_def memF_UnF)

lemma Int_choice_memF: 
  "(f :f [[P |~| Q]]F) = (f :f [[P]]F | f :f [[Q]]F)"
by (simp add: Int_choicef_memF evalF_def)

lemmas Int_choice_mem = Int_choice_memT  Int_choice_memF
                        Int_choicet_memT Int_choicef_memF

(*******************************
             domSF
 *******************************)

(* T2 *)

lemma Int_choice_T2 :
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_T2 (T |~|t S, F |~|f E)"
apply (simp add: HC_T2_def Int_choice_mem)
apply (simp add: domSF_def HC_T2_def)
by (auto)

(* F3 *)

lemma Int_choice_F3 :
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_F3 (T |~|t S, F |~|f E)"
apply (simp add: HC_F3_def Int_choice_mem)
apply (simp add: domSF_def HC_F3_def)
by (auto)

(* T3_F4 *)

lemma Int_choice_T3_F4 : 
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> HC_T3_F4 (T |~|t S, F |~|f E)"
apply (simp add: HC_T3_F4_def Int_choice_mem)
apply (simp add: domSF_iff HC_T3_F4_def)
by (auto)

(*** Int_choice_domSF ***)

lemma Int_choice_domSF : 
  "[| (T, F) : domSF ; (S, E) : domSF |]
     ==> (T |~|t S, F |~|f E) : domSF"
apply (simp (no_asm) add: domSF_iff)
apply (simp add: Int_choice_T2)
apply (simp add: Int_choice_F3)
apply (simp add: Int_choice_T3_F4)
done

(*********************************************************
                      mono
 *********************************************************)

(*** T ***)

lemma Int_choice_evalT_mono:
  "[| T1 <= T2 ; S1 <= S2 |] ==> T1 |~|t S1 <= T2 |~|t S2"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: Int_choice_mem)
apply (erule disjE)
by (simp_all)

(*** F ***)

lemma Int_choice_evalF_mono:
  "[| F1 <= F2 ; E1 <= E2 |] ==> F1 |~|f E1 <= F2 |~|f E2"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: Int_choice_mem)
apply (erule disjE)
by (simp_all)

(*** cong ***)

lemma Int_choice_evalT_cong:
  "[| T1 = T2 ; S1 = S2 |] ==> T1 |~|t S1 = T2 |~|t S2"
by (simp)

lemma Int_choice_evalF_cong:
  "[| F1 = F2 ; E1 = E2 |] ==> F1 |~|f E1 = F2 |~|f E2"
by (simp)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
