           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Int_choice_cms = Int_choice + Domain_SF_prod_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. T x |~|t S x) : non expanding
         2. (%x. F x |~|f E x) : non expanding
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Int_choice T
 *********************************************************)

(*** restT (subset) ***)

lemma Int_choice_restT_sub:
   "[| T x1 rest n <= T x2 rest n ;
       S x1 rest n <= S x2 rest n |]
    ==> (T x1 |~|t S x1) rest n <= (T x2 |~|t S x2) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Int_choice_mem)
by (auto)

(*** restT (equal) ***)

lemma Int_choice_restT:
   "[| T x1 rest n = T x2 rest n ;
       S x1 rest n = S x2 rest n |]
    ==> (T x1 |~|t S x1) rest n = (T x2 |~|t S x2) rest n"
apply (rule order_antisym)
by (simp_all add: Int_choice_restT_sub)

(*** distT lemma ***)

lemma Int_choice_distT:
   "TTs = {(T x1, T x2), (S x1, S x2)}
    ==> (EX TT. TT:TTs & 
             distance((T x1 |~|t S x1), (T x2 |~|t S x2))
          <= distance((fst TT), (snd TT)))"
apply (rule rest_to_dist_pair)
by (auto intro: Int_choice_restT)

(*** map_alpha T lemma ***)

lemma Int_choice_evalT_map_alpha_lm:
  "[| distance (T x1, T x2) <= alpha * distance (x1, x2) ;
      distance (S x1, S x2) <= alpha * distance (x1, x2) |]
    ==> distance (T x1 |~|t S x1, T x2 |~|t S x2)
     <= alpha * distance (x1, x2)"
apply (insert Int_choice_distT
       [of "{(T x1, T x2), (S x1, S x2)}" T x1 x2 S])
by (auto)

(*** Int_choice_evalT_non_expanding ***)

lemma Int_choice_evalT_map_alpha:
 "[| map_alpha T alpha ; map_alpha S alpha |]
  ==> map_alpha (%x. T x |~|t S x) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Int_choice_evalT_map_alpha_lm)

(*** Int_choice_evalT_non_expanding ***)

lemma Int_choice_evalT_non_expanding:
 "[| non_expanding T ; non_expanding S |]
  ==> non_expanding (%x. T x |~|t S x)"
by (simp add: non_expanding_def Int_choice_evalT_map_alpha)

(*** Int_choice_evalT_contraction_alpha ***)

lemma Int_choice_evalT_contraction_alpha:
 "[| contraction_alpha T alpha; contraction_alpha S alpha|]
  ==> contraction_alpha (%x. T x |~|t S x) alpha"
by (simp add: contraction_alpha_def Int_choice_evalT_map_alpha)

(*********************************************************
                   map Int_choice F
 *********************************************************)

(*** restF (subset) ***)

lemma Int_choice_restF_sub:
   "[| F x1 rest n <= F x2 rest n ;
       E x1 rest n <= E x2 rest n |]
    ==> (F x1 |~|f E x1) rest n <= (F x2 |~|f E x2) rest n"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_restF)
apply (simp add: Int_choice_mem)
apply (elim conjE exE disjE)
by (fast)+

(*** restF (equal) ***)

lemma Int_choice_restF:
   "[| F x1 rest n = F x2 rest n ;
       E x1 rest n = E x2 rest n |]
    ==> (F x1 |~|f E x1) rest n = (F x2 |~|f E x2) rest n"
apply (rule order_antisym)
by (simp_all add: Int_choice_restF_sub)

(*** distF lemma ***)

lemma Int_choice_distF:
  " FFs = {(F x1, F x2), (E x1, E x2)}
   ==> (EX FF. FF:FFs & 
               distance(F x1 |~|f E x1, F x2 |~|f E x2)
            <= distance((fst FF), (snd FF)))"
apply (rule rest_to_dist_pair)
by (auto intro: Int_choice_restF)

(*** map_alpha F lemma ***)

lemma Int_choice_evalF_map_alpha_lm:
  "[| distance (F x1, F x2) <= alpha * distance (x1, x2) ;
      distance (E x1, E x2) <= alpha * distance (x1, x2) |]
    ==> distance (F x1 |~|f E x1, F x2 |~|f E x2)
     <= alpha * distance (x1, x2)"
apply (insert Int_choice_distF
       [of "{(F x1, F x2), (E x1, E x2)}" F x1 x2 E])
by (auto)

(*** Int_choice_evalF_non_expanding ***)

lemma Int_choice_evalF_map_alpha:
 "[| map_alpha F alpha ; map_alpha E alpha |]
  ==> map_alpha (%x. F x |~|f E x) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Int_choice_evalF_map_alpha_lm)

(*** Int_choice_evalF_non_expanding ***)

lemma Int_choice_evalF_non_expanding:
 "[| non_expanding F ; non_expanding E |]
  ==> non_expanding (%x. F x |~|f E x)"
by (simp add: non_expanding_def Int_choice_evalF_map_alpha)

(*** Int_choice_evalF_contraction_alpha ***)

lemma Int_choice_evalF_contraction_alpha:
 "[| contraction_alpha F alpha; contraction_alpha E alpha|]
  ==> contraction_alpha (%x. F x |~|f E x) alpha"
by (simp add: contraction_alpha_def Int_choice_evalF_map_alpha)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
