           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Parallel_cms = Parallel + Domain_SF_prod_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. (T x) |[X]|t (S x)) : map
         2. (%x. (F x) |[X]|f (E x)) : map
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Parallel T
 *********************************************************)

(*** restT (subset) ***)

lemma Parallel_restT_sub:
   "[| T x1 rest n <= T x2 rest n ;
       S x1 rest n <= S x2 rest n |]
    ==> ((T x1) |[X]|t (S x1)) rest n <= ((T x2) |[X]|t (S x2)) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Parallel_mem)
apply (elim conjE exE)
apply (rule_tac x="s" in exI)
apply (rule_tac x="ta" in exI)
apply (simp)

apply (erule par_tr_lengthtE)
by (auto)

(*** restT (equal) ***)

lemma Parallel_restT:
   "[| T x1 rest n = T x2 rest n ;
       S x1 rest n = S x2 rest n |]
    ==> ((T x1) |[X]|t (S x1)) rest n = ((T x2) |[X]|t (S x2)) rest n"
apply (rule order_antisym)
by (simp_all add: Parallel_restT_sub)

(*** distT lemma ***)

lemma Parallel_distT:
  "TTs = {(T x1, T x2), (S x1, S x2)}
    ==> (EX TT. TT:TTs & 
             distance((T x1) |[X]|t (S x1), (T x2) |[X]|t (S x2))
          <= distance((fst TT), (snd TT)))"
apply (rule rest_to_dist_pair)
by (auto intro: Parallel_restT)

(*** map_alpha T lemma ***)

lemma Parallel_evalT_map_alpha_lm:
  "[| distance (T x1, T x2) <= alpha * distance (x1, x2) ;
      distance (S x1, S x2) <= alpha * distance (x1, x2) |]
    ==> distance((T x1) |[X]|t (S x1), (T x2) |[X]|t (S x2))
     <= alpha * distance (x1, x2)"
apply (insert Parallel_distT
       [of "{(T x1, T x2), (S x1, S x2)}" T x1 x2 S X])
by (auto)

(*** Parallel_evalT_map_alpha ***)

lemma Parallel_evalT_map_alpha:
 "[| map_alpha T alpha ; map_alpha S alpha |]
  ==> map_alpha (%x. (T x) |[X]|t (S x)) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Parallel_evalT_map_alpha_lm)

(*** Parallel_evalT_non_expanding ***)

lemma Parallel_evalT_non_expanding:
 "[| non_expanding T ; non_expanding S |]
  ==> non_expanding (%x. (T x) |[X]|t (S x))"
by (simp add: non_expanding_def Parallel_evalT_map_alpha)

(*** Parallel_evalT_contraction_alpha ***)

lemma Parallel_evalT_contraction_alpha:
 "[| contraction_alpha T alpha ; contraction_alpha S alpha |]
  ==> contraction_alpha (%x. (T x) |[X]|t (S x)) alpha"
by (simp add: contraction_alpha_def Parallel_evalT_map_alpha)

(*********************************************************
                   map Parallel F
 *********************************************************)

(*** restF (subset) ***)

lemma Parallel_restF_sub:
   "[| F x1 rest n <= F x2 rest n ;
       E x1 rest n <= E x2 rest n |]
    ==> ((F x1) |[X]|f (E x1)) rest n <= ((F x2) |[X]|f (E x2)) rest n"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_restF)
apply (simp add: Parallel_mem)
apply (elim conjE exE)
apply (rule_tac x="Y" in exI)
apply (rule_tac x="Z" in exI)
apply (simp)
apply (rule_tac x="sa" in exI)
apply (rule_tac x="t" in exI)
apply (simp)
apply (drule_tac x="sa" in spec)
apply (drule_tac x="t" in spec)
apply (drule_tac x="Y" in spec)
apply (drule_tac x="Z" in spec)

apply (erule disjE, simp)    (* lengtht s < n *)
apply (erule par_tr_lengthtE)
apply (simp)

apply (elim conjE exE, simp) (* lengtht s < n *)
apply (case_tac "~ notick s'", simp)
apply (simp add: par_tr_last)
apply (elim conjE exE, simp)
apply (erule par_tr_lengthtE)
apply (rotate_tac 4)
apply (drule sym, simp)
apply (force)
done

(*** restF (equal) ***)

lemma Parallel_restF:
   "[| F x1 rest n = F x2 rest n ;
       E x1 rest n = E x2 rest n |]
    ==> ((F x1) |[X]|f (E x1)) rest n = ((F x2) |[X]|f (E x2)) rest n"
apply (rule order_antisym)
by (simp_all add: Parallel_restF_sub)

(*** distF lemma ***)

lemma Parallel_distF:
  "FFs = {(F x1, F x2), (E x1, E x2)}
    ==> (EX FF. FF:FFs & 
             distance((F x1) |[X]|f (E x1), (F x2) |[X]|f (E x2))
          <= distance((fst FF), (snd FF)))"
apply (rule rest_to_dist_pair)
by (auto intro: Parallel_restF)

(*** map_alpha F lemma ***)

lemma Parallel_evalF_map_alpha_lm:
  "[| distance (F x1, F x2) <= alpha * distance (x1, x2) ;
      distance (E x1, E x2) <= alpha * distance (x1, x2) |]
    ==> distance((F x1) |[X]|f (E x1), (F x2) |[X]|f (E x2))
     <= alpha * distance (x1, x2)"
apply (insert Parallel_distF
       [of "{(F x1, F x2), (E x1, E x2)}" F x1 x2 E X])
by (auto)

(*** Parallel_evalF_map_alpha ***)

lemma Parallel_evalF_map_alpha:
 "[| map_alpha F alpha ; map_alpha E alpha |]
  ==> map_alpha (%x. (F x) |[X]|f (E x)) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Parallel_evalF_map_alpha_lm)

(*** Parallel_evalF_non_expanding ***)

lemma Parallel_evalF_non_expanding:
 "[| non_expanding F ; non_expanding E |]
  ==> non_expanding (%x. (F x) |[X]|f (E x))"
by (simp add: non_expanding_def Parallel_evalF_map_alpha)

(*** Parallel_evalF_contraction_alpha ***)

lemma Parallel_evalF_contraction_alpha:
 "[| contraction_alpha F alpha ; contraction_alpha E alpha |]
  ==> contraction_alpha (%x. (F x) |[X]|f (E x)) alpha"
by (simp add: contraction_alpha_def Parallel_evalF_map_alpha)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
