           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Renaming_cms = Renaming + Domain_SF_prod_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. (T x) [[r]]t) : non expanding
         2. (%x. (F x) [[r]]f) : non expanding
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Renaming T
 *********************************************************)

(*** restT (subset) ***)

lemma Renaming_restT_sub:
   "T x1 rest n <= T x2 rest n
    ==> ((T x1) [[r]]t) rest n <= ((T x2) [[r]]t) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Renaming_mem)
apply (elim conjE exE)
apply (rule_tac x="s" in exI)
apply (drule_tac x="s" in spec)
by (simp add: ren_tr_lengtht)

(*** restT (equal) ***)

lemma Renaming_restT:
   "T x1 rest n = T x2 rest n
    ==> ((T x1) [[r]]t) rest n = ((T x2) [[r]]t) rest n"
apply (rule order_antisym)
by (simp_all add: Renaming_restT_sub)

(*** distT lemma ***)

lemma Renaming_distT:
     "distance((T x1) [[r]]t, (T x2) [[r]]t) <= distance(T x1, T x2)"
apply (rule rest_distance_subset)
by (auto intro: Renaming_restT)

(*** map_alphaT lemma ***)

lemma Renaming_evalT_map_alpha_lm:
  "distance(T x1, T x2) <= alpha * distance (x1, x2)
    ==> distance((T x1) [[r]]t, (T x2) [[r]]t)
     <= alpha * distance (x1, x2)"
apply (insert Renaming_distT[of T x1 r x2])
by (simp)

(*** Renaming_evalT_map_alpha ***)

lemma Renaming_evalT_map_alpha:
 "map_alpha T alpha ==> map_alpha (%x. (T x) [[r]]t) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Renaming_evalT_map_alpha_lm)

(*** Renaming_evalT_non_expanding ***)

lemma Renaming_evalT_non_expanding:
 "non_expanding T ==> non_expanding (%x. (T x) [[r]]t)"
by (simp add: non_expanding_def Renaming_evalT_map_alpha)

(*** Renaming_evalT_contraction_alpha ***)

lemma Renaming_evalT_contraction_alpha:
 "contraction_alpha T alpha ==> contraction_alpha (%x. (T x) [[r]]t) alpha"
by (simp add: contraction_alpha_def Renaming_evalT_map_alpha)

(*********************************************************
                   map Renaming F
 *********************************************************)

(*** restF (subset) ***)

lemma Renaming_restF_sub:
   "F x1 rest n <= F x2 rest n
    ==> ((F x1) [[r]]f) rest n <= ((F x2) [[r]]f) rest n"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_restF)
apply (simp add: Renaming_mem)
apply (elim conjE exE)
apply (rule_tac x="sa" in exI)
apply (drule_tac x="sa" in spec)
apply (drule_tac x="[[r]]inv X" in spec)
apply (simp)

apply (erule disjE)
apply (simp add: ren_tr_lengtht)

apply (elim conjE exE)
apply (simp add: ren_tr_lengtht)
apply (simp add: not_None_F ren_tr_appt_decompo_right)
apply (elim conjE exE, simp)
by (fast)

(*** restF (equal) ***)

lemma Renaming_restF:
   "F x1 rest n = F x2 rest n
    ==> ((F x1) [[r]]f) rest n = ((F x2) [[r]]f) rest n"
apply (rule order_antisym)
by (simp_all add: Renaming_restF_sub)

(*** distF lemma ***)

lemma Renaming_distF:
     "distance((F x1) [[r]]f, (F x2) [[r]]f) <= distance(F x1, F x2)"
apply (rule rest_distance_subset)
by (auto intro: Renaming_restF)

(*** map_alphaF lemma ***)

lemma Renaming_evalF_map_alpha_lm:
  "distance(F x1, F x2) <= alpha * distance (x1, x2)
    ==> distance((F x1) [[r]]f, (F x2) [[r]]f)
     <= alpha * distance (x1, x2)"
apply (insert Renaming_distF[of F x1 r x2])
by (simp)

(*** Renaming_evalF_map_alpha ***)

lemma Renaming_evalF_map_alpha:
 "map_alpha F alpha ==> map_alpha (%x. (F x) [[r]]f) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Renaming_evalF_map_alpha_lm)

(*** Renaming_evalF_non_expanding ***)

lemma Renaming_evalF_non_expanding:
 "non_expanding F ==> non_expanding (%x. (F x) [[r]]f)"
by (simp add: non_expanding_def Renaming_evalF_map_alpha)

(*** Renaming_evalF_contraction_alpha ***)

lemma Renaming_evalF_contraction_alpha:
 "contraction_alpha F alpha ==> contraction_alpha (%x. (F x) [[r]]f) alpha"
by (simp add: contraction_alpha_def Renaming_evalF_map_alpha)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
