           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Rep_int_choice = CSP_semantics:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*********************************************************
                        Dom_T
 *********************************************************)

(*** Rep_int_choice_domT ***)

lemma Rep_int_choice_domT: 
  "{t. t = []t | (EX a:X. t :t Tf a) } : domT"
apply (simp add: domT_def HC_T1_def)
apply (rule conjI)
apply (rule_tac x="[]t" in exI, simp)

apply (simp add: prefix_closed_def)
apply (intro allI impI)
apply (elim conjE exE)

apply (erule disjE, simp)    (* []t *)

apply (elim conjE bexE)
apply (rule disjI2)
apply (rule_tac x="a" in bexI)
apply (rule memT_prefix_closed)
by (simp_all)

(*** Union_proc_memT ***)

lemma Union_proc_memT: 
   "(t :t {t. t = []t | (EX a:X. t :t Tf a)}t) 
  = (t = []t | (EX a:X. t :t Tf a))"
apply (insert Rep_int_choice_domT[of X Tf])
apply (simp add: evalt_def memT_def Abs_domT_inverse)
done

(*** Rep_int_choice_memT ***)

lemma Rep_int_choicet_memT: 
   "(t :t !! :X ..t Tf)  
  = (t = []t | (EX a:X. t :t Tf a))"
apply (simp add: evalt_def evalT_def)
apply (simp add: Union_proc_memT)
done

lemma Rep_int_choice_memT: 
   "(t :t [[!! :X .. Pf]]T) = (t = []t | (EX a:X. t :t [[Pf a]]T))"
by (simp add: Rep_int_choicet_memT evalT_def)

lemma Rep_int_choice_OR1_memT: 
   "(t :t [[! :X .. Pf]]T) = (t = []t | (EX a:X. t :t [[Pf a]]T))"
by (simp add: Rep_int_choicet_memT evalT_def)

lemma Rep_int_choice_OR2_memT: 
   "(t :t [[!nat :X .. Pf]]T) = (t = []t | (EX a:X. t :t [[Pf a]]T))"
by (simp add: Rep_int_choicet_memT evalT_def)

(*********************************************************
                        Dom_F
 *********************************************************)

(*** Rep_int_choice_domF ***)

lemma Rep_int_choice_domF: 
  "{f. EX a:X. f :f Ff a} : domF"
apply (simp add: domF_def HC_F2_def)
apply (intro allI impI)
apply (elim conjE bexE)
apply (rule_tac x="a" in bexI)
apply (rule memF_F2, simp_all)
done

(*** Union_proc_memF ***)

lemma Union_proc_memF:
   "f :f {f. EX a:X. f :f Ff a}f = (EX a:X. f :f Ff a)"
apply (insert Rep_int_choice_domF[of X Ff])
apply (simp add: memF_def Abs_domF_inverse)
done

(*** Rep_int_choice_memF ***)

lemma Rep_int_choicef_memF: 
  "(f :f !! :X ..f Ff) = (EX a:X. f :f Ff a)"
apply (simp add: evalf_def evalF_def)
apply (simp add: Union_proc_memF)
done

lemma Rep_int_choice_memF: 
  "(f :f [[!! :X .. Pf]]F) = (EX a:X. f :f [[Pf a]]F)"
by (simp add: Rep_int_choicef_memF evalF_def)

lemma Rep_int_choice_OR1_memF: 
  "(f :f [[! :X .. Pf]]F) = (EX a:X. f :f [[Pf a]]F)"
by (simp add: Rep_int_choicef_memF evalF_def)

lemma Rep_int_choice_OR2_memF: 
  "(f :f [[!nat :X .. Pf]]F) = (EX a:X. f :f [[Pf a]]F)"
by (simp add: Rep_int_choicef_memF evalF_def)

lemmas Rep_int_choice_mem = Rep_int_choicet_memT Rep_int_choicef_memF
                            Rep_int_choice_memT Rep_int_choice_memF
                            Rep_int_choice_OR1_memT Rep_int_choice_OR1_memF
                            Rep_int_choice_OR2_memT Rep_int_choice_OR2_memF

(*******************************
             domSF
 *******************************)

(* T2 *)

lemma Union_proc_T2:
  "ALL a. (Tf a, Ff a) : domSF
   ==> HC_T2 ({t. t = []t | (EX a:X. t :t Tf a)}t,
                         {f. EX a:X. f :f Ff a}f)"
apply (simp add: HC_T2_def)
apply (simp add: Union_proc_memT)
apply (simp add: Union_proc_memF)
apply (intro allI impI)
apply (elim conjE bexE exE)
apply (rule disjI2)
apply (drule_tac x="a" in spec)
apply (rule_tac x="a" in bexI)
apply (simp add: domSF_def HC_T2_def)
by (auto)

lemma Rep_int_choice_T2 :
  "ALL a. (Tf a, Ff a) : domSF
     ==> HC_T2 (!! :X ..t Tf, !! :X ..f Ff)"
apply (simp add: evalt_def evalf_def)
apply (simp add: Union_proc_T2)
done

(* F3 *)

lemma Union_proc_F3:
  "ALL a. (Tf a, Ff a) : domSF
   ==>
   HC_F3 ({t. t = []t | (EX a:X. t :t Tf a)}t,
          {f. EX a:X. f :f Ff a}f)"
apply (simp add: HC_F3_def)
apply (simp add: Union_proc_memT)
apply (simp add: Union_proc_memF)
apply (intro allI)
apply (intro impI)
apply (elim conjE bexE)
apply (rule_tac x="a" in bexI)
apply (drule_tac x="a" in spec)
apply (rule domSF_F3)
apply (auto)
done

lemma Rep_int_choice_F3 : 
  "ALL a. (Tf a, Ff a) : domSF
     ==> HC_F3 (!! :X ..t Tf, !! :X ..f Ff)"
apply (simp add: evalt_def evalf_def)
apply (simp add: Union_proc_F3)
done

(* T3_F4 *)

lemma Union_proc_T3_F4:
  "ALL a. (Tf a, Ff a) : domSF
   ==>
   HC_T3_F4 ({t. t = []t | (EX a:X. t :t Tf a)}t,
             {f. EX a:X. f :f Ff a}f)"
apply (simp add: HC_T3_F4_def)
apply (simp add: Union_proc_memT)
apply (simp add: Union_proc_memF)
apply (auto)
apply (rule_tac x="a" in bexI)
apply (rule domSF_F4)
apply (simp_all)
apply (rule_tac x="a" in bexI)
apply (rule domSF_T3)
apply (simp_all)
done

lemma Rep_int_choice_T3_F4 : 
  "ALL a. (Tf a, Ff a) : domSF
     ==> HC_T3_F4 (!! :X ..t Tf, !! :X ..f Ff)"
apply (simp add: evalt_def evalf_def)
apply (simp add: Union_proc_T3_F4)
done

(*** SF ***)

lemma Union_proc_domSF:
  "ALL a. (Tf a, Ff a) : domSF
   ==> ({t. t = []t | (EX a:X. t :t Tf a)}t,
        {f. EX a:X. f :f Ff a}f) : domSF"
apply (simp (no_asm) add: domSF_iff)
apply (simp add: Union_proc_T2)
apply (simp add: Union_proc_F3)
apply (simp add: Union_proc_T3_F4)
done

(*** Rep_int_choice_domSF ***)

lemma Rep_int_choice_domSF : 
  "ALL a. (Tf a, Ff a) : domSF
     ==> (!! :X ..t Tf, !! :X ..f Ff) : domSF"
apply (simp add: evalt_def evalf_def)
apply (simp add: Union_proc_domSF)
done

(*********************************************************
                      UNIV
 *********************************************************)

lemma UNIV_Union_proc_memT: 
   "(t :t {t. EX a. t :t Tf a}t) 
  = (EX a. t :t Tf a)"
apply (insert Union_proc_memT[of t UNIV Tf])
apply (simp)
apply (subgoal_tac "{t. t = []t | (EX a. t :t Tf a)}
                  = {t. EX a. t :t Tf a}")
by (auto)

lemma UNIV_Union_proc_memF:
   "f :f {f. EX a. f :f Ff a}f = (EX a. f :f Ff a)"
apply (insert Union_proc_memF[of f UNIV Ff])
apply (simp)
done

lemma UNIV_Union_proc_domSF:
  "ALL a. (Tf a, Ff a) : domSF
   ==> ({t. EX a. t :t Tf a}t,
        {f. EX a. f :f Ff a}f) : domSF"
apply (insert Union_proc_domSF[of Tf Ff UNIV])
apply (subgoal_tac "{t. t = []t | (EX a. t :t Tf a)}
                  = {t. EX a. t :t Tf a}")
by (auto)

(*********************************************************
                      mono
 *********************************************************)

(*** T ***)

lemma Rep_int_choice_evalT_mono:
  "ALL a:X. Tf a <= Sf a
   ==> !! :X ..t Tf <= !! :X ..t Sf"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: Rep_int_choice_mem)
apply (erule disjE, simp)

apply (elim conjE bexE)
apply (drule_tac x="a" in bspec, simp)
apply (drule_tac x="t" in spec, simp)
apply (rule disjI2)
apply (rule_tac x="a" in bexI)
apply (simp_all)
done

(*** F ***)

lemma Rep_int_choice_evalF_mono:
  "ALL a:X. Ff a <= Ef a
   ==> !! :X ..f Ff <= !! :X ..f Ef"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: Rep_int_choice_mem)

apply (elim conjE bexE)
apply (drule_tac x="a" in bspec, simp)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Xa" in spec)
apply (simp)
apply (rule_tac x="a" in bexI)
apply (simp_all)
done

(* cong *)

lemma Rep_int_choice_evalT_cong:
  "ALL a:X. Tf a = Sf a
   ==> !! :X ..t Tf = !! :X ..t Sf"
apply (rule order_antisym)
apply (simp_all add: Rep_int_choice_evalT_mono)
done

lemma Rep_int_choice_evalF_cong:
  "ALL a:X. Ff a = Ef a
   ==> !! :X ..f Ff = !! :X ..f Ef"
apply (rule order_antisym)
apply (simp_all add: Rep_int_choice_evalF_mono)
done

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
