           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Rep_int_choice_cms = Rep_int_choice + Domain_SF_prod_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. !! :X ..t (%a. Tf a x)) : non_expanding
         2. (%x. !! :X ..f (%a. Ff a x)) : non_expanding
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Rep_int_choice T
 *********************************************************)

(*** restT (subset) ***)

lemma Rep_int_choice_restT_sub:
   "[| ALL a : X. (Tf a x1) rest n <= (Tf a x2) rest n |]
    ==> (!! :X ..t (%a. (Tf a x1))) rest n <=
        (!! :X ..t (%a. (Tf a x2))) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Rep_int_choice_mem)
apply (elim conjE bexE disjE)
apply (simp_all)

apply (drule_tac x="a" in bspec, simp)
apply (drule_tac x="t" in spec, simp)
by (fast)

(*** restT (equal) ***)

lemma Rep_int_choice_restT:
   "[| ALL a : X. (Tf a x1) rest n = (Tf a x2) rest n |]
    ==> (!! :X ..t (%a. (Tf a x1))) rest n =
        (!! :X ..t (%a. (Tf a x2))) rest n"
apply (rule order_antisym)
by (simp_all add: Rep_int_choice_restT_sub)

(*** distT lemma ***)

lemma Rep_int_choice_distT_nonempty:
"[| X ~= {} ; TTs = {(Tf a x1, Tf a x2)|a. a : X} |]
 ==> (EX TT. TT:TTs & 
             distance(!! :X ..t (%a. (Tf a x1)), !! :X ..t (%a. (Tf a x2))) 
          <= distance((fst TT), (snd TT)))"
apply (rule rest_to_dist_pair)
apply (fast)

apply (intro allI impI)
apply (rule Rep_int_choice_restT)
apply (rule ballI)
apply (simp)
apply (drule_tac x="(Tf a x1)" in spec)
apply (drule_tac x="(Tf a x2)" in spec)
by (auto)

(*** map_alpha T lemma ***)

lemma Rep_int_choice_evalT_map_alpha_nonempty_lm:
  "[| X ~= {};
      ALL a. distance (Tf a x1, Tf a x2)
           <= alpha * distance (x1, x2) |]
    ==> distance (!! :X ..t (%a. (Tf a x1)), !! :X ..t (%a. (Tf a x2)))
     <= alpha *  distance (x1, x2)"
apply (insert Rep_int_choice_distT_nonempty
       [of X "{(Tf a x1, Tf a x2)|a. a : X}" Tf x1 x2])
apply (simp)
apply (elim conjE exE, simp)
apply (drule_tac x="aa" in spec)
by (force)

(*** Rep_int_choice_evalT_map_alpha ***)

lemma Rep_int_choice_evalT_map_alpha:
 "ALL a. map_alpha (Tf a) alpha ==> map_alpha (%x. !! :X ..t (%a. Tf a x)) alpha"
apply (simp add: map_alpha_def)
apply (case_tac "X = {}")
 apply (simp add: evalt_def)
 apply (simp add: real_mult_order_eq)
apply (simp add: Rep_int_choice_evalT_map_alpha_nonempty_lm)
done

(*** Rep_int_choice_evalT_non_expanding ***)

lemma Rep_int_choice_evalT_non_expanding:
 "ALL a. non_expanding (Tf a) ==> non_expanding (%x. !! :X ..t (%a. Tf a x))"
by (simp add: non_expanding_def Rep_int_choice_evalT_map_alpha)

(*** Rep_int_choice_evalT_contraction_alpha ***)

lemma Rep_int_choice_evalT_contraction_alpha:
 "ALL a. contraction_alpha (Tf a) alpha 
     ==> contraction_alpha (%x. !! :X ..t (%a. Tf a x)) alpha"
by (simp add: contraction_alpha_def Rep_int_choice_evalT_map_alpha)

(*********************************************************
                   map Rep_int_choice F
 *********************************************************)

(*** restF (subset) ***)

lemma Rep_int_choice_restF_sub:
   "[| ALL a : X. (Ff a x1) rest n <= (Ff a x2) rest n |]
    ==> (!! :X ..f (%a. (Ff a x1))) rest n <=
        (!! :X ..f (%a. (Ff a x2))) rest n"
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_restF)
apply (simp add: Rep_int_choice_mem)
apply (elim conjE exE bexE disjE)
apply (simp_all)

apply (drule_tac x="a" in bspec, simp)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Xa" in spec)
apply (rule_tac x="a" in bexI)
apply (simp_all)
by (fast)

(*** restF (equal) ***)

lemma Rep_int_choice_restF:
   "[| ALL a : X. (Ff a x1) rest n = (Ff a x2) rest n |]
    ==> (!! :X ..f (%a. (Ff a x1))) rest n =
        (!! :X ..f (%a. (Ff a x2))) rest n"
apply (rule order_antisym)
by (simp_all add: Rep_int_choice_restF_sub)

(*** distF lemma ***)

lemma Rep_int_choice_distF_nonempty:
"[| X ~= {} ; FFs = {(Ff a x1, Ff a x2)|a. a : X} |]
 ==> (EX FF. FF:FFs & 
             distance(!! :X ..f (%a. (Ff a x1)), !! :X ..f (%a. (Ff a x2))) 
          <= distance((fst FF), (snd FF)))"
apply (rule rest_to_dist_pair)
apply (fast)

apply (intro allI impI)
apply (rule Rep_int_choice_restF)
apply (rule ballI)
apply (simp)
apply (drule_tac x="Ff a x1" in spec)
apply (drule_tac x="Ff a x2" in spec)
by (auto)

(*** map_alpha F lemma ***)

lemma Rep_int_choice_evalF_map_alpha_nonempty_lm:
  "[| X ~= {};
      ALL a. distance (Ff a x1, Ff a x2)
           <= alpha * distance (x1, x2) |]
    ==>  distance(!! :X ..f (%a. (Ff a x1)), !! :X ..f (%a. (Ff a x2))) 
     <= alpha *  distance (x1, x2)"
apply (insert Rep_int_choice_distF_nonempty
       [of X "{(Ff a x1, Ff a x2)|a. a : X}" Ff x1 x2])
apply (simp)
apply (elim conjE exE, simp)
apply (drule_tac x="aa" in spec)
by (force)

(*** Rep_int_choice_evalF_map_alpha ***)

lemma Rep_int_choice_evalF_map_alpha:
 "ALL a. map_alpha (Ff a) alpha ==> map_alpha (%x. !! :X ..f (%a. Ff a x)) alpha"
apply (simp add: map_alpha_def)
apply (case_tac "X = {}")
 apply (simp add: evalf_def)
 apply (simp add: real_mult_order_eq)
apply (simp add: Rep_int_choice_evalF_map_alpha_nonempty_lm)
done

(*** Rep_int_choice_evalF_non_expanding ***)

lemma Rep_int_choice_evalF_non_expanding:
 "ALL a. non_expanding (Ff a) ==> non_expanding (%x. !! :X ..f (%a. Ff a x))"
by (simp add: non_expanding_def Rep_int_choice_evalF_map_alpha)

(*** Rep_int_choice_evalF_contraction_alpha ***)

lemma Rep_int_choice_evalF_contraction_alpha:
 "ALL a. contraction_alpha (Ff a) alpha 
     ==> contraction_alpha (%x. !! :X ..f (%a. Ff a x)) alpha"
by (simp add: contraction_alpha_def Rep_int_choice_evalF_map_alpha)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
