           (*-------------------------------------------*
            |                CSP-Prover                 |
            |                 May 2005                  |
            |                June 2005  (modified)      |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Rep_parallel = Alpha_parallel:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(**************************************************************
                         lemmas
 **************************************************************)

(*------------------------*
 |           T            |
 *------------------------*)

(*** Rep_parallel_evalT ***)

lemma Rep_parallel_memT_lm1: 
  "(P, X) : set PXs ==> X <= Union (snd ` (set PXs))"
by (auto)

(* main *)

lemma Rep_parallel_memT_lm:
  "PXs ~= [] --> (ALL u.
    (u :t [[ [||] PXs ]]T) = 
     (sett(u) <= insert Tick (Ev ` (Union (snd ` (set PXs)))) & 
      (ALL P X. (P,X):(set PXs) --> (u --tr (- X)) :t [[P]]T)))"
apply (induct_tac PXs)

(* case 0 *)
 apply (simp)

(* case 1 *)
 apply (case_tac "list = []")
 apply (simp)
 apply (intro allI)
 apply (simp add: Alpha_parallel_memT)
 apply (simp add: SKIP_memT)
 apply (simp add: pair_eq_decompo)
 apply (case_tac "u = None", simp)
 apply (simp add: hide_tr_UNIV)
 apply (rule iffI)
 (* => *)
  apply (simp)
 (* <= *)
  apply (simp)

(* step case *)
apply (simp add: Alpha_parallel_memT)
apply (intro allI impI)
apply (rule iffI)

(* => *)
 apply (simp)
 apply (intro allI)
 apply (rule conjI)

  apply (intro impI)
  apply (simp add: pair_eq_decompo)

  apply (intro impI)
  apply (elim conjE)
  apply (drule_tac x="P" in spec)
  apply (drule_tac x="X" in spec)
  apply (simp)
  apply (subgoal_tac "(- Union (snd ` set list)) <= (- X)")
  apply (simp add: hide_tr_of_hide_tr_subset)
  apply (simp add: Rep_parallel_memT_lm1)

(* <= *)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (intro allI impI)
 apply (elim conjE)
 apply (drule_tac x="P" in spec)
 apply (drule_tac x="X" in spec)
 apply (subgoal_tac "(- Union (snd ` set list)) <= (- X)")
 apply (simp add: hide_tr_of_hide_tr_subset)
 apply (simp add: Rep_parallel_memT_lm1)
done

(*** remove ALL ***)

lemma Rep_parallel_memT:
  "PXs ~= [] 
   ==> (u :t [[ [||] PXs ]]T) = 
       (sett(u) <= insert Tick (Ev ` (Union (snd ` (set PXs)))) & 
        (ALL P X. (P,X):(set PXs) --> (u --tr (- X)) :t [[P]]T))"
by (simp add: Rep_parallel_memT_lm)

(*** Semantics for replicated alphabetized parallel on T ***)

lemma Rep_parallel_evalT:
  "PXs ~= []
   ==> [[ [||] PXs ]]T = 
       {u. sett(u) <= insert Tick (Ev ` (Union (snd ` (set PXs)))) & 
        (ALL P X. (P,X):(set PXs) --> (u --tr (- X)) :t [[P]]T)}t"
apply (simp add: Rep_parallel_memT[THEN sym])
apply (simp add: CollectT_open)
done

(*------------------------*
 |           F            |
 *------------------------*)

(*** Rep_parallel_evalF ***)

lemma Rep_parallel_memF_lm1: 
       "Y Int insert Tick (Ev ` snd a) Un
        Union {Y Int insert Tick (Ev ` X) |X Y. EX P. ((P, X), Y) : set PXYs} =
          Union {Ya Int insert Tick (Ev ` X) |X Ya.
                 EX P. P = fst a & X = snd a & Ya = Y | ((P, X), Ya) : set PXYs}"
by (auto)

lemma Rep_parallel_memF_lm2: 
  "((P, X), Y) : set s ==> X <= Union (snd ` fst ` set s)"
apply (auto)
apply (rule_tac x="((P, X), Y)" in bexI)
by (simp_all)

lemma Rep_parallel_memF_lm3:
   "Union {Y Int insert Tick (Ev ` X) |X Y. EX P. ((P, X), Y) : set zs} <=
    insert Tick (Ev ` Union (snd ` fst ` set zs))"
apply (rule)
apply (simp)
apply (elim conjE exE)
apply (simp)
apply (elim conjE disjE)
apply (simp)
apply (rule disjI2)
apply (auto)
apply (simp add: image_iff)
apply (rule_tac x="((P, X), Y)" in bexI)
apply (auto)
done

lemma Rep_parallel_memF_lm4:
   "Union {Y Int insert Tick (Ev ` X) |X Y. EX P. ((P, X), Y) : set zs} Int
    insert Tick (Ev ` Union (snd ` fst ` set zs))
    = Union {Y Int insert Tick (Ev ` X) |X Y. EX P. ((P, X), Y) : set zs}"
apply (rule Int_subset_eq)
apply (simp add: Rep_parallel_memF_lm3)
done

(* main *)

lemma Rep_parallel_memF_lm:
 "PXs ~= [] --> (ALL f.
   (f :f [[ [||] PXs ]]F) = 
         (EX u. (sett(u) <= insert Tick (Ev ` Union (snd ` set PXs)) & 
          (EX Z. f = (u, Z) &
          (EX PXYs. map fst PXYs = PXs &
          (Z Int insert Tick (Ev ` Union (snd ` set PXs))
           = Union {(Y Int insert Tick (Ev ` X))|X Y. EX P. ((P,X),Y) : set PXYs} &
          (ALL P X Y. ((P,X),Y) : set PXYs --> ((u --tr (- X)), Y) :f [[P]]F)))))))"
apply (induct_tac PXs)

(* case 0 *)
 apply (simp)

(* case 1 *)
 apply (case_tac "list = []")
 apply (simp)
 apply (intro allI)
 apply (simp add: Alpha_parallel_memF)
 apply (simp add: SKIP_memF)
 apply (rule iffI)
 (* => *)
  apply (elim conjE exE)
  apply (case_tac "aa = None", simp)
  apply (simp)
  apply (erule disjE)

   apply (rule_tac x="[(a,Y)]" in exI)
   apply (rule conjI)
    apply (rule_tac x="fst a" in exI)
    apply (rule_tac x="snd a" in exI)
    apply (simp)
    apply (rule conjI)
    apply (simp add: Evset_def pair_eq_decompo)
    apply (fast)
    apply (intro allI impI)
    apply (simp add: pair_eq_decompo)

   apply (case_tac "Tick ~: Z")
    apply (rule_tac x="[(a,Y)]" in exI)
    apply (simp add: Evset_def pair_eq_decompo)
    apply (fast)
   (* Tick : Z *)
    apply (rule_tac x="[(a,insert Tick Y)]" in exI)
    apply (simp)
    apply (rule conjI)
    apply (fast)
    apply (simp add: hide_tr_Tick_sett)
    apply (elim conjE exE)
    apply (simp add: pair_eq_decompo)
    apply (rule conjI)
    apply (fast)
    apply (rule proc_T2_T3, simp)

 (* <= *)
  apply (elim conjE exE)
  apply (simp)
  apply (case_tac "aa = None", simp)
  apply (subgoal_tac "aa --tr UNIV = []t | aa --tr UNIV = [Tick]t")
  apply (erule disjE)

   (* []t *)
   apply (rule_tac x="bb" in exI)
   apply (rule_tac x="{}" in exI)
   apply (simp add: pair_eq_decompo)

   (* [Tick]t *)
   apply (rule_tac x="bb" in exI)
   apply (rule_tac x="{}" in exI)
   apply (simp add: pair_eq_decompo)

   apply (simp add: hide_tr_UNIV)

(* step case *)
apply (simp add: Alpha_parallel_memF)
apply (intro allI impI)
apply (rule iffI)

(* => *)
 apply (elim conjE exE, simp)
 apply (rule_tac x="(a,Y) # PXYs" in exI)
 apply (simp add: pair_eq_decompo)
 apply (simp add: Rep_parallel_memF_lm1)

 apply (case_tac "aa = None", simp)
 apply (intro allI impI)
 apply (drule_tac x="P" in spec)
 apply (drule_tac x="X" in spec)
 apply (drule_tac x="Ya" in spec)
 apply (simp)
 apply (subgoal_tac "(- Union (snd ` set list)) <= (- X)")
 apply (simp add: hide_tr_of_hide_tr_subset)

 apply (rotate_tac -5)
 apply (drule sym)
 apply (simp add: Rep_parallel_memF_lm2)

(* <= *)
 apply (simp add: hide_tr_sett_comp_subseteq)
 apply (elim conjE exE)

 apply (rule_tac x="bb" in exI)
 apply (rule_tac x=
    "Union {Y Int insert Tick (Ev ` X) |X Y. EX P. ((P, X), Y) : set zs}" in exI)
 apply (simp)
 apply (rule conjI)
 apply (simp add: pair_eq_decompo)
 apply (rotate_tac -1)
 apply (drule sym)
 apply (simp)
 apply (simp add: Rep_parallel_memF_lm4)
 apply (simp add: Rep_parallel_memF_lm1)

 apply (rule_tac x="zs" in exI)
 apply (rotate_tac -1)
 apply (drule sym)
 apply (simp add: Rep_parallel_memF_lm4)

 apply (intro allI impI)
 apply (subgoal_tac "(- Union (snd ` set list)) <= (- X)")
 apply (simp add: hide_tr_of_hide_tr_subset)
 apply (simp add: Rep_parallel_memF_lm2)
done

(*** remove ALL ***)

lemma Rep_parallel_memF:
 "PXs ~= [] ==> 
   (f :f [[ [||] PXs ]]F) = 
         (EX u. (sett(u) <= insert Tick (Ev ` Union (snd ` set PXs)) & 
          (EX Z. f = (u, Z) &
          (EX PXYs. map fst PXYs = PXs &
          (Z Int insert Tick (Ev ` Union (snd ` set PXs))
           = Union {(Y Int insert Tick (Ev ` X))|X Y. EX P. ((P,X),Y) : set PXYs} &
          (ALL P X Y. ((P,X),Y) : set PXYs --> ((u --tr (- X)), Y) :f [[P]]F))))))"
by (simp add: Rep_parallel_memF_lm)

(*** Semantics for replicated alphabetized parallel on F ***)

lemma Rep_parallel_evalF:
  "PXs ~= []
   ==> [[ [||] PXs ]]F = 
      {f. (EX u. (sett(u) <= insert Tick (Ev ` Union (snd ` set PXs)) & 
          (EX Z. f = (u, Z) &
          (EX PXYs. map fst PXYs = PXs &
          (Z Int insert Tick (Ev ` Union (snd ` set PXs))
           = Union {(Y Int insert Tick (Ev ` X))|X Y. EX P. ((P,X),Y) : set PXYs} &
          (ALL P X Y. ((P,X),Y) : set PXYs --> ((u --tr (- X)), Y) :f [[P]]F))))))}f"
apply (simp add: Rep_parallel_memF[THEN sym])
apply (auto simp add: CollectF_open)
done

(************************************
 |              traces              |
 ************************************)

lemma domT_Rep_parallel_trace:
  "[| PXs ~= [] ; t :t [[[||] PXs]]T |] 
   ==> sett t <= insert Tick (Ev ` Union (snd ` set PXs))"
by (simp add: Rep_parallel_memT)

lemma domF_Rep_parallel_trace:
  "[| PXs ~= [] ; (t,X) :f [[[||] PXs]]F |] 
   ==> sett t <= insert Tick (Ev ` Union (snd ` set PXs))"
by (simp add: Rep_parallel_memF)

(*---------------------------------------------------------*
 |        another expression of Rep_parallel_eval          |
 *---------------------------------------------------------*)

lemma Rep_parallel_memT_nth:
  "PXs ~= [] 
   ==> (u :t [[ [||] PXs ]]T) = 
       (sett(u) <= insert Tick (Ev ` (Union (snd ` (set PXs)))) & 
        (ALL i. i < length PXs --> (u --tr (- (snd (PXs!i)))) :t [[fst (PXs!i)]]T))"
apply (simp add: Rep_parallel_memT)
apply (simp add: set_nth)
apply (simp add: pair_eq_decompo)
by (auto)

lemma Rep_parallel_memF_nth:
 "PXs ~= [] ==> 
   (f :f [[ [||] PXs ]]F) = 
         (EX u. (sett(u) <= insert Tick (Ev ` Union (snd ` set PXs)) & 
          (EX Z. f = (u, Z) &
          (EX Ys. length PXs = length Ys &
          (Z Int insert Tick (Ev ` Union (snd ` set PXs))
           = Union {((Ys!i) Int insert Tick (Ev ` (snd (PXs!i))))|i. i < length PXs} &
          (ALL i. i < length PXs --> ((u --tr (- (snd (PXs!i)))), Ys!i) :f 
            [[fst (PXs!i)]]F))))))"
apply (simp add: Rep_parallel_memF)
apply (simp add: set_nth)
apply (rule iffI)

(* => *)
 apply (elim conjE exE)
 apply (simp)
 apply (rule_tac x="map snd PXYs" in exI)
 apply (simp)
 apply (rotate_tac 3)
 apply (drule sym)
 apply (simp)
 apply (rule conjI)
 apply (simp add: pair_eq_decompo)

 apply (rule equalityI)
 (* <= *)
  apply (rule)
  apply (simp)
  apply (elim conjE exE)
  apply (simp)
  apply (elim conjE disjE)
   apply (simp)
   apply (rule_tac x="map snd PXYs ! i Int
                      insert Tick (Ev ` snd (map fst PXYs ! i))" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

   apply (rule_tac x="map snd PXYs ! i Int
                      insert Tick (Ev ` snd (map fst PXYs ! i))" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

 (* => *)
  apply (rule)
  apply (simp)
  apply (elim conjE exE)
  apply (simp)
  apply (elim conjE disjE)
   apply (simp)
   apply (rule_tac x=
     "(snd (PXYs ! i)) Int insert Tick (Ev ` snd (fst (PXYs ! i)))" in exI)
   apply (simp)
   apply (rule_tac x="snd (fst (PXYs ! i))" in exI)
   apply (rule_tac x="snd (PXYs ! i)" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

   apply (rule_tac x=
     "(snd (PXYs ! i)) Int insert Tick (Ev ` snd (fst (PXYs ! i)))" in exI)
   apply (simp)
   apply (rule_tac x="snd (fst (PXYs ! i))" in exI)
   apply (rule_tac x="snd (PXYs ! i)" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

  apply (intro allI impI)
  apply (drule_tac x="fst (fst (PXYs ! i))" in spec)
  apply (drule_tac x="snd (fst (PXYs ! i))" in spec)
  apply (drule_tac x="snd (PXYs ! i)" in spec)
  apply (simp)
  apply (fast)

 (* => *)
 apply (elim conjE exE)
 apply (simp)
 apply (rule_tac x="zip PXs Ys" in exI)
 apply (simp add: map_fst_zip_eq)

 apply (rule conjI)
 apply (simp add: pair_eq_decompo)

 apply (rule equalityI)
 (* <= *)
  apply (rule)
  apply (simp)
  apply (elim conjE exE)
  apply (simp)
  apply (elim conjE disjE)
   apply (simp)
   apply (rule_tac x="Ys ! i Int insert Tick (Ev ` snd (PXs ! i))" in exI)
   apply (simp)
   apply (rule_tac x="snd (PXs ! i)" in exI)
   apply (rule_tac x="Ys ! i" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

   apply (rule_tac x="Ys ! i Int insert Tick (Ev ` snd (PXs ! i))" in exI)
   apply (simp)
   apply (rule_tac x="snd (PXs ! i)" in exI)
   apply (rule_tac x="Ys ! i" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

 (* => *)
  apply (rule)
  apply (simp)
  apply (elim conjE exE)
  apply (simp)
  apply (elim conjE disjE)
   apply (simp)
   apply (rule_tac x="Ys ! i Int insert Tick (Ev ` snd (PXs ! i))" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

   apply (rule_tac x="Ys ! i Int insert Tick (Ev ` snd (PXs ! i))" in exI)
   apply (simp)
   apply (rule_tac x="i" in exI)
   apply (simp)

  apply (intro allI impI)
  apply (elim conjE exE)
  apply (drule_tac x="i" in spec)
  apply (simp add: pair_eq_decompo)
done

(****************** to add it again ******************)

declare disj_not1   [simp] 
declare not_None_eq [simp]

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
