           (*-------------------------------------------*
            |                CSP-Prover                 |
            |               December 2004               |
            |               June 2005 (modified)        |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Seq_compo_cms = Seq_compo + Domain_SF_prod_cms:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. (%x. (T x) ;;t (S x)) : non expanding
         2. (%x. (T x,F x) ;;f (E x)) : non expanding
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                   map Seq_compo T
 *********************************************************)

(*** restT (subset) ***)

lemma Seq_compo_restT_sub:
   "[| T x1 rest n <= T x2 rest n ;
       S x1 rest n <= S x2 rest n |]
    ==> ((T x1) ;;t (S x1)) rest n <= ((T x2) ;;t (S x2)) rest n"
apply (simp add: subsetT_iff)
apply (intro allI impI)
apply (simp add: in_restT)
apply (simp add: Seq_compo_mem)
apply (elim conjE exE disjE)

 (* case 1 *)
 apply (rule disjI1)
 apply (insert trace_last_notick_or_tick)
 apply (rotate_tac -1)
 apply (drule_tac x="s" in spec, simp add: not_None_T)
 apply (erule disjE, simp)
 apply (rule_tac x="s" in exI, simp)

 apply (elim conjE exE)
 apply (rule_tac x="sa" in exI, simp)
 apply (drule_tac x="sa" in spec, simp)
 apply (drule mp)
 apply (rule memT_prefix_closed, simp_all, simp)

 (* case 2 *)
 apply (case_tac "~ notick s", simp add: not_None_T)
 apply (insert trace_last_nil_or_unnil)
 apply (rotate_tac -1)
 apply (drule_tac x="ta" in spec, simp add: not_None_T)
 apply (erule disjE, simp)

  apply (rule disjI1)                (* ta = []t *)
  apply (rule_tac x="s" in exI, simp)
  apply (drule_tac x="s" in spec, simp)
  apply (drule mp)
  apply (rule memT_prefix_closed, simp_all, simp)

  apply (rule disjI2)                (* ta ~= []t *)
  apply (elim conjE exE, simp)
  apply (simp add: appt_ass_rev del: appt_ass)

  apply (drule_tac x="s @t [Tick]t" in spec)
  apply (drule_tac x=" sa @t [a]t" in spec)
  apply (simp)
  apply (rule_tac x="s" in exI)
  apply (rule_tac x="sa @t [a]t" in exI)
  apply (simp add: not_None)
done

(*** restT (equal) ***)

lemma Seq_compo_restT:
   "[| T x1 rest n = T x2 rest n ;
       S x1 rest n = S x2 rest n |]
    ==> ((T x1) ;;t (S x1)) rest n = ((T x2) ;;t (S x2)) rest n"
apply (rule order_antisym)
by (simp_all add: Seq_compo_restT_sub)

(*** distT lemma ***)

lemma Seq_compo_distT:
  "TTs = {(T x1, T x2), (S x1, S x2)}
    ==> (EX TT. TT:TTs & 
             distance((T x1) ;;t (S x1), (T x2) ;;t (S x2))
          <= distance((fst TT), (snd TT)))"
apply (rule rest_to_dist_pair)
by (auto intro: Seq_compo_restT)

(*** map_alpha T lemma ***)

lemma Seq_compo_evalT_map_alpha_lm:
  "[| distance (T x1, T x2) <= alpha * distance (x1, x2) ;
      distance (S x1, S x2) <= alpha * distance (x1, x2) |]
    ==> distance((T x1) ;;t (S x1), (T x2) ;;t (S x2))
     <= alpha * distance (x1, x2)"
apply (insert Seq_compo_distT
       [of "{(T x1, T x2), (S x1, S x2)}" T x1 x2 S])
by (auto)

(*** Seq_compo_evalT_map_alpha ***)

lemma Seq_compo_evalT_map_alpha:
 "[| map_alpha T alpha ; map_alpha S alpha |]
  ==> map_alpha (%x. (T x) ;;t (S x)) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Seq_compo_evalT_map_alpha_lm)

(*** Seq_compo_evalT_non_expanding ***)

lemma Seq_compo_evalT_non_expanding:
 "[| non_expanding T ; non_expanding S |]
  ==> non_expanding (%x. (T x) ;;t (S x))"
by (simp add: non_expanding_def Seq_compo_evalT_map_alpha)

(*** Seq_compo_evalT_contraction_alpha ***)

lemma Seq_compo_evalT_contraction_alpha:
 "[| contraction_alpha T alpha; contraction_alpha S alpha|]
  ==> contraction_alpha (%x. (T x) ;;t (S x)) alpha"
by (simp add: contraction_alpha_def Seq_compo_evalT_map_alpha)

(*********************************************************
                   map Seq_compo F
 *********************************************************)

(*** restF (subset) ***)

lemma Seq_compo_restF_sub:
   "[| T x1 rest n <= T x2 rest n ;
       F x1 rest n <= F x2 rest n ;
       E x1 rest n <= E x2 rest n |]
    ==> ((T x1, F x1) ;;f (E x1)) rest n <= ((T x2, F x2) ;;f (E x2)) rest n"
apply (simp add: subsetT_iff subsetF_iff)
apply (intro allI impI)
apply (simp add: in_restT in_restF)
apply (simp add: Seq_compo_mem)
apply (elim conjE)

 (* case 1 *)
 apply (erule disjE)
 apply (rotate_tac 1)
 apply (drule_tac x="s" in spec)
 apply (drule_tac x="insert Tick X" in spec, simp)

 (* case 2 *)
 apply (elim conjE exE)
 apply (case_tac "~ notick sa", simp add: not_None_T)
 apply (case_tac "sa @t t = None", simp add: not_None)
 apply (rule disjI2)
 apply (rule_tac x="sa" in exI)
 apply (rule_tac x="t" in exI)
 apply (simp)

  (* < n *)
  apply (erule disjE)
  apply (simp)

  (* = n *)
  apply (elim conjE exE, simp)
  apply (insert trace_last_nil_or_unnil)
  apply (rotate_tac -1)
  apply (drule_tac x="t" in spec, simp add: not_None_F)
  apply (erule disjE, simp)
  apply (elim conjE exE, simp)
  apply (simp add: appt_ass_rev del: appt_ass)

  apply (rotate_tac 2)
  apply (drule_tac x="sb @t [Tick]t" in spec)
  apply (drule_tac x="X" in spec, simp)
  apply (force)
done

(*** restF (equal) ***)

lemma Seq_compo_restF:
   "[| T x1 rest n = T x2 rest n ;
       F x1 rest n = F x2 rest n ;
       E x1 rest n = E x2 rest n |]
    ==> ((T x1, F x1) ;;f (E x1)) rest n = ((T x2, F x2) ;;f (E x2)) rest n"
apply (rule order_antisym)
by (simp_all add: Seq_compo_restF_sub)

(*** distF lemma ***)

lemma Seq_compo_distF:
  "[| TTs = {(T x1, T x2)} ; FFs = {(F x1, F x2), (E x1, E x2)} |]
   ==> (EX TT. TT:TTs & 
             distance((T x1, F x1) ;;f (E x1), (T x2, F x2) ;;f (E x2))
          <= distance((fst TT), (snd TT))) |
       (EX FF. FF:FFs & 
             distance((T x1, F x1) ;;f (E x1), (T x2, F x2) ;;f (E x2))
          <= distance((fst FF), (snd FF)))"
apply (rule rest_to_dist_pair_two)
by (auto intro: Seq_compo_restF)

(*** map_alpha F lemma ***)

lemma Seq_compo_evalF_map_alpha_lm:
  "[| distance (T x1, T x2) <= alpha * distance (x1, x2) ;
      distance (F x1, F x2) <= alpha * distance (x1, x2) ;
      distance (E x1, E x2) <= alpha * distance (x1, x2) |]
    ==> distance((T x1, F x1) ;;f (E x1), (T x2, F x2) ;;f (E x2))
     <= alpha * distance (x1, x2)"
apply (insert Seq_compo_distF
       [of "{(T x1, T x2)}" T x1 x2
           "{(F x1, F x2), (E x1, E x2)}" F E])
by (auto)

(*** Seq_compo_evalF_map_alpha ***)

lemma Seq_compo_evalF_map_alpha:
 "[| map_alpha T alpha ; map_alpha F alpha ; map_alpha E alpha |]
  ==> map_alpha (%x. (T x,F x) ;;f (E x)) alpha"
apply (simp add: map_alpha_def)
apply (intro allI)
apply (erule conjE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
by (simp add: Seq_compo_evalF_map_alpha_lm)

(*** Seq_compo_evalF_non_expanding ***)

lemma Seq_compo_evalF_non_expanding:
 "[| non_expanding T ; non_expanding F ; non_expanding E |]
  ==> non_expanding (%x. (T x,F x) ;;f (E x))"
by (simp add: non_expanding_def Seq_compo_evalF_map_alpha)

(*** Seq_compo_evalF_contraction_alpha ***)

lemma Seq_compo_evalF_contraction_alpha:
 "[| contraction_alpha T alpha; 
     contraction_alpha F alpha; contraction_alpha E alpha|]
  ==> contraction_alpha (%x. (T x,F x) ;;f (E x)) alpha"
by (simp add: contraction_alpha_def Seq_compo_evalF_map_alpha)

(****************** to add them again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

end
