           (*-------------------------------------------*
            |                DFP package                |
            |                 May 2005                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory DFP_Network = DFP_evalF_cut:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

         1. 
         2. 
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                    definitions
 *********************************************************)

(*** network ***)

types ('i,'a) Network  = "('i set * ('i => ('a proc * 'a set)))"

consts
  PAR  :: "('i,'a) Network => 'a proc" ("(1PAR _)" [77] 77)

defs
  PAR_def :
    "PAR V == [||]i:(fst V) .. ((snd V) i)"

syntax
  "@Network" :: 
     "('a proc * 'a set) => pttrn => 'i set
      => ('i,'a) Network"  ("(1{ _ | /_:_ }net)" [77,900,90] 900)

translations
  "{ PX | i:I }net"  == "(I, (%i. PX))"


(*** failure set of network ***)

types ('i,'a) NetworkF = "('i set * ('i => (('a failure set) * 'a set)))"

syntax
  "@NetworkF" :: 
     "(('a failure set) * 'a set) => pttrn => 'i set
                      => ('i,'a) NetworkF"  ("(1{ _ | /_:_ }Fnet)" [77,900,90] 900)

translations
  "{ FX | i:I }Fnet"  == "(I, (%i. FX))"

(*** Network and NetworkF ***)

syntax
  "@Network_pro" :: 
     "('i set * ('i => ('b * 'a set))) => 'i => 'b"
                          ("netElement _<_>" [900,0] 1000)
  "@Network_alp" :: 
     "('i set * ('i => ('b * 'a set))) => 'i => 'a set"
                          ("netAlpha _<_>" [900,0] 1000)

translations
  "netElement V<i>" == "fst ((snd V) i)"
  "netAlpha V<i>"   == "snd ((snd V) i)"

consts
  isFailureOf ::
   "('i,'a) NetworkF => ('i,'a) Network => bool"
                                           ("(1_ /isFailureOf /_)" [55,55] 55)

defs
  isFailureOf_def : 
   "VF isFailureOf V == 
        fst V = fst VF &
       (ALL i : fst VF.
                fst (snd VF i) <=cut[Ev ` snd (snd VF i)]
                [[fst (snd V i)]]F &
                snd (snd V i) = snd (snd VF i))"

(*** short notations ***)

consts
  ALP  :: "('i set * ('i => ('b * 'a set))) => 'a set"

defs
  ALP_def  : "ALP V == {a. EX i:(fst V). a : snd((snd V) i)}"

(*** state ***)

types ('i,'a) net_state = "('a trace * ('i => 'a event set))"

consts
  isStateOf ::
   "('i,'a) net_state => ('i,'a) NetworkF => bool"
                                           ("(1_ /isStateOf /_)" [55,55] 55)
defs
  isStateOf_def : 
   "sigma isStateOf VF == 
     sett(fst sigma) <= Ev ` (ALP VF) &
    (ALL i: fst VF. (fst sigma --tr (- snd ((snd VF) i)), (snd sigma) i) 
                     : fst ((snd VF) i) &
                    (snd sigma) i <= Ev ` snd ((snd VF) i))"

(*********************************************************
                    small lemmas
 *********************************************************)

lemma decompo_V:
  "EX I PXf. (V::('i,'a) Network) = (I,PXf)"
apply (rule_tac x="fst V" in exI)
apply (rule_tac x="snd V" in exI)
by (simp)

lemma isFailureOf_subset_index:
   "[| (I, FXf) isFailureOf (I, PXf) ; J <= I |]
    ==> (J, FXf) isFailureOf (J, PXf)"
apply (simp add: isFailureOf_def)
by (auto)

lemma isFailureOf_subset_alpha1:
   "[| (I, FXf) isFailureOf (I, PXf) ; i : I ; (s, Y) : fst (FXf i) ; e : Y |]
    ==> e : Ev ` snd (PXf i)"
apply (simp add: isFailureOf_def)
apply (drule_tac x="i" in bspec, simp)
apply (simp add: evalF_cut_def)
apply (auto)
done

lemma isFailureOf_subset_alpha2:
   "[| (I, FXf) isFailureOf (I, PXf) ; i : I ; (s, Y) : fst (FXf i) ; e : Y |]
    ==> e : Ev ` snd (FXf i)"
apply (simp add: isFailureOf_def)
apply (drule_tac x="i" in bspec, simp)
apply (simp add: evalF_cut_def)
apply (auto)
done

lemmas isFailureOf_subset_alpha = isFailureOf_subset_alpha1
                                  isFailureOf_subset_alpha2

lemma isFailureOf_not_Tick[simp]:
   "[| (I, FXf) isFailureOf (I, PXf) ; i : I ; (s, Y) : fst (FXf i) |]
    ==> Tick ~: Y"
apply (simp add: isFailureOf_def)
apply (drule_tac x="i" in bspec, simp)
apply (simp add: evalF_cut_def)
apply (auto)
done

lemma isFailureOf_same_alpha:
   "(I, FXf) isFailureOf (I, PXf)
    ==> ALL i:I. snd (PXf i) = snd (FXf i)"
by (simp add: isFailureOf_def)

(*********************************************************
                 StateOf subset
 *********************************************************)

lemma isStateOf_subset_alpha1:
   "[| (I, FXf) isFailureOf (I, PXf) ; 
       (t, Yf) isStateOf (I, FXf) ;
       i : I ; e : Yf i|]
    ==> e : (Ev ` (snd (PXf i)))"
apply (simp add: isStateOf_def)
apply (auto simp add: isFailureOf_same_alpha)
done

lemma isStateOf_subset_alpha2:
   "[| (I, FXf) isFailureOf (I, PXf) ; 
       (t, Yf) isStateOf (I, FXf) ;
       i : I ; e : Yf i|]
    ==> e : (Ev ` (snd (FXf i)))"
apply (simp add: isStateOf_def)
apply (auto)
done

lemmas isStateOf_subset_alpha = isStateOf_subset_alpha1 
                                isStateOf_subset_alpha2

lemma isStateOf_each_element:
   "[| (t, Yf) isStateOf (I, FXf) ; i : I |]
    ==> (t --tr (- snd (FXf i)), Yf) isStateOf ({i}, FXf)"
apply (simp add: isStateOf_def)
apply (rule conjI)

 apply (simp add: ALP_def)
 apply (insert hide_tr_sett_comp_subseteq[of t "snd (FXf i)"])
 apply (case_tac "Tick : sett t")
  apply (force)
  apply (subgoal_tac "sett (t --tr (- snd (FXf i))) <= sett (t --tr (- UNIV))")
  apply (case_tac "Tick : sett (t --tr (- snd (FXf i)))", simp)
  apply (blast)
  apply (rule hide_tr_sett_subseteq_sett_comp)
  apply (simp)

apply (simp add: hide_tr_of_hide_tr_subset)
done

(*********************************************************
                        PAR
 *********************************************************)

(*---------------------------------*
 |           flattening            |
 *---------------------------------*)

(*** SF ***)

(* lemmas *)

lemma domSF_PAR_flattening_lm1:
   "ALP (V x) = Union (snd ` (%(i, j). snd (V j) i) ` {(i, x) |i. i : fst (V x)})"
by (auto simp add: ALP_def)

lemma domSF_PAR_flattening_lm2:
   "Union (snd ` (%i. ([||]:(fst (V i)) (snd (V i)), ALP (V i))) ` F) =
    Union (snd ` (%(i, j). snd (V j) i) ` {(i, j). j : F & i : fst (V j)})"
apply (auto simp add: ALP_def)
apply (rule_tac x="ia" in exI)
apply (rule_tac x="i" in exI)
apply (simp)
apply (rule_tac x="ba" in bexI)
apply (rule_tac x="aa" in bexI)
apply (simp_all)
done

(* main *)

lemma domSF_PAR_flattening:
   "[| finite J ; ALL j:J. finite (fst (V j)) |] ==>
       [[PAR { (PAR (V j), ALP (V j)) | j:J }net]]SF =
       [[PAR { (snd (V j)) i | (i,j):{(i, j). j : J & i : fst (V j)} }net]]SF"
apply (induct set: Finites)

(* base *)
apply (simp add: PAR_def)

(* step *)
apply (simp add: PAR_def)
apply (simp add: domSF_Index_parallel_induct)

apply (subgoal_tac 
   "[[[||]:(fst (V x)) (snd (V x))
              |[ALP (V x),
                Union (snd ` (%i. ([||]:(fst (V i)) (snd (V i)), ALP (V i))) ` F)]|
      ([||] i:F .. ([||]:(fst (V i)) (snd (V i)), ALP (V i)))]]SF
  = [[[||]:(fst (V x)) (snd (V x))
              |[ALP (V x),
                Union (snd ` (%i. ([||]:(fst (V i)) (snd (V i)), ALP (V i))) ` F)]|
      ([||] (i, j):{(i, j). j : F & i : (fst (V j))} .. (snd (V j)) i)]]SF")
                                      (* ... sub 1 *)
apply (simp)

apply (subgoal_tac 
   "{(i, j). (j = x | j : F) & i : fst (V j)} = 
    {(i,x) |i. i : fst (V x)} Un {(i, j). j : F & i : fst (V j)}")
                                      (* ... sub 2 *)
apply (simp)

apply (subgoal_tac "finite {(i, x)|i. i : fst (V x)}")
                                      (* ... sub 3 *)
apply (subgoal_tac "finite {(i, j). j : F & i : fst (V j)}")
                                      (* ... sub 4 *)
apply (subgoal_tac
   "{(i, x)|i. i : fst (V x)} Int {(i, j). j : F & i : fst (V j)} = {}")
                                      (* ... sub 5 *)
apply (simp add: domSF_Index_parallel_ass)
apply (rule Alpha_parallel_evalSF_cong)
apply (simp add: domSF_PAR_flattening_lm1)
apply (simp add: domSF_PAR_flattening_lm2)

apply (rule domSF_Index_parallel_index_eq)
 apply (simp)
 apply (rule_tac x="(%i . (i, x))" in exI)
 apply (simp)
 apply (force)
 apply (simp)

(* sub 5 *)
apply (force)

(* sub 4 *)
apply (simp add: finite_pair_set)

(* sub 3 *)
apply (subgoal_tac "{(i, x) |i. i : fst (V x)} = (%i. (i,x)) ` (fst (V x))")
apply (force)
apply (force)

(* sub 2 *)
apply (force)

(* sub 1 *)
apply (rule Alpha_parallel_evalSF_cong)
apply (simp_all)
done

(*------------------*
 |      csp law     |
 *------------------*)

lemma csp_PAR_flattening:
   "[| finite J ; ALL j:J. finite (fst (V j)) |] 
    ==> (PAR { (PAR (V j), ALP (V j)) | j:J }net) =SF 
        (PAR { (snd (V j)) i | (i,j):{(i, j). j : J & i : fst (V j)}}net)"
by (simp add: eqSF_def domSF_PAR_flattening)

(*** T and F ***)

lemma domT_F_PAR_flattening:
   "[| finite J ; ALL j:J. finite (fst (V j)) |] ==>
       [[PAR { (PAR (V j), ALP (V j)) | j:J }net]]T =
       [[PAR { (snd (V j)) i | (i,j):{(i, j). j : J & i : fst (V j)} }net]]T &
       [[PAR { (PAR (V j), ALP (V j)) | j:J }net]]F =
       [[PAR { (snd (V j)) i | (i,j):{(i, j). j : J & i : fst (V j)} }net]]F"
apply (simp add: proc_eqSF_decompo[THEN sym])
apply (simp add: domSF_PAR_flattening)
done

lemma domT_PAR_flattening:
   "[| finite J ; ALL j:J. finite (fst (V j)) |] ==>
       [[PAR { (PAR (V j), ALP (V j)) | j:J }net]]T =
       [[PAR { (snd (V j)) i | (i,j):{(i, j). j : J & i : fst (V j)} }net]]T"
by (simp add: domT_F_PAR_flattening)

lemma domF_PAR_flattening:
   "[| finite J ; ALL j:J. finite (fst (V j)) |] ==>
       [[PAR { (PAR (V j), ALP (V j)) | j:J }net]]F =
       [[PAR { (snd (V j)) i | (i,j):{(i, j). j : J & i : fst (V j)} }net]]F"
by (simp add: domT_F_PAR_flattening)

(*********************************************************
                      sub network
 *********************************************************)

(*** state ***)

lemma isStateOf_subset: 
  "[| (t,Yf) isStateOf (I,FXf) ; J <= I ; 
      X = {sY. EX i:J. sY : snd (FXf i)} |]
  ==> (t --tr (- X),Yf) isStateOf (J,FXf)"
apply (simp add: isStateOf_def ALP_def)
apply (rule conjI)
apply (insert hide_tr_sett_comp_subseteq[of t "X"])
apply (subgoal_tac "Tick ~: sett (t --tr (- X))")
apply (blast)
apply (force)

apply (rule ballI)
apply (elim conjE)
apply (drule_tac x="i" in bspec, fast)

apply (subgoal_tac "snd (FXf i) <= {sY. EX i:J. sY : snd (FXf i)}")
apply (simp add: hide_tr_of_hide_tr_subset)
by (auto)

lemma isStateOf_subsetI:
  "[| (t,Yf) isStateOf (I,FXf) ; 
      J <= I ; X = Union {snd (FXf i)|i. i:J} |]
  ==> (t --tr (- X),Yf) isStateOf (J,FXf)"
apply (subgoal_tac "Union {snd (FXf i)|i. i:J} = {sY. EX i:J. sY : snd (FXf i)}")
apply (simp add: isStateOf_subset)
by (auto)

(*********************************************************
                      isFailureOf
 *********************************************************)

lemma isFailureOf_largest:
   "{ ({(s,Y Int (Ev ` snd (PXf i)))|s Y. 
        (s,Y) :f [[fst (PXf i)]]F}, snd (PXf i)) | i:I }Fnet 
     isFailureOf (I, PXf)"
apply (simp add: isFailureOf_def)
apply (simp add: evalF_cut_largest)
done

lemma isFailureOf_EX: "EX VF. VF isFailureOf V"
apply (rule_tac x=
  "{ ({(s,Y Int (Ev ` snd ((snd V) i))) |s Y. 
        (s,Y) :f [[fst ((snd V) i)]]F}, 
      snd ((snd V) i)) | i: fst V }Fnet" in exI)
apply (insert decompo_V[of V])
apply (elim exE)
apply (simp add: isFailureOf_largest)
done

(****************** to add it again ******************)

declare disj_not1   [simp]
declare not_None_eq [simp]

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
