           (*-------------------------------------------*
            |                DFP package                |
            |                 June 2005                 |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory DFP_evalF_cut = CSP:

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*****************************************************************

  1. safe subsets of [[P]]F, used for deadlock-free verification.

 *****************************************************************)

(*********************************************************
                    safe cut of [[_]]F
        This is used for deadlock-free verification.
 *********************************************************)

consts
  evalF_cut  :: "'a failure set => 'a event set => 'a domF => bool"
                                     ("(0_/ <=cut[_] /_)" [50,0,50] 50)

defs
  evalF_cut_def :
    "FS <=cut[A] FD == 
       (ALL s Y. ((s,Y) : FS  --> ((s,Y) :f FD) & Y <= A)) &
       (ALL s Y. (s,Y) :f FD --> (EX Z. (s,Z) : FS & (Y Int A) <= Z))"

(*********************************************************
            How to prove FS <=cut [[P]]F ev
 *********************************************************)

(*--------------------*
 |        DIV         |
 *--------------------*)

lemma evalF_cut_DIV[simp]: "{} <=cut[A] [[DIV]]F"
apply (simp add: evalF_cut_def)
apply (simp add: DIV_memF)
done

(*-----------*
 | csp rules |
 *-----------*)

lemma csp_evalF_cut_DIV: 
  "P =SF DIV ==> {} <=cut[A] [[P]]F"
by (simp add: eqSF_def eqSF_decompo)

(*--------------------*
 |     Int_choice     |
 *--------------------*)

(* eq *)

lemma evalF_cut_Int_choice: 
  "[| FS1 <=cut[A] [[P1]]F ; FS2 <=cut[A] [[P2]]F |]
    ==> FS1 Un FS2 <=cut[A] [[ P1 |~| P2]]F"
apply (simp add: evalF_cut_def)
apply (simp add: Int_choice_mem)
apply (intro allI impI)
apply (elim conjE)
apply (rule conjI)

 apply (intro impI)
 apply (rotate_tac 1)
 apply (drule_tac x="s" in spec)
 apply (drule_tac x="Y" in spec)
 apply (fast)

 apply (intro impI)
 apply (rotate_tac 3)
 apply (drule_tac x="s" in spec)
 apply (drule_tac x="Y" in spec)
 apply (fast)
done

(*-----------*
 | csp rules |
 *-----------*)

lemma csp_evalF_cut_Int_choice: 
  "[| P =SF P1 |~| P2 ;
      FS = FS1 Un FS2 ;
      FS1 <=cut[A] [[P1]]F ;
      FS2 <=cut[A] [[P2]]F |]
   ==> FS <=cut[A] [[P]]F"
by (simp add: eqSF_def eqSF_decompo evalF_cut_Int_choice)

(*--------------------*
 |   Rep_int_choice   |
 *--------------------*)

(* eq *)

lemma evalF_cut_Rep_int_choice: 
  "ALL a : X. FSf a <=cut[A] [[Pf a]]F
    ==> Union {FSf a |a. a : X} <=cut[A] [[!! :X .. Pf]]F"
apply (simp add: evalF_cut_def)
apply (simp add: Rep_int_choice_mem)
apply (rule conjI)

 (* 1 *)
 apply (intro allI impI)
 apply (elim conjE exE)
 apply (drule_tac x="a" in bspec, simp, simp)
 apply (rule_tac x="a" in bexI, simp, simp)

 (* 2 *)
 apply (intro allI impI)
 apply (elim conjE bexE)
 apply (drule_tac x="a" in bspec, simp)
 apply (elim conjE)
 apply (rotate_tac -1)
 apply (drule_tac x="s" in spec)
 apply (drule_tac x="Y" in spec)
 apply (auto)
done

(*-----------*
 | csp rules |
 *-----------*)

lemma csp_evalF_cut_Rep_int_choice0: 
  "[| P =SF !! :X .. Pf ;
      FS = Union {FSf a |a. a : X} ;
      ALL a : X. FSf a <=cut[A] [[Pf a]]F |]
   ==> FS <=cut[A] [[P]]F"
by (simp add: eqSF_def eqSF_decompo evalF_cut_Rep_int_choice)

lemma csp_evalF_cut_Rep_int_choice1: 
  "[| P =SF ! :X .. Pf ;
      FS = Union {FSf a |a. a : X} ;
      ALL a : X. FSf a <=cut[A] [[Pf a]]F |]
   ==> FS <=cut[A] [[P]]F"
apply (rule csp_evalF_cut_Rep_int_choice0
      [of P "(OR1 ` X)" "(%a. Pf (inv OR1 a))" FS "(%a. FSf (inv OR1 a))" A])
apply (simp_all add: Rep_int_choice_OR1_def)
apply (simp add: image_iff)
apply (simp add: Union_def)
apply (auto)
apply (rule_tac x="FSf aa" in exI)
apply (simp)
apply (rule_tac x="OR1 aa" in exI)
apply (simp)
done

lemma csp_evalF_cut_Rep_int_choice2: 
  "[| P =SF !nat :X .. Pf ;
      FS = Union {FSf a |a. a : X} ;
      ALL a : X. FSf a <=cut[A] [[Pf a]]F |]
   ==> FS <=cut[A] [[P]]F"
apply (rule csp_evalF_cut_Rep_int_choice0
      [of P "(OR2 ` X)" "(%a. Pf (inv OR2 a))" FS "(%a. FSf (inv OR2 a))" A])
apply (simp_all add: Rep_int_choice_OR2_def)
apply (simp add: image_iff)
apply (simp add: Union_def)
apply (auto)
apply (rule_tac x="FSf aa" in exI)
apply (simp)
apply (rule_tac x="OR2 aa" in exI)
apply (simp)
done

lemmas csp_evalF_cut_Rep_int_choice = csp_evalF_cut_Rep_int_choice1
lemmas csp_evalF_cut_Rep_int_choice_nat = csp_evalF_cut_Rep_int_choice2

lemma csp_evalF_cut_Rep_int_choice_nat_UNIV: 
  "[| P =SF !nat n .. Pf n;
      FS = Union {FSf n |n. True} ;
      ALL n. FSf n <=cut[A] [[Pf n]]F |]
   ==> FS <=cut[A] [[P]]F"
apply (insert csp_evalF_cut_Rep_int_choice_nat[of P UNIV Pf FS FSf A])
by (simp_all)

(*--------------------*
 |   Ext_pre_choice   |
 *--------------------*)

(* eq *)

lemma evalF_cut_Ext_pre_choice: 
  "ALL a : X. FSf a <=cut[A] [[Pf a]]F
    ==> insert ([]t, A - Ev ` X) 
        {([Ev a]t @t s, Y) |a s Y. (s,Y) : FSf a & a : X}
        <=cut[A] [[? :X -> Pf]]F"
apply (simp add: evalF_cut_def)
apply (rule conjI)

 (* 1 *)
 apply (simp add: Ext_pre_choice_mem)
 apply (intro allI)
 apply (rule conjI)
 apply (fast)

 apply (intro impI)
 apply (elim conjE exE)
 apply (drule_tac x="a" in bspec, simp)
 apply (simp)
 apply (fast)

 (* 2 *)
 apply (simp add: Ext_pre_choice_mem)
 apply (intro allI impI)
 apply (rule conjI)
  apply (fast)

  apply (intro impI)
  apply (elim conjE exE)
  apply (drule_tac x="a" in bspec, simp)
  apply (elim conjE)
  apply (rotate_tac -1)
  apply (drule_tac x="sa" in spec)
  apply (drule_tac x="Y" in spec)
  apply (simp)
  apply (fast)
done

(*-----------*
 | csp rules |
 *-----------*)

lemma csp_evalF_cut_Ext_pre_choice: 
  "[| P =SF ? :X -> Pf ;
      FS = insert ([]t, A - Ev ` X) 
           {([Ev a]t @t s, Y) |a s Y. (s,Y) : FSf a & a : X} ;
      ALL a : X. FSf a <=cut[A] [[Pf a]]F |]
   ==> FS <=cut[A] [[P]]F"
by (simp add: eqSF_def eqSF_decompo evalF_cut_Ext_pre_choice)

(*-----------*
 | csp rules |
 *-----------*)

lemma csp_evalF_cut_eqSF:
  "[| P =SF Q ;
       FS <=cut[A] [[Q]]F |]
   ==> FS <=cut[A] [[P]]F"
by (simp add: eqSF_def eqSF_decompo)

(*********************************************************
                evalF_cut existency
 *********************************************************)

lemma evalF_cut_largest:
   "{(s,Y Int A)|s Y. (s,Y) :f [[P]]F}
    <=cut[A] [[P]]F"
apply (simp add: evalF_cut_def)
apply (rule conjI)

 (* 1 *)
 apply (intro allI impI)
 apply (elim conjE exE, simp)
 apply (rule conjI)
 apply (rule proc_F2)
 apply (simp)
 apply (fast)
 apply (fast)

 (* 2 *)
 apply (intro allI impI)
 apply (rule_tac x="Y Int A" in exI)
 apply (auto)
done

(****************** to add them again ******************)

declare not_None_eq    [simp]

end

