           (*-------------------------------------------*
            |    Example 1 [Roscoe_Dathi_1987 P.10]     |
            |             WITH computation              |
            |  Self-timed version of a systolic array   |
            |                 June 2005                 |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory SA_definition = DFP + Ring_and_Field:

(*  The following simplification is sometimes unexpected.              *)
(*                                                                     *)
(*             not_None_eq: (x ~= None) = (EX y. x = Some y)           *)

declare not_None_eq [simp del]

(*********************************************************
                         event
 *********************************************************)

datatype 'r Event = vert "nat * nat" 'r | hori "nat * nat" 'r

(*********************************************************
                  element process
 *********************************************************)

types index_type = "nat * nat"

datatype 'r ProcName = pe "index_type" 'r | pe' "index_type" 'r 'r 'r

consts 
  Def  :: "(('r::ring) ProcName, ('r::ring) Event) procXnc"
  Fun :: "(('r::ring) ProcName, ('r::ring) Event) procX"
  Fix :: "('r::ring) ProcName => ('r::ring) Event proc"

recdef Def "{}"
  "Def (X, (pe (i,j) r)) = (vert (i,j) ? x -> hori (i,j) ? y -> (X (pe' (i,j) r x y))) [+]
                           (hori (i,j) ? y -> vert (i,j) ? x -> (X (pe' (i,j) r x y)))"

  "Def (X, (pe' (i,j) r x y)) = ((vert (i+1,j) ! x -> SKIP) ||| 
                                 (hori (i,j+1) ! y -> SKIP)) ;;
                                (X (pe (i,j) (r + x * y)))"

defs Fun_def : "Fun == curry Def"

recdef Fix "{}" "Fix n = (FIX Fun) n"

(*** ProcX ***)

lemma Fun_ProcX[simp]: "Fun : ProcX"
apply (simp add: ProcX_def)
apply (rule, induct_tac pn)
by (auto simp add: Fun_def csp_ss_def)

consts
  Alpha_pe :: "index_type => ('r::ring) Event set"

recdef Alpha_pe "{}"
  "Alpha_pe (i,j) = {a. EX x. a = vert (i  ,j) x | a = hori (i,j  ) x |
                              a = vert (i+1,j) x | a = hori (i,j+1) x }"

declare Alpha_pe.simps [simp del]

consts
  Array_Index :: "nat => index_type set"

defs
  Array_Index_def : "Array_Index N == {(i,j)|i j. i < N & j < N}"

consts 
  Systolic_Array :: "nat => (index_type,  ('r::ring) Event) Network"

defs Systolic_Array_def: 
  "Systolic_Array N == 
   { (Fix (pe (i,j) 0), Alpha_pe (i,j)) | (i,j) : (Array_Index N)}net"


(*---------------------------------*
 | Failures sets of Systolic_Array |
 *---------------------------------*)

(*** out term ***)

consts 
  Faiures_out_hori :: "('r::ring) => index_type
                         => 'r Event failure set => 'r Event failure set"
  Faiures_out_vert :: "('r::ring) => index_type
                         => 'r Event failure set => 'r Event failure set"
  Faiures_out:: "('r::ring) => 'r => index_type 
                         => 'r Event failure set => 'r Event failure set"

defs 
  Faiures_out_hori_def :
  "Faiures_out_hori y ==
   (%(i,j). (%F.
      {([]t, {Ev a |a. EX z. (a = vert (i,j) z | a = hori (i,j) z | a = vert (i+1,j) z |
                              (a = hori (i,j+1) z & z ~= y))})} Un
      {([Ev (hori (i,j+1) y)]t @t s, Y) |s Y. (s, Y) : F}))"

  Faiures_out_vert_def :
  "Faiures_out_vert x ==
   (%(i,j). (%F.
      {([]t, {Ev a |a. EX z. (a = hori (i,j) z | a = vert (i,j) z | a = hori (i,j+1) z |
                              (a = vert (i+1,j) z & z ~= x))})} Un
      {([Ev (vert (i+1,j) x)]t @t s, Y) |s Y. (s, Y) : F}))"

  Faiures_out_def :
  "Faiures_out x y ==
   (%(i,j). (%F.
      {([]t, {Ev a |a. EX z. (a = hori (i,j) z | a = vert (i,j) z |
                              (a = hori (i,j+1) z & z ~= y) |
                              (a = vert (i+1,j) z & z ~= x))})} Un
      {([Ev (vert (i+1,j) x)]t @t s, Y) |s Y. (s, Y) : Faiures_out_hori y (i,j) F} Un
      {([Ev (hori (i,j+1) y)]t @t s, Y) |s Y. (s, Y) : Faiures_out_vert x (i,j) F}))"

(*** in term ***)

consts 
  Faiures_in_hori :: "('r::ring) => index_type
                       => 'r Event failure set => 'r Event failure set"
  Faiures_in_vert :: "('r::ring) => index_type
                       => 'r Event failure set => 'r Event failure set"
  Faiures_in      :: "index_type => ('r::ring) Event failure set => 'r Event failure set"

defs 
  Faiures_in_hori_def :
  "Faiures_in_hori x ==
   (%(i,j) F.
      {([]t, {Ev a |a. EX x. (a = vert (i,j) x | a = hori (i,j+1) x | 
                              a = vert (i+1,j) x)})} Un
      {([Ev (hori (i,j) y)]t @t s, Y) |y s Y. (s, Y) : Faiures_out x y (i,j) F})"

  Faiures_in_vert_def :
  "Faiures_in_vert y ==
   (%(i,j) F.
      {([]t, {Ev a |a. EX x. (a = hori (i,j) x | a = vert (i+1,j) x | 
                              a = hori (i,j+1) x)})} Un
      {([Ev (vert (i,j) x)]t @t s, Y) |x s Y. (s, Y) : Faiures_out x y (i,j) F})"

  Faiures_in_def :
  "Faiures_in ==
   (%(i,j) F.
      {([]t, {Ev a |a. EX x. (a = vert (i+1,j) x | a = hori (i,j+1) x)})} Un
      {([Ev (vert (i,j) x)]t @t s, Y) |x s Y. (s, Y) : Faiures_in_hori x (i,j) F} Un
      {([Ev (hori (i,j) y)]t @t s, Y) |y s Y. (s, Y) : Faiures_in_vert y (i,j) F})"

(*** induction ***)

consts 
  peF_rec :: "nat => index_type => ('r::ring) Event failure set"
  peF     :: "index_type => ('r::ring) Event failure set"

primrec
  "peF_rec 0       = (%ij. {})"
  "peF_rec (Suc n) = (%ij. Faiures_in ij (peF_rec n ij))"

defs
   peF_def : "peF ij == Union {peF_rec n ij |n. True}"

(*** NetworkF ***)

consts 
  Systolic_ArrayF :: "nat => (index_type, ('r::ring) Event) NetworkF"

defs Systolic_ArrayF_def: 
  "Systolic_ArrayF == 
   (%N. { (peF ij, Alpha_pe ij) | ij : (Array_Index N)}Fnet)"

(******************************************************************
                        small lemmas 
 ******************************************************************)

(* index *)

lemma Example1_index_mem:
  "ij : Array_Index N = (EX i j. ij = (i,j) & i < N & j < N)"
apply (simp add: Array_Index_def)
by (auto)

(* trace *)

lemma Faiures_out_hori_F:
  "[| (s, X) : Faiures_out_hori y (i,j) F ; ALL s X. (s, X) : F --> s ~= None |]
   ==> s ~= None"
apply (simp add: Faiures_out_hori_def)
apply (elim disjE conjE exE)
by (auto)

lemma Faiures_out_vert_F:
  "[| (s, X) : Faiures_out_vert x (i,j) F ; ALL s X. (s, X) : F --> s ~= None |]
   ==> s ~= None"
apply (simp add: Faiures_out_vert_def)
apply (elim disjE conjE exE)
by (auto)

lemma Faiures_out_F:
  "[| (s, X) : Faiures_out x y (i,j) F ; ALL s X. (s, X) : F --> s ~= None |]
   ==> s ~= None"
apply (simp add: Faiures_out_def)
apply (elim disjE conjE exE)
by (auto simp add: Faiures_out_hori_F Faiures_out_vert_F)

lemma Faiures_in_hori_F:
  "[| (s, X) : Faiures_in_hori x (i,j) F ; ALL s X. (s, X) : F --> s ~= None |]
   ==> s ~= None"
apply (simp add: Faiures_in_hori_def)
apply (elim disjE conjE exE)
by (auto simp add: Faiures_out_F)

lemma Faiures_in_vert_F:
  "[| (s, X) : Faiures_in_vert y (i,j) F ; ALL s X. (s, X) : F --> s ~= None |]
   ==> s ~= None"
apply (simp add: Faiures_in_vert_def)
apply (elim disjE conjE exE)
by (auto simp add: Faiures_out_F)

lemma Faiures_in_F:
  "[| (s, X) : Faiures_in (i,j) F ; ALL s X. (s, X) : F --> s ~= None |]
   ==> s ~= None"
apply (simp add: Faiures_in_def)
apply (elim disjE conjE exE)
by (auto simp add: Faiures_in_hori_F Faiures_in_vert_F)

lemma peF_rec_not_None_ALL: 
  "ALL s X. (s, X) : peF_rec n (i,j) --> s ~= None"
apply (induct_tac n)
apply (simp_all add: Faiures_in_F)
done

lemma peF_rec_not_None[simp]: 
  "(s, X) : peF_rec n (i,j) ==> s ~= None"
by (simp add: peF_rec_not_None_ALL)

lemma peF_not_None[simp]: "(s,X): peF (i,j) ==> s ~= None"
apply (simp add: peF_def)
apply (elim conjE exE)
apply (simp)
done

lemma Faiures_out_hori[simp]:
  "(s, X) : Faiures_out_hori y (i,j) (peF_rec n (i,j))
   ==> s ~= None"
by (simp add: Faiures_out_hori_F)

lemma Faiures_out_vert[simp]:
  "(s, X) : Faiures_out_vert x (i,j) (peF_rec n (i,j))
   ==> s ~= None"
by (simp add: Faiures_out_vert_F)

lemma Faiures_out[simp]:
  "(s, X) : Faiures_out x y (i,j) (peF_rec n (i,j))
   ==> s ~= None"
by (simp add: Faiures_out_F)

lemma Faiures_in_hori[simp]:
  "(s, X) : Faiures_in_hori x (i,j) (peF_rec n (i,j))
   ==> s ~= None"
by (simp add: Faiures_in_hori_F)

lemma Faiures_in_vert[simp]:
  "(s, X) : Faiures_in_vert y (i,j) (peF_rec n (i,j))
   ==> s ~= None"
by (simp add: Faiures_in_vert_F)

lemma Faiures_in[simp]:
  "(s, X) : Faiures_in (i,j) (peF_rec n (i,j))
   ==> s ~= None"
by (simp add: Faiures_in_F)

(****************** to add it again ******************)

declare not_None_eq [simp]

end
